/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkitemfactory.h>
#include <gtk/gtkcheckmenuitem.h>
#include <gtk/gtkoptionmenu.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkclist.h>
#include <gtk/gtkctree.h>
#include <gtk/gtkvpaned.h>
#include <gtk/gtktext.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkeditable.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkhandlebox.h>
#include <gtk/gtktoolbar.h>
#include <gtk/gtktable.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkthemes.h>
#include <gtk/gtkdnd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#if (HAVE_WCTYPE_H && HAVE_WCHAR_H)
#  include <wchar.h>
#  include <wctype.h>
#endif

#include "main.h"
#include "mainwindow.h"
#include "compose.h"
#include "addressbook.h"
#include "procmsg.h"
#include "menu.h"
#include "send.h"
#include "prefs_common.h"
#include "prefs_account.h"
#include "account.h"
#include "filesel.h"
#include "procheader.h"
#include "procmime.h"
#include "statusbar.h"
#include "about.h"
#include "base64.h"
#include "cpconfigparse.h"
#include "conv_iconv.h"
#include "utils.h"
#include "gtkutils.h"
#include "socket.h"
#include "alertpanel.h"
#include "manage_window.h"
#include "gtkspell.h"

#include "pixmaps/stock_mail_compose_icon.xpm"

extern GList *addresbook_recursive_list_items(GtkCTreeNode *node);
extern GList *account_list_addresses(void);
FolderInfo *m_finfo;
typedef enum
{
	COL_MIMETYPE = 0,
	COL_SIZE     = 1,
	COL_NAME     = 2
} AttachColumnPos;

#define N_ATTACH_COLS	3

#define B64_LINE_SIZE	57
#define B64_BUFFSIZE	77


static GList *compose_list = NULL;

static Compose *compose_create			(void);
static void compose_toolbar_create		(Compose	*compose,
						 GtkWidget	*container);
static void compose_destroy			(Compose	*compose);

static gint compose_parse_header		(Compose	*compose,
						 MsgInfo	*msginfo);
static void compose_quote_parse_fmt		(Compose	*compose,
						 MsgInfo	*msginfo);

static void compose_reply_set_entry		(Compose	*compose,
						 MsgInfo	*msginfo);
static void compose_reedit_set_entry		(Compose	*compose,
						 MsgInfo	*msginfo);

static void compose_insert_sig			(Compose	*compose);
static void compose_insert_file			(Compose	*compose,
						 const gchar	*file, gboolean isexec);
static void compose_attach_append		(Compose	*compose,
						 const gchar	*file,
						 ContentType	 cnttype);

static void compose_wrap_line			(Compose	*compose);

static gint compose_send			(Compose	*compose);
static gint compose_write_to_file_for_smtp	(Compose	*compose,
						 const gchar	*file,
						 gboolean	 is_draft);

static gint compose_write_to_file_for_nntp	(Compose	*compose,
						 const gchar	*file,
						 gboolean	 is_draft);

static gint compose_write_to_file_for_queue	(Compose	*compose,
						 const gchar	*file,
						 gboolean	 is_draft);

static gint compose_write_body_to_file		(Compose	*compose,
						 const gchar	*file);
static gint compose_save_to_outbox		(Compose	*compose,
						 const gchar	*file);
static gint compose_queue			(Compose	*compose,
						 const gchar	*file);
static void compose_write_attach		(Compose	*compose,
						 FILE		*fp);
static gint compose_write_headers_smtp		(Compose	*compose,
						 FILE		*fp,
						 const gchar	*charset,
						 EncodingType	 encoding,
						 gboolean	 is_draft);

static gint compose_write_headers_nntp		(Compose	*compose,
						 FILE		*fp,
						 const gchar	*charset,
						 EncodingType	 encoding,
						 gboolean	 is_draft);

static gint compose_write_headers_queue		(Compose 	*compose,
						 FILE 		*fp,
						 const gchar 	*charset,
						 EncodingType 	encoding,
						 gboolean 	is_draft);

static void compose_convert_header		(gchar		*dest,
						 gint		 len,
						 gchar		*src,
						 gint		 header_len);
static void compose_convert_header_misc 	(gchar		*dest,
						 gint		 len,
						 gchar		*src,
						 gint		 header_len);
static void compose_generate_msgid		(gchar		*buf,
						 gint		 len);

static void compose_attach_info_free		(AttachInfo	*ainfo);
static void compose_attach_remove_selected	(Compose	*compose);

static void compose_attach_property		(Compose	*compose);
static void compose_attach_property_create	(gboolean	*cancelled);
static void attach_property_ok			(GtkWidget	*widget,
						 gboolean	*cancelled);
static void attach_property_cancel		(GtkWidget	*widget,
						 gboolean	*cancelled);
static gint attach_property_delete_event	(GtkWidget	*widget,
						 GdkEventAny	*event,
						 gboolean	*cancelled);
static void attach_property_key_pressed		(GtkWidget	*widget,
						 GdkEventKey	*event,
						 gboolean	*cancelled);

static void compose_exec_ext_editor		(Compose	   *compose);
static gint compose_exec_ext_editor_real	(const gchar	   *file);
static gboolean compose_ext_editor_kill		(Compose	   *compose);
static void compose_input_cb			(gpointer	    data,
						 gint		    source,
						 GdkInputCondition  condition);
static void compose_set_ext_editor_sensitive	(Compose	   *compose,
						 gboolean	    sensitive);

/* callback functions */

static void toolbar_send_cb		(GtkWidget	*widget,
					 gpointer	 data);
static void toolbar_send_later_cb	(GtkWidget	*widget,
					 gpointer	 data);
static void toolbar_draft_cb		(GtkWidget	*widget,
					 gpointer	 data);
static void toolbar_insert_cb		(GtkWidget	*widget,
					 gpointer	 data);
static void toolbar_attach_cb		(GtkWidget	*widget,
					 gpointer	 data);
static void toolbar_sig_cb		(GtkWidget	*widget,
					 gpointer	 data);
static void toolbar_ext_editor_cb	(GtkWidget	*widget,
					 gpointer	 data);

static void toolbar_linewrap_cb		(GtkWidget	*widget,
					 gpointer	 data);

static void toolbar_checkspell_cb	(GtkWidget	*widget,
					 gpointer	 data);

static void toolbar_address_cb		(GtkWidget	*widget,
					 gpointer	 data);

static void attach_selected		(GtkCList	*clist,
					 gint		 row,
					 gint		 column,
					 GdkEvent	*event,
					 gpointer	 data);
static void attach_button_pressed	(GtkWidget	*widget,
					 GdkEventButton	*event,
					 gpointer	 data);
static void attach_key_pressed		(GtkWidget	*widget,
					 GdkEventKey	*event,
					 gpointer	 data);

static void compose_send_cb		(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_send_later_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_draft_cb		(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_attach_cb		(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_insert_file_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);

static void compose_close_cb		(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_address_cb		(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);

static void compose_ext_editor_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);

static void compose_checkspell_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);

static gint compose_delete_cb		(GtkWidget	*widget,
					 GdkEventAny	*event,
					 gpointer	 data);
static void compose_destroy_cb		(GtkWidget	*widget,
					 Compose	*compose);

static void compose_cut_cb		(Compose	*compose);
static void compose_copy_cb		(Compose	*compose);
static void compose_paste_cb		(Compose	*compose);
static void compose_allsel_cb		(Compose	*compose);

static void compose_grab_focus_cb	(GtkWidget	*widget,
					 Compose	*compose);

static void compose_changed_cb		(GtkEditable	*editable,
					 Compose	*compose);
static void compose_button_press_cb	(GtkWidget	*widget,
					 GdkEventButton	*event,
					 Compose	*compose);
#if 0
static void compose_key_press_cb	(GtkWidget	*widget,
					 GdkEventKey	*event,
					 Compose	*compose);
#endif

static void compose_toggle_bcc_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_toggle_replyto_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);
static void compose_toggle_attach_cb	(gpointer	 data,
					 guint		 action,
					 GtkWidget	*widget);

static void compose_attach_drag_received_cb (GtkWidget		*widget,
					     GdkDragContext	*drag_context,
					     gint		 x,
					     gint		 y,
					     GtkSelectionData	*data,
					     guint		 info,
					     guint		 time,
					     gpointer		 user_data);
static void compose_insert_drag_received_cb (GtkWidget		*widget,
					     GdkDragContext	*drag_context,
					     gint		 x,
					     gint		 y,
					     GtkSelectionData	*data,
					     guint		 info,
					     guint		 time,
					     gpointer		 user_data);

static void to_activated		(GtkWidget	*widget,
					 Compose	*compose);

static void ng_activated		(GtkWidget	*widget,
					 Compose	*compose);

static void from_activated		(GtkWidget	*widget,
					 Compose	*compose);
static void subject_activated		(GtkWidget	*widget,
					 Compose	*compose);
static void cc_activated		(GtkWidget	*widget,
					 Compose	*compose);
static void bcc_activated		(GtkWidget	*widget,
					 Compose	*compose);
static void replyto_activated		(GtkWidget	*widget,
					 Compose	*compose);

static GtkItemFactoryEntry compose_popup_entries[] =
{
	{N_("/_Add..."),	NULL, compose_attach_cb, 0, NULL},
	{N_("/_Remove"),	NULL, compose_attach_remove_selected, 0, NULL},
	{N_("/---"),		NULL, NULL, 0, "<Separator>"},
	{N_("/_Property..."),	NULL, compose_attach_property, 0, NULL}
};

static GtkItemFactoryEntry compose_entries[] =
{
	{N_("/_File"),				NULL, NULL, 0, "<Branch>"},
	{N_("/_File/_Attach file"),		"<control>M", compose_attach_cb, 0, NULL},
	{N_("/_File/_Insert file"),		"<control>I", compose_insert_file_cb, 0, NULL},
	{N_("/_File/Insert si_gnature"),	"<control>G", compose_insert_sig, 0, NULL},
	{N_("/_File/---"),			NULL, NULL, 0, "<Separator>"},
	{N_("/_File/_Close"),			"<alt>W", compose_close_cb, 0, NULL},

	{N_("/_Edit"),		   NULL,	 NULL,  0, "<Branch>"},
	{N_("/_Edit/_Undo"),	   "<control>Z", NULL,	0, NULL},
	{N_("/_Edit/_Redo"),	   "<control>Y", NULL,	0, NULL},
	{N_("/_Edit/---"),	   NULL,	 NULL,	0, "<Separator>"},
	{N_("/_Edit/Cu_t"),	   "<control>X", compose_cut_cb,    0, NULL},
	{N_("/_Edit/_Copy"),	   "<control>C", compose_copy_cb,   0, NULL},
	{N_("/_Edit/_Paste"),	   "<control>V", compose_paste_cb,  0, NULL},
	{N_("/_Edit/Select _all"), "<control>A", compose_allsel_cb, 0, NULL},
	{N_("/_Edit/---"),	   NULL,	 NULL,	0, "<Separator>"},
	{N_("/_Edit/Wrap long _lines"), "<alt>L", compose_wrap_line, 0, NULL},
	{N_("/_Edit/Edit with e_xternal editor"), "<alt>X",compose_ext_editor_cb,	0, NULL},

	{N_("/_Message"),		NULL,		NULL,	0, "<Branch>"},
	{N_("/_Message/_Send"),		"<alt>S",	compose_send_cb, 0, NULL},
	{N_("/_Message/Send _later"),	"<shift><alt>S",
					compose_send_later_cb,  0, NULL},
	{N_("/_Message/Save to _draft folder"),
					"<alt>D",	compose_draft_cb, 0, NULL},

	{N_("/_Message/---"),		NULL,		NULL,	0, "<Separator>"},
	{N_("/_Message/_Bcc"),		NULL, compose_toggle_bcc_cb, 0, "<ToggleItem>"},
	{N_("/_Message/_Reply to"),	NULL, compose_toggle_replyto_cb, 0, "<ToggleItem>"},
	{N_("/_Message/_Attach"),	NULL, compose_toggle_attach_cb, 0, "<ToggleItem>"},

	{N_("/_Tool"),			NULL, NULL,	0, "<Branch>"},
	{N_("/_Tool/_Address book"),	"<alt>A",	compose_address_cb, 0, NULL},
	{N_("/_Help"),			NULL, NULL,	0, "<LastBranch>"},
	{N_("/_Help/_About"),		NULL,		about_show,	0, NULL}
};

static GtkTargetEntry compose_mime_types[] =
{
	{"text/uri-list", 0, 0}
};

static gchar* fido_quot(gchar* from, gchar* quotemark) {
    gchar* from_name;
    gchar* result;
    int index, i;

    if(from == NULL)
        return g_strdup(quotemark);

    if((from_name = procheader_get_fromname(from)) == NULL)
        return g_strdup(quotemark);

    result = g_new(gchar, strlen(from_name) + strlen(quotemark) + 1 + 1);

    for(index = 0; isspace(quotemark[index]); index++)
        result[index] = ' ';

    result[index] = toupper(from_name[0]);
    index++;

    for(i = 1; from_name[i]; i++) {
        if(isspace(from_name[i-1])) {
            result[index] = toupper(from_name[i]);
            index++;
        }
    }
    for(i = 0; isspace(quotemark[i]); i++)
        ;
    strcpy(&result[index], &quotemark[i]);
    g_free(from_name);
    return result;
}

void compose_new(void)
{
	Compose *compose;
	extern cmdl_mail_info minfo;

	if (cur_account == NULL) {
	alertpanel_message(_("Warrning"), _("Account not exist"));
	    return;
	};

	compose = compose_create();
	compose->mode = COMPOSE_NEW;
	if (prefs_common.auto_sig)
		compose_insert_sig(compose);
	gtk_editable_set_position(GTK_EDITABLE(compose->text), 0);
	gtk_text_set_point(GTK_TEXT(compose->text), 0);

	if (minfo.ext_mailto && minfo.mailto!=NULL) {
	    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry), minfo.mailto);
	    minfo.ext_mailto = FALSE;
	}

	if (minfo.ext_subject && minfo.subject!=NULL) {
	    gtk_entry_set_text(GTK_ENTRY(compose->subject_entry), minfo.subject);
	    minfo.ext_subject = FALSE;
	}

	if (minfo.ext_body && minfo.fname!=NULL) {
	    compose_insert_file(compose, minfo.fname, TRUE);
	    minfo.ext_body = FALSE;
	}
	gtk_widget_grab_focus(compose->text);

	if (prefs_common.auto_ispell) {
	    gtkspell_attach(GTK_TEXT(compose->text));
	    gtkspell_check_all(GTK_TEXT(compose->text));
	};


}

void compose_reply(MsgInfo *msginfo)
{
	Compose *compose;
	if (cur_account == NULL) {
	alertpanel_message(_("Warrning"), _("Account not exist"));
	    return;
	};

	if (msginfo == NULL) {
	    alertpanel_message(_("Warrning"), _("Current message not exist"));
	    return;
	};

	compose = compose_create();
	compose->mode = COMPOSE_REPLY;
	if (compose_parse_header(compose, msginfo) < 0) return;

	compose_reply_set_entry(compose, msginfo);

	if (prefs_common.auto_sig)
		compose_insert_sig(compose);
	gtk_editable_set_position(GTK_EDITABLE(compose->text), 0);
	gtk_text_set_point(GTK_TEXT(compose->text), 0);
	gtk_widget_grab_focus(compose->text);
	if (prefs_common.auto_ispell) {
	    gtkspell_attach(GTK_TEXT(compose->text));
	    gtkspell_check_all(GTK_TEXT(compose->text));
	};

}


void compose_quote(MsgInfo *msginfo)
{
	Compose *compose;
	GtkText *text;
	FILE *fp;
	gchar *msgfile;
	gchar buf[BUFFSIZE];
	gchar *qmark;
 	gchar *quotemark = NULL;
	gint number_of_spaces_before;
	gchar *text_pointer;
	gchar *quotemark_pos;
	gchar *buffer_pos;
        gint i, str_len;
        gint current_line_len;
        gchar *citation_prefix = NULL;


	if (cur_account == NULL) {
	    alertpanel_message(_("Warrning"), _("Account not exist"));
	    return;
	};

	if (msginfo == NULL) {
	    alertpanel_message(_("Warrning"), _("Current message not exist"));
	    return;
	};

	compose = compose_create();
	compose->mode = COMPOSE_QUOTE;

	if (compose_parse_header(compose, msginfo) < 0) return;
	compose_reply_set_entry(compose, msginfo);

	text = GTK_TEXT(compose->text);
	gtk_text_freeze(text);

	if ((fp = procmime_get_text_part(msginfo)) == NULL)
		g_warning(_("Can't get text part\n"));
	else {
		compose_quote_parse_fmt(compose, msginfo);

		/* quote body */
		if (prefs_common.quotemark && *prefs_common.quotemark)
			qmark = prefs_common.quotemark;
		else
			qmark = "> ";

                if (prefs_common.quote_fido) {
                    quotemark = fido_quot(msginfo->from, qmark);
                } else {
		    quotemark = g_strdup(qmark);
                }
		for(number_of_spaces_before = 0;
                    isspace(qmark[number_of_spaces_before]);
                    number_of_spaces_before++)
                    ;

                while (fgets(buf, sizeof(buf), fp) != NULL) {
                    str_len = strlen(buf);
                    if (str_len > 2 && buf[str_len-1] == '\n' && buf[str_len-2] == '\r') {
                        buf[str_len-2] = '\n';
                        buf[str_len-1] = '\0';
                    }

                    quotemark_pos = get_quotemark_position(buf);
                    for(text_pointer = buf; isspace(*text_pointer); text_pointer++)
                        ;
                    if(*text_pointer == '\n' || *text_pointer == '\0') {
                        gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);
                        continue;
                    }
                    if(quotemark_pos) {/* re-cite */
                        citation_prefix = g_new(gchar, number_of_spaces_before \
                                                + quotemark_pos - text_pointer + 1 \
                                                + 1 + 1);
                        for(i = 0; i < number_of_spaces_before; i++)
                            citation_prefix[i] = ' ';
                        strncpy(&citation_prefix[i], text_pointer,
                                quotemark_pos - text_pointer + 1);
                        i += quotemark_pos - text_pointer + 1;
                        citation_prefix[i++] = '>';
                        citation_prefix[i] = '\0';
                        text_pointer = quotemark_pos + 1;
                    } else {
                        citation_prefix = g_strdup(quotemark);
                    }

                    gtk_text_insert(text, NULL, NULL, NULL, citation_prefix, -1);

                    if(prefs_common.linewrap_quote) {
                        current_line_len = strlen(citation_prefix);

                        /* insert leading spaces */
                        /*buffer_pos = quotemark_pos;
                        while(isspace(*buffer_pos)) {
                            gtk_text_insert(text, NULL, NULL, NULL, buffer_pos, 1);
                            buffer_pos++;
                            current_line_len++;
                         }*/

                        buffer_pos = text_pointer;
                        while(*buffer_pos != '\n' && *buffer_pos != '\0') {
                            if(!isspace(*buffer_pos)) {
                                gtk_text_insert(text, NULL, NULL, NULL, buffer_pos, 1);
                            } else { /* split line only on spaces */
                                if(current_line_len >= prefs_common.linewrap_len - 1) {
                                    gtk_text_insert(text, NULL, NULL, NULL, "\n", -1);
                                    gtk_text_insert(text, NULL, NULL, NULL, citation_prefix, -1);
                                    current_line_len = strlen(citation_prefix);
                                    /* strip spaces */
                                    while(isspace(*buffer_pos))
                                        buffer_pos++;
                                    /*if(*buffer_pos == '\n' || *buffer_pos == '\0')*/
                                        buffer_pos--;
                                } else {
                                  gtk_text_insert(text, NULL, NULL, NULL, buffer_pos, 1);
                                }
                            }
                            buffer_pos++;
                            current_line_len++;
                        }
                        gtk_text_insert(text, NULL, NULL, NULL, "\n", -1);
                    } else {
                        gtk_text_insert(text, NULL, NULL, NULL, text_pointer, -1);
                    }
                   g_free(citation_prefix);
		}
		fclose(fp);
	}
	g_free(quotemark);

	if (prefs_common.auto_sig)
		compose_insert_sig(compose);
	gtk_editable_set_position(GTK_EDITABLE(text), 0);
	gtk_text_set_point(GTK_TEXT(text), 0);

	gtk_text_thaw(text);
	gtk_widget_grab_focus(compose->text);

	if (prefs_common.auto_ispell) {
	    gtkspell_attach(GTK_TEXT(compose->text));
	    gtkspell_check_all(GTK_TEXT(compose->text));
	};
}

#define INSERT_FW_HEADER(var, hdr) \
if (msginfo->var && *msginfo->var) { \
	gtk_text_insert(text, NULL, NULL, NULL, hdr, -1); \
	gtk_text_insert(text, NULL, NULL, NULL, msginfo->var, -1); \
	gtk_text_insert(text, NULL, NULL, NULL, "\n", 1); \
}

void compose_forward(MsgInfo *msginfo)
{
	Compose *compose;
	GtkText *text;
	FILE *fp;
	gchar buf[BUFFSIZE];
	gchar *msgfile;
	int len;

	if (cur_account == NULL) {
	    alertpanel_message(_("Warrning"), _("Account not exist"));
	    return;
	};

	if (msginfo == NULL) {
	    alertpanel_message(_("Warrning"), _("Current message not exist"));
	    return;
	};


	compose = compose_create();
	compose->mode = COMPOSE_REEDIT_DRAFT;
	if (msginfo->subject && *msginfo->subject) {
		gtk_entry_set_text(GTK_ENTRY(compose->subject_entry), "Fw: ");
		gtk_entry_append_text(GTK_ENTRY(compose->subject_entry),
				      msginfo->subject);
	}

	text = GTK_TEXT(compose->text);
	gtk_text_freeze(text);

	msgfile = (gchar*) procmsg_get_message_file_path(msginfo);

	if ((fp = procmime_get_text_part(msginfo)) == NULL)
			g_warning(_("Can't get text part\n"));
	else {
		/* insert header */
		gtk_text_insert(text, NULL, NULL, NULL,
				_("\n\nBegin forwarded message:\n\n"), -1);
		INSERT_FW_HEADER(date,    "Date: ");
		INSERT_FW_HEADER(from,    "From: ");
		INSERT_FW_HEADER(to,      "To: ");
		INSERT_FW_HEADER(subject, "Subject: ");
		gtk_text_insert(text, NULL, NULL, NULL, "\n\n", 2);

		/* forward body */
		while (fgets(buf, sizeof(buf), fp) != NULL) {
			len = strlen(buf);
			if (buf[len-1] == '\n' && buf[len-2] == '\r') {
			    buf[len-2] = '\n';
			    buf[len-1] = '\0';
			    };
			gtk_text_insert(text, NULL, NULL, NULL, buf, -1);
		}
		fclose(fp);
	}

	g_free(msgfile);

	if (prefs_common.auto_sig)
		compose_insert_sig(compose);
	gtk_editable_set_position(GTK_EDITABLE(compose->text), 0);
	gtk_text_set_point(GTK_TEXT(compose->text), 0);

	gtk_text_thaw(text);
	gtk_widget_grab_focus(compose->to_entry);

	if (prefs_common.auto_ispell) {
	    gtkspell_attach(GTK_TEXT(compose->text));
	    gtkspell_check_all(GTK_TEXT(compose->text));
	};

}

void compose_forward_as_attach(SummaryView *summaryview)
{
        MsgInfo *msginfo;
	Compose *compose;
	GtkText *text;
	FILE *fp;
	gchar buf[BUFFSIZE];
	gchar *msgfile;
	guint num;
        GList *cur;

	if (cur_account == NULL) {
	    alertpanel_message(_("Warrning"), _("Account not exist"));
	    return;
	};

	if (summaryview == NULL) {
	    alertpanel_message(_("Warrning"), _("Summaryview not exist? =/"));
	    return;
	};

        for(cur=GTK_CLIST(summaryview->ctree)->selection,num=0;cur!=NULL;cur=cur->next,num++);

        if(num==0) {
	    alertpanel_message(_("Warrning"), _("No messages selected!"));
	    return;
        }

	compose = compose_create();
	compose->mode = COMPOSE_REEDIT_DRAFT;
        msginfo=gtk_ctree_node_get_row_data(GTK_CTREE(summaryview->ctree),summaryview->selected);
        if(num==1) {
	        if (msginfo->subject && *msginfo->subject) {
		        gtk_entry_set_text(GTK_ENTRY(compose->subject_entry), "Fw: ");
		        gtk_entry_append_text(GTK_ENTRY(compose->subject_entry),
				                        msginfo->subject);
	        }
        } else {
	        gtk_entry_set_text(GTK_ENTRY(compose->subject_entry), "Multi Forward");
        }

	text = GTK_TEXT(compose->text);
	gtk_text_freeze(text);

        for(cur=GTK_CLIST(summaryview->ctree)->selection;cur!=NULL;cur=cur->next) {
                msginfo=gtk_ctree_node_get_row_data(GTK_CTREE(summaryview->ctree),GTK_CTREE_NODE(cur->data));
	        msgfile = (gchar*) procmsg_get_message_file_path(msginfo);
	        if (!is_file_exist(msgfile)) {
		        alertpanel_message(_("Warrning"), _("Current message not exist"));
	        } else {
		        compose_attach_append(compose, msgfile, MIME_MESSAGE_RFC822);
	        };
	        g_free(msgfile);
        }

	if (prefs_common.auto_sig)
		compose_insert_sig(compose);
	gtk_editable_set_position(GTK_EDITABLE(compose->text), 0);
	gtk_text_set_point(GTK_TEXT(compose->text), 0);

	gtk_text_thaw(text);
	gtk_widget_grab_focus(compose->to_entry);

	if (prefs_common.auto_ispell) {
	    gtkspell_attach(GTK_TEXT(compose->text));
	    gtkspell_check_all(GTK_TEXT(compose->text));
	};
}

void compose_reedit(MsgInfo *msginfo)
{
	Compose *compose;
	GtkText *text;
	FILE *fp;
	gchar buf[BUFFSIZE];

	if (msginfo == NULL) {
	    alertpanel_message(_("Warrning"), _("Current message not exist"));
	    return;
	};



	compose = compose_create();
	compose->mode = COMPOSE_REEDIT_DRAFT;
	compose->targetinfo = procmsg_msginfo_copy(msginfo);

	if (compose_parse_header(compose, msginfo) < 0) return;
	compose_reedit_set_entry(compose, msginfo);

	text = GTK_TEXT(compose->text);
	gtk_text_freeze(text);

	if ((fp = procmime_get_text_part(msginfo)) == NULL)
		g_warning(_("Can't get text part\n"));
	else {
		while (fgets(buf, sizeof(buf), fp) != NULL)
			gtk_text_insert(text, NULL, NULL, NULL, buf, -1);
		fclose(fp);
	}

	gtk_text_thaw(text);
	gtk_widget_grab_focus(compose->text);

	if (prefs_common.auto_ispell) {
	    gtkspell_attach(GTK_TEXT(compose->text));
	    gtkspell_check_all(GTK_TEXT(compose->text));
	};

}

GList *compose_get_compose_list(void)
{
	return compose_list;
}

void compose_entry_append(Compose *compose, const gchar *address,
			  ComposeEntryType type)
{
	GtkEntry *entry;
	const gchar *text;

	if (!address || *address == '\0') return;

	switch (type) {
	case COMPOSE_TO:
		entry = GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry);
		break;
	case COMPOSE_CC:
		entry = GTK_ENTRY(GTK_COMBO(compose->cc_entry)->entry);
		break;
	case COMPOSE_BCC:
		entry = GTK_ENTRY(GTK_COMBO(compose->bcc_entry)->entry);
		break;
	case COMPOSE_FROM:
		entry = GTK_ENTRY(GTK_COMBO(compose->from_entry)->entry);
		break;
	default:
		entry = GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry);
		break;
	}

	text = gtk_entry_get_text(entry);
	if (*text != '\0')
		gtk_entry_append_text(entry, ", ");
	gtk_entry_append_text(entry, address);
}

static gchar *compose_convert_and_free(gchar *str)
{
	gchar tmp[BUFFSIZE*4];
	compose_convert_header(tmp, sizeof(tmp), str, strlen(str));
	g_free(str);
	return g_strdup(tmp);
}

static gint compose_parse_header(Compose *compose, MsgInfo *msginfo)
{
	static HeaderEntry hentry[] = {{"Reply-To:",	NULL, TRUE},
				       {"Cc:",		NULL, FALSE},
				       {"References:",	NULL, FALSE},
				       {"Bcc:",		NULL, FALSE},
				       {NULL,		NULL, FALSE}};

	enum
	{
		H_REPLY_TO	= 0,
		H_CC		= 1,
		H_REFERENCES	= 2,
		H_BCC		= 3
	};

	FILE *fp;

	/* open parsing message */
	g_return_val_if_fail(msginfo != NULL, -1);
	if ((fp = procmsg_open_message(msginfo)) == NULL) return -1;
	procheader_get_header_fields(fp, hentry);
	fclose(fp);

	if (hentry[H_REPLY_TO].body != NULL) {
		compose->replyto = compose_convert_and_free(hentry[H_REPLY_TO].body);
		hentry[H_REPLY_TO].body = NULL;
	}
	if (hentry[H_CC].body != NULL) {
		compose->cc = compose_convert_and_free(hentry[H_CC].body);
		hentry[H_CC].body = NULL;
	}
	if (hentry[H_REFERENCES].body != NULL) {
		gchar *p;

		if ((p = strrchr(hentry[H_REFERENCES].body, '<')) == NULL ||
		    strchr(p + 1, '>') == NULL)
			g_free(hentry[H_REFERENCES].body);
		else if (compose->mode != COMPOSE_REEDIT_DRAFT &&
			 msginfo->msgid && *msginfo->msgid) {
			compose->references =
				g_strconcat(hentry[H_REFERENCES].body,
					    "\n\t", "<", msginfo->msgid, ">",
					    NULL);
			g_free(hentry[H_REFERENCES].body);
		} else
			compose->references = hentry[H_REFERENCES].body;

		hentry[H_REFERENCES].body = NULL;
	}
	if (hentry[H_BCC].body != NULL) {
		if (compose->mode == COMPOSE_REEDIT_DRAFT)
			compose->bcc = compose_convert_and_free(hentry[H_BCC].body);
		else
			g_free(hentry[H_BCC].body);
		hentry[H_BCC].body = NULL;
	}

	if (compose->mode == COMPOSE_REEDIT_DRAFT && msginfo->inreplyto)
		compose->inreplyto = g_strdup(msginfo->inreplyto);
	else if (compose->mode != COMPOSE_REEDIT_DRAFT &&
	    msginfo->msgid && *msginfo->msgid) {
		compose->inreplyto = g_strdup(msginfo->msgid);

		if (!compose->references) {
			if (msginfo->inreplyto && *msginfo->inreplyto)
				compose->references =
					g_strdup_printf("<%s>\n\t<%s>",
							msginfo->inreplyto,
							msginfo->msgid);
			else
				compose->references =
					g_strconcat("<", msginfo->msgid, ">",
						    NULL);
		}
	}

	return 0;
}

static void compose_quote_parse_fmt(Compose *compose, MsgInfo *msginfo)
{
	gchar *str;
	gchar *mbs;
	wchar_t *wcsfmt;
	wchar_t *sp;
	GtkText *text = GTK_TEXT(compose->text);

	if (!prefs_common.quotefmt || *prefs_common.quotefmt == '\0') return;

	Xalloca(mbs, sizeof(wchar_t) + 1, return);
	Xalloca(wcsfmt, (strlen(prefs_common.quotefmt) + 1) * sizeof(wchar_t),
		return);
	mbstowcs(wcsfmt, prefs_common.quotefmt, strlen(prefs_common.quotefmt) + 1);
	sp = wcsfmt;

	while (*sp) {
		gint len;

		len = safe_wctomb(mbs, *sp);
		mbs[len] = '\0';

		if (*mbs == '%') {
			safe_wctomb(mbs, *(++sp));
			str = NULL;

			switch (*mbs) {
			case 'd':
				str = msginfo->date;
				sp++;
				break;
			case 'f':
				str = msginfo->from;
				sp++;
				break;
			case 's':
				str = msginfo->subject;
				sp++;
				break;
			case 't':
				str = msginfo->to;
				sp++;
				break;
			case 'i':
				str = g_strconcat("<", msginfo->msgid, ">",
						  NULL);
				gtk_text_insert(text, NULL, NULL, NULL,
						str, -1);
				g_free(str);
				str = NULL;
				sp++;
				break;
			case '%':
				str = "%";
				sp++;
				break;
			default:
				break;
			}
			if (str)
				gtk_text_insert
					(text, NULL, NULL, NULL, str, -1);
		} else if (*mbs == '\\') {
			safe_wctomb(mbs, *(++sp));
			str = NULL;

			switch (*mbs) {
			case 'n':
				str = "\n";
				break;
			case 't':
				str = "\t";
				break;
			case '\\':
				str = "\\";
				break;
			default:
				break;
			}
			if (str) {
				gtk_text_insert
					(text, NULL, NULL, NULL, str, -1);
				sp++;
			}
		} else {
			gtk_text_insert(text, NULL, NULL, NULL, mbs, -1);
			sp++;
		}
	}

}


#ifdef NO_CC_ACCOUNTS
GHashTable *repl_comp_to_table;

static gint insert_account_tos(PrefsAccount *acco, gpointer data )
{
	if (acco)
		g_hash_table_insert(repl_comp_to_table, acco->address,
				    GINT_TO_POINTER(1));
	return(0);
}

#endif

#define SET_ENTRY(entry, str) \
{ \
	if (str && *str) \
		gtk_entry_set_text(GTK_ENTRY(compose->entry), str); \
}

static void compose_reply_set_entry(Compose *compose, MsgInfo *msginfo)
{
	GSList *cc_list;
	GSList *cur;
	gchar *from;
#ifndef NO_CC_ACCOUNTS
	GHashTable *repl_comp_to_table;
#endif

	if (msginfo == NULL) {
	    alertpanel_message(_("Warrning"), _("Current message not exist"));
	    return;
	};
	if ( MSG_IS_NEWS( msginfo->flags ) ) {
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(compose->ng_entry)->entry),
				     compose->newsgroups ? compose->newsgroups : "None");
	} else {
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry),
			   compose->replyto ? compose->replyto
				: msginfo->from ? msginfo->from : "");
	};

	if (msginfo->subject && *msginfo->subject) {
		gchar *buf, *buf2, *p;

		buf = g_strdup(msginfo->subject);
		while (!strncasecmp(buf, "Re:", 3)) {
			p = buf + 3;
			while (isspace(*p)) p++;
			memmove(buf, p, strlen(p) + 1);
		}

		buf2 = g_strdup_printf("Re: %s", buf);
		gtk_entry_set_text(GTK_ENTRY(compose->subject_entry), buf2);
		g_free(buf2);
		g_free(buf);
	} else
		gtk_entry_set_text(GTK_ENTRY(compose->subject_entry), "Re: ");

	from = g_strdup(compose->replyto ? compose->replyto :
			msginfo->from ? msginfo->from : "");
	extract_address(from);

    if ( !MSG_IS_NEWS( msginfo->flags ) ) {
	cc_list = address_list_append(NULL, msginfo->to);
	cc_list = address_list_append(cc_list, compose->cc);

	repl_comp_to_table = g_hash_table_new(g_str_hash, g_str_equal);
	g_hash_table_insert(repl_comp_to_table, from, GINT_TO_POINTER(1));

#ifdef NO_CC_ACCOUNTS
	account_foreach( insert_account_tos, NULL );

#else
	if (cur_account)
		g_hash_table_insert(repl_comp_to_table, cur_account->address,
				    GINT_TO_POINTER(1));
#endif

	/* remove address on To: and that of current account */
	for (cur = cc_list; cur != NULL; ) {
		GSList *next = cur->next;

		if (g_hash_table_lookup(repl_comp_to_table, cur->data) != NULL)
			cc_list = g_slist_remove(cc_list, cur->data);
		else
			g_hash_table_insert(repl_comp_to_table, cur->data, cur);

		cur = next;
	}
	g_hash_table_destroy(repl_comp_to_table);
	g_free(from);

	if (cc_list) {
		for (cur = cc_list; cur != NULL; cur = cur->next)
			compose_entry_append(compose, (gchar *)cur->data,
					     COMPOSE_CC);
		slist_free_strings(cc_list);
		g_slist_free(cc_list);
	}
    };
}

static void compose_reedit_set_entry(Compose *compose, MsgInfo *msginfo)
{
	if (msginfo == NULL) {
	    alertpanel_message(_("Warrning"), _("Current message not exist"));
	    return;
	};

	if (msginfo->to && *(msginfo->to))
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry), msginfo->to);

	if (msginfo->newsgroups && *(msginfo->newsgroups))
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(compose->ng_entry)->entry), msginfo->newsgroups);

	SET_ENTRY(subject_entry, msginfo->subject);

	if (compose->cc && *(compose->cc))
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(compose->cc_entry)->entry), compose->cc);

	SET_ENTRY(bcc_entry, compose->bcc);
	SET_ENTRY(reply_entry, compose->replyto);

	if (compose->bcc) {
		GtkItemFactory *ifactory;
		GtkWidget *menuitem;

		ifactory = gtk_item_factory_from_widget(compose->menubar);
		menuitem = gtk_item_factory_get_item(ifactory, "/Message/Bcc");
		gtk_check_menu_item_set_active
			(GTK_CHECK_MENU_ITEM(menuitem), TRUE);
	}
	if (compose->replyto) {
		GtkItemFactory *ifactory;
		GtkWidget *menuitem;

		ifactory = gtk_item_factory_from_widget(compose->menubar);
		menuitem = gtk_item_factory_get_item
			(ifactory, "/Message/Reply to");
		gtk_check_menu_item_set_active
			(GTK_CHECK_MENU_ITEM(menuitem), TRUE);
	}
}


static void compose_insert_sig(Compose *compose)
{
	gtk_text_freeze(GTK_TEXT(compose->text));

	if (prefs_common.sig_sep) {
		gtk_text_insert(GTK_TEXT(compose->text), NULL, NULL, NULL,
				"\n", -1);
		gtk_text_insert(GTK_TEXT(compose->text), NULL, NULL, NULL,
				prefs_common.sig_sep, -1);
		gtk_text_insert(GTK_TEXT(compose->text), NULL, NULL, NULL,
				"\n", -1);
	}

	if (cur_account && cur_account->sig_path && is_file_exist(cur_account->sig_path)) compose_insert_file(compose, cur_account->sig_path,TRUE);
	else {
		gchar *sigfile;

		sigfile = g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S,
				      DEFAULT_SIGNATURE, NULL);
		if (is_file_exist(sigfile))
			compose_insert_file(compose, sigfile, TRUE);
		g_free(sigfile);
	}
	gtk_text_thaw(GTK_TEXT(compose->text));
}

static void compose_insert_file(Compose *compose, const gchar *file,const gboolean isexec)
{
	GtkText *text = GTK_TEXT(compose->text);
	gchar buf[BUFFSIZE];
	gboolean weareexecute;
	FILE *fp;

	if (file == NULL || (!is_file_exist(file))) {
	    alertpanel_message(_("Warrning"), _("File not exist"));
	    return;
	};
	if (is_file_executable(file) && isexec) {
	    fp=popen(file, "r");
	    weareexecute=TRUE;
	} else {
	    fp = fopen(file, "r");
	    weareexecute=FALSE;
	}

	if ( fp == NULL) { FILE_OP_ERROR(file, "fopen");return; };
	gtk_text_freeze(text);

	while (fgets(buf, sizeof(buf), fp) != NULL)
		gtk_text_insert(text, NULL, NULL, NULL, buf, -1);

	gtk_text_thaw(text);
	if (weareexecute) pclose(fp); else fclose(fp);
}

static void compose_attach_append(Compose *compose, const gchar *file,ContentType cnttype)
{
	AttachInfo *ainfo;
	gchar *text[N_ATTACH_COLS];
	off_t size;
	gint row;

	if (!is_file_exist(file)) {
		g_warning(_("File %s doesn't exist\n"), file);
		return;
	}
	if ((size = get_file_size(file)) < 0) {
		g_warning(_("Can't get file size of %s\n"), file);
		return;
	}
	if (size == 0) {
		alertpanel_notice(_("File %s is empty\n"), file);
		return;
	}

	if (!compose->use_attach) {
		GtkItemFactory *ifactory;
		GtkWidget *menuitem;

		ifactory = gtk_item_factory_from_widget(compose->menubar);
		menuitem = gtk_item_factory_get_item(ifactory,
						     "/Message/Attach");
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(menuitem),
					       TRUE);
	}

	ainfo = g_new0(AttachInfo, 1);

	ainfo->file = g_strdup(file);
	if (cnttype == MIME_MESSAGE_RFC822) {
		ainfo->content_type = g_strdup("message/rfc822");
		ainfo->encoding = ENC_7BIT;
		ainfo->name = g_strdup_printf(_("Message: %s"),
					      g_basename(file));
	} else {
		ainfo->content_type = procmime_get_mime_type(file);
		if (!ainfo->content_type)
			ainfo->content_type =
				g_strdup("text/plain");
		ainfo->encoding = ENC_BASE64;
		ainfo->name = g_strdup(g_basename(file));
	}


	ainfo->size = size;

	text[COL_MIMETYPE] = ainfo->content_type;
	text[COL_SIZE] = to_human_readable(size);
	text[COL_NAME] = ainfo->name;

	row = gtk_clist_append(GTK_CLIST(compose->attach_clist), text);
	gtk_clist_set_row_data(GTK_CLIST(compose->attach_clist), row, ainfo);

	if (cnttype != MIME_MESSAGE_RFC822)
		compose_changed_cb(NULL, compose);

}

static gint compose_send(Compose *compose)
{
	gchar tmps[22];
	gchar tmpn[22];
	gchar *sbuf;
	gchar *nbuf;
	gboolean ismail,isnews;
	gint ok;
	static gboolean lock = FALSE;

	if (lock) return 1;
	lock = TRUE;
	ismail=isnews=FALSE;

	g_return_val_if_fail(cur_account != NULL, -1);

	sbuf = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry));
	nbuf = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->ng_entry)->entry));
	if (*sbuf != '\0') ismail=TRUE;
	if (*nbuf != '\0') isnews=TRUE;
	if (!isnews && !ismail) {
		alertpanel_error(_("Recipient is not specified."));
		lock = FALSE;
		return 1;
	}

	/* write to temporary file */
	if (ismail) g_snprintf(tmps, sizeof(tmps), "%s%ctmpmsgs%d",
			    g_get_tmp_dir(), G_DIR_SEPARATOR, (gint)compose);
	if (isnews) g_snprintf(tmpn, sizeof(tmpn), "%s%ctmpmsgn%d",
			    g_get_tmp_dir(), G_DIR_SEPARATOR, (gint)compose);

	if (prefs_common.linewrap_at_send)
		compose_wrap_line(compose);

	if (ismail) if (compose_write_to_file_for_smtp(compose, tmps, FALSE) <0) return -1;
	if (isnews) if (compose_write_to_file_for_nntp(compose, tmpn, FALSE) <0) return -1;
	ok = send_message(compose->to_list,compose->ng_list,compose->nserver, tmps, tmpn, ismail, isnews);

	statusbar_pop_all();
	/* queue message if failed to send */
	if (ok < 0 && prefs_common.queue_msg) {
		AlertValue val;

		val = alertpanel(_("Queueing"),
				 _("Put this message into queue folder?"),
				 _("OK"), _("Cancel"), NULL);
		if (G_ALERTDEFAULT == val) {
			ok = compose_queue(compose, tmps);
			if (ok < 0)
				alertpanel_error
					(_("Can't queue the message."));
		}
	}

	/* save message to outbox */
	if (ismail && ok == 0 && prefs_common.savemsg) {
		if (compose_save_to_outbox(compose, tmps) < 0)
			alertpanel_error
				(_("Can't save the message to outbox."));
	}

	if (ismail) if (unlink(tmps) < 0)
		    FILE_OP_ERROR(tmps, "unlink");

	if (isnews) if (unlink(tmpn) < 0)
		    FILE_OP_ERROR(tmpn, "unlink");

	lock = FALSE;
	return ok;
}

static gint compose_write_to_file_for_smtp(Compose *compose, const gchar *file,
				  gboolean is_draft)
{
	FILE *fp;
	size_t len;
	gchar *chars,*hdrchars;
	gchar *buf;
	const gchar *out_codeset;
	const gchar *out_iconv_codeset;
	EncodingType encoding;

	if ((fp = fopen(file, "w")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	/* chmod for security */
	if (chmod(file, S_IRUSR|S_IWUSR) < 0) {
		FILE_OP_ERROR(file, "chmod");
		g_warning(_("can't change file mode\n"));
	}

	/* get all composed text */
	chars = gtk_editable_get_chars(GTK_EDITABLE(compose->text), 0, -1);
	len = strlen(chars);
	hdrchars = gtk_editable_get_chars(GTK_EDITABLE(compose->subject_entry), 0, -1);
	if (is_ascii_str(chars) && is_ascii_str(hdrchars)) {
		buf = g_strdup(chars);
		out_codeset = "US-ASCII";
		encoding = ENC_7BIT;
	} else {
            gint retval;

		out_iconv_codeset = get_iconv_charset(prefs_common.outgoing_charset);
		if (out_iconv_codeset==NULL) {
		    out_codeset=g_strdup(curr_charset);
		    out_iconv_codeset=g_strdup(curr_charset);
		} else {
		    out_codeset=prefs_common.outgoing_charset;
		}
		encoding = prefs_common.outgoing_encoding;

		buf=g_malloc((strlen(chars)+1)*MAX_CHAR_LEN);

                retval=code_convert(buf, ((strlen(chars)+1)*MAX_CHAR_LEN) , chars, strlen(chars)+1, (gchar*)curr_charset,(gchar*)out_iconv_codeset);
                if (retval==-1) {
                    g_free(chars);
			fclose(fp);
			unlink(file);
			alertpanel_error(_("Can't convert the codeset of the message."));
			return -1;
		};
		buf[retval]=0;
	}
	if (chars!=NULL) g_free(chars);
	/* write headers */
	if (compose_write_headers_smtp
		(compose, fp, out_codeset, encoding, is_draft) < 0)
	 {

		g_warning(_("can't write headers\n"));
		fclose(fp);
		unlink(file);
		g_free(buf);
		return -1;
	}
	if (compose->use_attach) {
		fputs("This is a multi-part message in MIME format.\n", fp);
		fprintf(fp, "\n--%s\n", compose->boundary);
		fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n",
			out_codeset);
		fprintf(fp, "Content-Transfer-Encoding: %s\n",
			procmime_get_encoding_str(encoding));
		fputc('\n', fp);
	}
	/* write body */
	len = strlen(buf);
	if (encoding == ENC_BASE64) {
		gchar outbuf[B64_BUFFSIZE];
		gint i, l;
		for (i = 0; i < len; i += B64_LINE_SIZE) {
			l = MIN(B64_LINE_SIZE, len - i);
			to64frombits(outbuf, buf + i, l);
			fputs(outbuf, fp);
			fputc('\n', fp);
		}
	} else if (fwrite(buf, sizeof(gchar), len, fp) != len) {
		FILE_OP_ERROR(file, "fwrite");
		fclose(fp);
		unlink(file);
		g_free(buf);
		return -1;
	}
	g_free(buf);
	if (compose->use_attach)
		compose_write_attach(compose, fp);

	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(file, "fclose");
		unlink(file);
		return -1;
	}

	return 0;
}

static gint compose_write_to_file_for_nntp(Compose *compose, const gchar *file,
				  gboolean is_draft)
{
	FILE *fp;
	size_t len;
	gchar *chars,*hdrchars;
	gchar *buf;
	const gchar *out_codeset;
	const gchar *out_iconv_codeset;
	EncodingType encoding;

	if ((fp = fopen(file, "w")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	/* chmod for security */
	if (chmod(file, S_IRUSR|S_IWUSR) < 0) {
		FILE_OP_ERROR(file, "chmod");
		g_warning(_("can't change file mode\n"));
	}

	/* get all composed text */
	chars = gtk_editable_get_chars(GTK_EDITABLE(compose->text), 0, -1);
	len = strlen(chars);
	hdrchars = gtk_editable_get_chars(GTK_EDITABLE(compose->subject_entry), 0, -1);
	if (is_ascii_str(chars) && is_ascii_str(hdrchars)) {
		buf = g_strdup(chars);
		out_codeset = "US-ASCII";
		encoding = ENC_7BIT;
	} else {

		gint retval;
		out_iconv_codeset = get_iconv_charset(prefs_common.outgoing_charset);
		if (out_iconv_codeset==NULL) {
		    out_codeset=g_strdup(curr_charset);
		    out_iconv_codeset=g_strdup(curr_charset);
		} else {
		    out_codeset=prefs_common.outgoing_charset;
		}
		encoding = prefs_common.outgoing_encoding;

		buf=g_malloc((strlen(chars)+1)*MAX_CHAR_LEN);
		retval=code_convert(buf, ((strlen(chars)+1)*MAX_CHAR_LEN) , chars, strlen(chars)+1, (gchar*) curr_charset,(gchar*) out_iconv_codeset);
		if (retval==-1) {
			g_free(chars);
			fclose(fp);
			unlink(file);
			alertpanel_error(_("Can't convert the codeset of the message."));
			return -1;
		};
		buf[retval]=0;
	}
	g_free(chars);

	/* write headers */
	if (compose_write_headers_nntp
		(compose, fp, out_codeset, encoding, is_draft) < 0)
	 {

		g_warning(_("can't write headers\n"));
		fclose(fp);
		unlink(file);
		g_free(buf);
		return -1;
	}

	if (compose->use_attach) {
		fputs("This is a multi-part message in MIME format.\n", fp);
		fprintf(fp, "\n--%s\n", compose->boundary);
		fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n",
			out_codeset);
		fprintf(fp, "Content-Transfer-Encoding: %s\n",
			procmime_get_encoding_str(encoding));
		fputc('\n', fp);
	}

	/* write body */
	len = strlen(buf);
	if (encoding == ENC_BASE64) {
		gchar outbuf[B64_BUFFSIZE];
		gint i, l;

		for (i = 0; i < len; i += B64_LINE_SIZE) {
			l = MIN(B64_LINE_SIZE, len - i);
			to64frombits(outbuf, buf + i, l);
			fputs(outbuf, fp);
			fputc('\n', fp);
		}
	} else if (fwrite(buf, sizeof(gchar), len, fp) != len) {
		FILE_OP_ERROR(file, "fwrite");
		fclose(fp);
		unlink(file);
		g_free(buf);
		return -1;
	}
	g_free(buf);

	if (compose->use_attach)
		compose_write_attach(compose, fp);

	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(file, "fclose");
		unlink(file);
		return -1;
	}

	return 0;
}

static gint compose_write_to_file_for_queue(Compose *compose, const gchar *file,
				  gboolean is_draft)
{
	FILE *fp;
	size_t len;
	gchar *chars,*hdrchars;
	gchar *buf;
	const gchar *out_codeset;
	const gchar *out_iconv_codeset;
	EncodingType encoding;

	if ((fp = fopen(file, "w")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	/* chmod for security */
	if (chmod(file, S_IRUSR|S_IWUSR) < 0) {
		FILE_OP_ERROR(file, "chmod");
		g_warning(_("can't change file mode\n"));
	}

	/* get all composed text */
	chars = gtk_editable_get_chars(GTK_EDITABLE(compose->text), 0, -1);
	len = strlen(chars);
	hdrchars = gtk_editable_get_chars(GTK_EDITABLE(compose->subject_entry), 0, -1);
	if (is_ascii_str(chars) && is_ascii_str(hdrchars)) {
		buf = g_strdup(chars);
		out_codeset = "US-ASCII";
		encoding = ENC_7BIT;
	} else {

		gint retval;
		out_iconv_codeset = get_iconv_charset(prefs_common.outgoing_charset);
		if (out_codeset==NULL) {
		    out_codeset=g_strdup(curr_charset);
		    out_iconv_codeset=g_strdup(curr_charset);
		} else {
		    out_codeset=prefs_common.outgoing_charset;
		}
		encoding = prefs_common.outgoing_encoding;

		buf=malloc((strlen(chars)+1)*MAX_CHAR_LEN);
		retval=code_convert(buf, ((strlen(chars)+1)*MAX_CHAR_LEN) , chars, strlen(chars)+1, (gchar*)curr_charset,(gchar*) out_iconv_codeset);

		if (retval==-1) {
			g_free(chars);
			fclose(fp);
			unlink(file);
			alertpanel_error(_("Can't convert the codeset of the message."));
			return -1;
		};
		buf[retval]=0;
	}
	g_free(chars);

	/* write headers */
	if (compose_write_headers_queue
		(compose, fp, out_codeset, encoding, is_draft) < 0)
	 {

		g_warning(_("can't write headers\n"));
		fclose(fp);
		unlink(file);
		g_free(buf);
		return -1;
	}

	if (compose->use_attach) {
		fputs("This is a multi-part message in MIME format.\n", fp);
		fprintf(fp, "\n--%s\n", compose->boundary);
		fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n",
			out_codeset);
		fprintf(fp, "Content-Transfer-Encoding: %s\n",
			procmime_get_encoding_str(encoding));
		fputc('\n', fp);
	}

	/* write body */
	len = strlen(buf);
	if (encoding == ENC_BASE64) {
		gchar outbuf[B64_BUFFSIZE];
		gint i, l;

		for (i = 0; i < len; i += B64_LINE_SIZE) {
			l = MIN(B64_LINE_SIZE, len - i);
			to64frombits(outbuf, buf + i, l);
			fputs(outbuf, fp);
			fputc('\n', fp);
		}
	} else if (fwrite(buf, sizeof(gchar), len, fp) != len) {
		FILE_OP_ERROR(file, "fwrite");
		fclose(fp);
		unlink(file);
		g_free(buf);
		return -1;
	}
	g_free(buf);

	if (compose->use_attach)
		compose_write_attach(compose, fp);

	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(file, "fclose");
		unlink(file);
		return -1;
	}

	return 0;
}


static gint compose_write_body_to_file(Compose *compose, const gchar *file)
{
	FILE *fp;
	size_t len;
	gchar *chars;

	if ((fp = fopen(file, "w")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		return -1;
	}

	/* chmod for security */
	if (chmod(file, S_IRUSR|S_IWUSR) < 0) {
		FILE_OP_ERROR(file, "chmod");
		g_warning(_("can't change file mode\n"));
	}

	chars = gtk_editable_get_chars(GTK_EDITABLE(compose->text), 0, -1);

	/* write body */
	len = strlen(chars);
	if (fwrite(chars, sizeof(gchar), len, fp) != len) {
		FILE_OP_ERROR(file, "fwrite");
		g_free(chars);
		fclose(fp);
		unlink(file);
		return -1;
	}

	g_free(chars);

	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(file, "fclose");
		unlink(file);
		return -1;
	}
	return 0;
}

static gint compose_save_to_outbox(Compose *compose, const gchar *file)
{
	FILE *fp;
	gint num;

	debug_print(_("saving sent message...\n"));

	if ((num = procmsg_copy_message(OUTBOX_DIR, file)) < 0) {
		g_warning(_("can't save message\n"));
		return -1;
	} else if ((fp = procmsg_open_mark_file(OUTBOX_DIR, TRUE)) == NULL)
		g_warning(_("can't open mark file\n"));
	else {
		MsgInfo newmsginfo;

		newmsginfo.msgnum = num;
		newmsginfo.flags = 0;
		procmsg_write_flags(&newmsginfo, fp);
		fclose(fp);
	}

	return 0;
}

static gint compose_queue(Compose *compose, const gchar *file)
{
	FILE *fp, *src_fp;
	gint num;
	gchar *destfile;
	GSList *cur;
	gchar buf[BUFFSIZE];

	debug_print(_("queueing message that failed to send...\n"));
//	g_return_val_if_fail(compose->to_list != NULL, -1);

	num = procmsg_get_last_message_number(QUEUE_DIR);
	if (num < 0) return -1;

	destfile = g_strdup_printf
		("%s%c%d", QUEUE_DIR, G_DIR_SEPARATOR, num + 1);
	if ((fp = fopen(destfile, "w")) == NULL) {
		FILE_OP_ERROR(destfile, "fopen");
		g_free(destfile);
		return -1;
	}
	if ((src_fp = fopen(file, "r")) == NULL) {
		FILE_OP_ERROR(file, "fopen");
		fclose(fp);
		unlink(destfile);
		g_free(destfile);
		return -1;
	}
	if (chmod(destfile, S_IRUSR|S_IWUSR) < 0) {
		FILE_OP_ERROR(destfile, "chmod");
		g_warning(_("can't change file mode\n"));
	}

	/* queueing variables */
	fprintf(fp, "AF:\n");
	fprintf(fp, "NF:0\n");
	fprintf(fp, "PS:10\n");
	fprintf(fp, "SRH:1\n");
	fprintf(fp, "SFN:\n");
	fprintf(fp, "DSR:\n");
	if (compose->msgid)
		fprintf(fp, "MID:<%s>\n", compose->msgid);
	else
		fprintf(fp, "MID:\n");
	fprintf(fp, "CFG:\n");
	fprintf(fp, "PT:0\n");
	fprintf(fp, "S:%s\n", cur_account->address);
	fprintf(fp, "RQ:\n");
	fprintf(fp, "SSV:%s\n", cur_account->smtp_server);
	if (compose->nserver!=NULL) fprintf(fp, "NSV:%s\n", compose->nserver);
	fprintf(fp, "SSH:\n");
	if (compose->to_list!=NULL) {
	    fprintf(fp, "R:<%s>", (gchar *)compose->to_list->data);
	    for (cur = compose->to_list->next; cur != NULL; cur = cur->next)
		    fprintf(fp, ",<%s>", (gchar *)cur->data);
	    fprintf(fp, "\n");
	} else fprintf(fp, "R:\n");
	if (compose->ng_list!=NULL) {
	    fprintf(fp, "N:%s", (gchar *)compose->ng_list->data);
	    for (cur = compose->ng_list->next; cur != NULL; cur = cur->next)
		    fprintf(fp, ",%s", (gchar *)cur->data);
	    fprintf(fp, "\n");
	} else fprintf(fp, "N:\n");
	fprintf(fp, "\n");

	while (fgets(buf, sizeof(buf), src_fp) != NULL) {
		if (fputs(buf, fp) == EOF) {
			FILE_OP_ERROR(destfile, "fputs");
			fclose(fp);
			fclose(src_fp);
			unlink(destfile);
			g_free(destfile);
			return -1;
		}
	}

	fclose(src_fp);
	if (fclose(fp) == EOF) {
		FILE_OP_ERROR(destfile, "fclose");
		unlink(destfile);
		g_free(destfile);
		return -1;
	}
	g_free(destfile);

	if ((fp = procmsg_open_mark_file(QUEUE_DIR, TRUE)) == NULL)
		g_warning(_("can't open mark file\n"));
	else {
		MsgInfo newmsginfo;

		newmsginfo.msgnum = num + 1;
		newmsginfo.flags = 0;
		procmsg_write_flags(&newmsginfo, fp);
		if (fclose(fp) == EOF)
			FILE_OP_ERROR("markfile", "fclose");
	}

	return 0;
}

static void compose_write_attach(Compose *compose, FILE *fp)
{
	AttachInfo *ainfo;
	GtkCList *clist = GTK_CLIST(compose->attach_clist);
	gint row;
	gchar inbuf[B64_LINE_SIZE], outbuf[B64_BUFFSIZE];
	FILE *attach_fp;
	gchar filename[BUFFSIZE];
	gint len;

	for (row = 0; (ainfo = gtk_clist_get_row_data(clist, row)) != NULL;
	     row++) {
		if ((attach_fp = fopen(ainfo->file, "r")) == NULL) {
			g_warning(_("Can't open file %s\n"), ainfo->file);
			continue;
		}

/*   		conv_encode_header(filename, sizeof(filename), ainfo->name, 12); */

		fprintf(fp, "\n--%s\n", compose->boundary);

		if (!strcmp2(ainfo->content_type, "message/rfc822")) {
			fprintf(fp, "Content-Type: %s\n", ainfo->content_type);
			fprintf(fp, "Content-Disposition: inline\n");
		} else {
                    compose_convert_header(filename, sizeof(filename),
					   ainfo->name, 12);
			fprintf(fp, "Content-Type: %s;\n"
				    " name=\"%s\"\n",
				ainfo->content_type, filename);
			fprintf(fp, "Content-Disposition: attachment;\n"
				    " filename=\"%s\"\n", filename);
		}
		fprintf(fp, "Content-Transfer-Encoding: %s\n\n",
			procmime_get_encoding_str(ainfo->encoding));

		if (ainfo->encoding == ENC_7BIT) {
			gchar buf[BUFFSIZE];

			while (fgets(buf, sizeof(buf), attach_fp) != NULL) {
				len = strlen(buf);
				if (len > 1 && buf[len - 1] == '\n' &&
				    buf[len - 2] == '\r') {
					buf[len - 2] = '\n';
					buf[len - 1] = '\0';
				}
				fputs(buf, fp);
			}
		} else {
			gchar inbuf[B64_LINE_SIZE], outbuf[B64_BUFFSIZE];

			while ((len = fread(inbuf, sizeof(gchar),
					    B64_LINE_SIZE, attach_fp))== B64_LINE_SIZE) {
				to64frombits(outbuf, inbuf, B64_LINE_SIZE);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}
			if (len > 0 && feof(attach_fp)) {
				to64frombits(outbuf, inbuf, len);
				fputs(outbuf, fp);
				fputc('\n', fp);
			}

		}

		fclose(attach_fp);
	}

	fprintf(fp, "\n--%s--\n", compose->boundary);
}

static gint compose_write_headers_smtp(Compose *compose, FILE *fp,
				  const gchar *charset, EncodingType encoding,
				  gboolean is_draft)

{
	gchar buf[BUFFSIZE];
	gchar *str;
	struct utsname utsbuf;
	g_return_val_if_fail(fp != NULL, -1);
	g_return_val_if_fail(charset != NULL, -1);
	g_return_val_if_fail(cur_account != NULL, -1);
	g_return_val_if_fail(cur_account->address != NULL, -1);

	/* Date */
	if (cur_account->add_date) {
		get_rfc822_date(buf, sizeof(buf));
		fprintf(fp, "Date: %s\n", buf);
	}

	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->from_entry)->entry));
	/* From */
        if(str) {
		compose_convert_header(buf, sizeof(buf), str,
				       strlen("From: "));
		fprintf(fp, "From: %s\n", buf);
	} else if (cur_account->name && *cur_account->name) {
		compose_convert_header(buf, sizeof(buf), cur_account->name,
				       strlen("From: "));
		fprintf(fp, "From: %s <%s>\n", buf, cur_account->address);
	} else
		fprintf(fp, "From: %s\n", cur_account->address);

	slist_free_strings(compose->to_list);
	g_slist_free(compose->to_list);
	compose->to_list = NULL;

	/* To */
	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry));
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose->to_list =
				address_list_append(compose->to_list, str);
			compose_convert_header(buf, sizeof(buf), str,
					       strlen("To: "));
			fprintf(fp, "To: %s\n", buf);
		} else if (!is_draft)return -1;
	} else
		return -1;

	/* Cc */
	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->cc_entry)->entry));
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose->to_list =
				address_list_append(compose->to_list, str);
			compose_convert_header(buf, sizeof(buf), str,
					       strlen("Cc: "));
			fprintf(fp, "Cc: %s\n", buf);
		}
	}

	/* Bcc */
	if (compose->use_bcc) {
		str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->bcc_entry)->entry));
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose->to_list = address_list_append
					(compose->to_list, str);
				if (is_draft) {
					compose_convert_header
						(buf, sizeof(buf), str,
						 strlen("Bcc: "));
					fprintf(fp, "Bcc: %s\n", buf);
    			    }
			}
		}
	}

	/* Subject */
	str = gtk_entry_get_text(GTK_ENTRY(compose->subject_entry));
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose_convert_header_misc(buf, sizeof(buf), str,
					       strlen("Subject: "));
			fprintf(fp, "Subject: %s\n", buf);
		}
	}

	/* Message-ID */
	if (cur_account->gen_msgid) {
		compose_generate_msgid(buf, sizeof(buf));
		fprintf(fp, "Message-Id: <%s>\n", buf);
		compose->msgid = g_strdup(buf);
	}

	/* In-Reply-To */
	if (compose->inreplyto)
		fprintf(fp, "In-Reply-To: <%s>\n", compose->inreplyto);

	/* References */
	if (compose->references)
		fprintf(fp, "References: %s\n", compose->references);

	/* Reply-To */
	if (compose->use_replyto) {
		str = gtk_entry_get_text(GTK_ENTRY(compose->reply_entry));
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose_convert_header(buf, sizeof(buf), str,
						       strlen("Reply-To: "));
				fprintf(fp, "Reply-To: %s\n", buf);
			}
		}
	}

	/* Program version and system info */
	uname(&utsbuf);
	fprintf(fp, "X-Mailer: %s (GTK+ %d.%d.%d; %s %s; %s)\n",
		prog_version,
		gtk_major_version, gtk_minor_version, gtk_micro_version,
		utsbuf.sysname, utsbuf.release, utsbuf.machine);

	/* Organization */
	if (cur_account->organization && *cur_account->organization) {
		compose_convert_header_misc(buf, sizeof(buf),
				       cur_account->organization,
				       strlen("Organization: "));
		fprintf(fp, "Organization: %s\n", buf);
	}

	/* MIME */
	fprintf(fp, "Mime-Version: 1.0\n");
	if (compose->use_attach) {
		get_rfc822_date(buf, sizeof(buf));
		subst_char(buf, ' ', '_');
		subst_char(buf, ',', '_');
		compose->boundary = g_strdup_printf("Multipart_%s_%08x",
						    buf, (guint)compose);
		fprintf(fp,
			"Content-Type: multipart/mixed;\n"
			" boundary=\"%s\"\n", compose->boundary);
	} else {
		fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n", charset);
		fprintf(fp, "Content-Transfer-Encoding: %s\n",
			procmime_get_encoding_str(encoding));
	}

	/* separator between header and body */
	fputs("\n", fp);

	return 0;
}

static gint compose_write_headers_nntp(Compose *compose, FILE *fp,
				  const gchar *charset, EncodingType encoding,
				  gboolean is_draft)

{
	gchar buf[BUFFSIZE];
	gchar *str;
	struct utsname utsbuf;
	g_return_val_if_fail(fp != NULL, -1);
	g_return_val_if_fail(charset != NULL, -1);
	g_return_val_if_fail(cur_account != NULL, -1);
	g_return_val_if_fail(cur_account->address != NULL, -1);

	/* Date */
	if (cur_account->add_date) {
		get_rfc822_date(buf, sizeof(buf));
		fprintf(fp, "Date: %s\n", buf);
	}

	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->from_entry)->entry));
	/* From */
        if(str) {
		compose_convert_header(buf, sizeof(buf), str,
				       strlen("From: "));
		fprintf(fp, "From: %s\n", buf);
	} else if (cur_account->name && *cur_account->name) {
		compose_convert_header(buf, sizeof(buf), cur_account->name,
				       strlen("From: "));
		fprintf(fp, "From: %s <%s>\n", buf, cur_account->address);
	} else
		fprintf(fp, "From: %s\n", cur_account->address);

	slist_free_strings(compose->ng_list);
	g_slist_free(compose->ng_list);
	compose->ng_list = NULL;

	/* Newsgroups */
	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->ng_entry)->entry));
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose->ng_list =
				address_list_append(compose->ng_list, str);
			compose_convert_header(buf, sizeof(buf), str,
					       strlen("Newsgroups: "));
			fprintf(fp, "Newsgroups: %s\n", buf);
		} else if (!is_draft)return -1;
	} else
		return -1;

	/* Subject */
	str = gtk_entry_get_text(GTK_ENTRY(compose->subject_entry));
	if (*str == '\0') str=g_strdup("(no subject)");
	Xstrdup_a(str, str, return -1);
	g_strstrip(str);
	compose_convert_header_misc(buf, sizeof(buf), str,
			            strlen("Subject: "));
	fprintf(fp, "Subject: %s\n", buf);

	/* Message-ID */
	compose_generate_msgid(buf, sizeof(buf));
	fprintf(fp, "Message-Id: <%s>\n", buf);
	compose->msgid = g_strdup(buf);

	/* In-Reply-To
	*/
	if (compose->inreplyto)
		fprintf(fp, "In-Reply-To: <%s>\n", compose->inreplyto);


	/* References */
	if (compose->references)
		fprintf(fp, "References: %s\n", compose->references);

	/* Reply-To */
	if (compose->use_replyto) {
		str = gtk_entry_get_text(GTK_ENTRY(compose->reply_entry));
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose_convert_header(buf, sizeof(buf), str,
						       strlen("Reply-To: "));
				fprintf(fp, "Reply-To: %s\n", buf);
			}
		}
	}

	/* Program version and system info */
	uname(&utsbuf);
	fprintf(fp, "X-Newsreader: %s (GTK+ %d.%d.%d; %s %s; %s)\n",
		prog_version,
		gtk_major_version, gtk_minor_version, gtk_micro_version,
		utsbuf.sysname, utsbuf.release, utsbuf.machine);

	/* Organization */
	if (cur_account->organization && *cur_account->organization) {
		compose_convert_header_misc(buf, sizeof(buf),
				            cur_account->organization,
				            strlen("Organization: "));
		fprintf(fp, "Organization: %s\n", buf);
	}

	/* MIME */
	fprintf(fp, "Mime-Version: 1.0\n");
	if (compose->use_attach) {
		get_rfc822_date(buf, sizeof(buf));
		subst_char(buf, ' ', '_');
		subst_char(buf, ',', '_');
		compose->boundary = g_strdup_printf("Multipart_%s_%08x",
						    buf, (guint)compose);
		fprintf(fp,
			"Content-Type: multipart/mixed;\n"
			" boundary=\"%s\"\n", compose->boundary);
	} else {
		fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n", charset);
		fprintf(fp, "Content-Transfer-Encoding: %s\n",
			procmime_get_encoding_str(encoding));
	}

	/* separator between header and body */
	fputs("\n", fp);

	return 0;
}

static gint compose_write_headers_queue(Compose *compose, FILE *fp,
				  const gchar *charset, EncodingType encoding,
				  gboolean is_draft)
{
	gchar buf[BUFFSIZE];
	gchar *strs,*strn,*str;
	struct utsname utsbuf;
	g_return_val_if_fail(fp != NULL, -1);
	g_return_val_if_fail(charset != NULL, -1);
	g_return_val_if_fail(cur_account != NULL, -1);
	g_return_val_if_fail(cur_account->address != NULL, -1);

	/* Date */
	if (cur_account->add_date) {
		get_rfc822_date(buf, sizeof(buf));
		fprintf(fp, "Date: %s\n", buf);
	}

	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->from_entry)->entry));
	/* From */
        if(str) {
		compose_convert_header(buf, sizeof(buf), str,
				       strlen("From: "));
		fprintf(fp, "From: %s\n", buf);
	} else if (cur_account->name && *cur_account->name) {
		compose_convert_header(buf, sizeof(buf), cur_account->name,
				       strlen("From: "));
		fprintf(fp, "From: %s <%s>\n", buf, cur_account->address);
	} else
		fprintf(fp, "From: %s\n", cur_account->address);

	slist_free_strings(compose->to_list);
	g_slist_free(compose->to_list);
	compose->to_list = NULL;

	/* To */
	strs = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry));
	strn = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->ng_entry)->entry));
	if (strs==NULL && strn==NULL) return -1;
	str=strs;
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose->to_list =
				address_list_append(compose->to_list, str);
			compose_convert_header(buf, sizeof(buf), str,
					       strlen("To: "));
			fprintf(fp, "To: %s\n", buf);
		} else if (!is_draft)return -1;
	}
	/* Newsgroups */
	str=strn;
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose->ng_list =
				address_list_append(compose->ng_list, str);
			compose_convert_header(buf, sizeof(buf), str,
					       strlen("Newsgroups: "));
			fprintf(fp, "Newsgroups: %s\n", buf);
		} else if (!is_draft)return -1;
	};


	/* Cc */
	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->cc_entry)->entry));
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose->to_list =
				address_list_append(compose->to_list, str);
			compose_convert_header(buf, sizeof(buf), str,
					       strlen("Cc: "));
			fprintf(fp, "Cc: %s\n", buf);
		}
	}

	/* Bcc */
	if (compose->use_bcc) {
		str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->bcc_entry)->entry));
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose->to_list = address_list_append
					(compose->to_list, str);
				if (is_draft) {
					compose_convert_header
						(buf, sizeof(buf), str,
						 strlen("Bcc: "));
					fprintf(fp, "Bcc: %s\n", buf);
    			    }
			}
		}
	}

	/* Subject */
	str = gtk_entry_get_text(GTK_ENTRY(compose->subject_entry));
	if (*str != '\0') {
		Xstrdup_a(str, str, return -1);
		g_strstrip(str);
		if (*str != '\0') {
			compose_convert_header_misc(buf, sizeof(buf), str,
					            strlen("Subject: "));
			fprintf(fp, "Subject: %s\n", buf);
		}
	}

	/* Message-ID */
	if (cur_account->gen_msgid) {
		compose_generate_msgid(buf, sizeof(buf));
		fprintf(fp, "Message-Id: <%s>\n", buf);
		compose->msgid = g_strdup(buf);
	}

	/* In-Reply-To */
	if (compose->inreplyto)
		fprintf(fp, "In-Reply-To: <%s>\n", compose->inreplyto);

	/* References */
	if (compose->references)
		fprintf(fp, "References: %s\n", compose->references);

	/* Reply-To */
	if (compose->use_replyto) {
		str = gtk_entry_get_text(GTK_ENTRY(compose->reply_entry));
		if (*str != '\0') {
			Xstrdup_a(str, str, return -1);
			g_strstrip(str);
			if (*str != '\0') {
				compose_convert_header(buf, sizeof(buf), str,
						       strlen("Reply-To: "));
				fprintf(fp, "Reply-To: %s\n", buf);
			}
		}
	}

	/* Program version and system info */
	uname(&utsbuf);
	fprintf(fp, "X-Mailer: %s (GTK+ %d.%d.%d; %s %s; %s)\n",
		prog_version,
		gtk_major_version, gtk_minor_version, gtk_micro_version,
		utsbuf.sysname, utsbuf.release, utsbuf.machine);

	/* Organization */
	if (cur_account->organization && *cur_account->organization) {
		compose_convert_header_misc(buf, sizeof(buf),
				            cur_account->organization,
                                            strlen("Organization: "));
		fprintf(fp, "Organization: %s\n", buf);
	}

	/* MIME */
	fprintf(fp, "Mime-Version: 1.0\n");
	if (compose->use_attach) {
		get_rfc822_date(buf, sizeof(buf));
		subst_char(buf, ' ', '_');
		subst_char(buf, ',', '_');
		compose->boundary = g_strdup_printf("Multipart_%s_%08x",
						    buf, (guint)compose);
		fprintf(fp,
			"Content-Type: multipart/mixed;\n"
			" boundary=\"%s\"\n", compose->boundary);
	} else {
		fprintf(fp, "Content-Type: text/plain; charset=\"%s\"\n", charset);
		fprintf(fp, "Content-Transfer-Encoding: %s\n",
			procmime_get_encoding_str(encoding));
	}

	/* separator between header and body */
	fputs("\n", fp);

	return 0;
}


static void compose_convert_header(gchar *dest, gint len, gchar *src,
				   gint header_len)
{
	if (len < 1) return;

	remove_return(src);

	if (is_ascii_str(src)) {
		strncpy2(dest, src, len);
		dest[len - 1] = '\0';
		return;
	} else
		conv_encode_header(dest, len, src, header_len);
}

static void compose_convert_header_misc(gchar *dest, gint len, gchar *src,
				   gint header_len)
{
	if (len < 1) return;

	remove_return(src);

	if (is_ascii_str(src)) {
		strncpy2(dest, src, len);
		dest[len - 1] = '\0';
		return;
	} else
		conv_encode_header_misc(dest, len, src, header_len);
}

static void compose_generate_msgid(gchar *buf, gint len)
{
	struct tm *lt;
	time_t t;
	gchar *addr;

	t = time(NULL);
	lt = localtime(&t);

	if (cur_account && cur_account->address && *cur_account->address) {
		if (strchr(cur_account->address, '@'))
			addr = g_strdup(cur_account->address);
		else
			addr = g_strconcat(cur_account->address, "@",
					   get_domain_name(), NULL);
	} else
		addr = g_strconcat(g_get_user_name(), "@", get_domain_name(),
				   NULL);

	g_snprintf(buf, len, "%04d%02d%02d%02d%02d%02d.%08x.%s",
		   lt->tm_year + 1900, lt->tm_mon + 1,
		   lt->tm_mday, lt->tm_hour,
		   lt->tm_min, lt->tm_sec,
		   (guint)random(), addr);

	debug_print(_("generated Message-ID: %s\n"), buf);

	g_free(addr);
}

static Compose *compose_create(void)
{
	Compose   *compose;
	GtkWidget *window;
	GtkWidget *vbox;
	GtkWidget *menubar;
	GtkWidget *handlebox;
	GtkWidget *vbox2;
	GtkWidget *table_vbox;
	GtkWidget *to_entry;
        GList     *to_glist=NULL;
	GtkWidget *ng_entry;
        GList     *ng_glist=NULL;
        GList     *from_glist=NULL;
	GtkWidget *subject_entry;
	GtkWidget *cc_entry;
	GtkWidget *bcc_entry;
	GtkWidget *bcc_hbox;
	GtkWidget *from_entry;
	GtkWidget *reply_entry;
	GtkWidget *reply_hbox;

	GtkWidget *paned;

	GtkWidget *attach_scrwin;
	GtkWidget *attach_clist;

	GtkWidget *scrolledwin;
	GtkWidget *text;

	GtkWidget *table;
	GtkWidget *label;
	GtkWidget *hbox;

        GdkPixmap *compose_icon;
        GdkBitmap *compose_icon_mask;

	gchar *titles[] = {_("MIME type"), _("Size"), _("Name")};
	guint      n_menu_entries;
	GtkStyle  *style, *new_style;
	GdkFont   *font;
	GtkWidget *popupmenu;
	GtkItemFactory *popupfactory;
	GtkItemFactory *ifactory;
	gint n_entries;
	gint i;

        char *somebuf;

	if (cur_account == NULL) {
	    alertpanel_message(_("Warrning"), _("Account not exist"));
	    return;
	};

	debug_print(_("Creating compose window...\n"));
	compose = g_new0(Compose, 1);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if (cur_account->address) {
		gchar *buf;

		buf = g_strdup_printf(_("%s - Compose message"),
				      cur_account->address);
		gtk_window_set_title(GTK_WINDOW(window), buf);
		g_free(buf);
	} else
		gtk_window_set_title(GTK_WINDOW(window), _("Compose message"));
	gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
	gtk_widget_set_usize(window, -1, prefs_common.compose_height);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			   GTK_SIGNAL_FUNC(compose_delete_cb), compose);
	gtk_signal_connect(GTK_OBJECT(window), "destroy",
			   GTK_SIGNAL_FUNC(compose_destroy_cb), compose);
	gtk_signal_connect(GTK_OBJECT(window), "focus_in_event",
			   GTK_SIGNAL_FUNC(manage_window_focus_in), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "focus_out_event",
			   GTK_SIGNAL_FUNC(manage_window_focus_out), NULL);
	gtk_widget_realize(window);

        compose_icon=gdk_pixmap_create_from_xpm_d(window->window,&compose_icon_mask,NULL,stock_mail_compose_icon_xpm);
        gdk_window_set_icon(window->window,NULL,compose_icon,compose_icon_mask);


	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	n_menu_entries = sizeof(compose_entries) / sizeof(compose_entries[0]);
	menubar = menubar_create(window, compose_entries,
				 n_menu_entries, "<Compose>", compose);
	gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, TRUE, 0);

	handlebox = gtk_handle_box_new();
	gtk_box_pack_start(GTK_BOX(vbox), handlebox, FALSE, FALSE, 0);

	compose_toolbar_create(compose, handlebox);

	vbox2 = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), vbox2, TRUE, TRUE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(vbox2), BORDER_WIDTH);

	table_vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox2), table_vbox, FALSE, TRUE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(table_vbox),
				       BORDER_WIDTH * 2);

	table = gtk_table_new(6, 2, FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table), 4);
	gtk_table_set_col_spacings(GTK_TABLE(table), 4);
	gtk_box_pack_start(GTK_BOX(table_vbox), table, FALSE, TRUE, 0);

	/* header labels */
	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("From:")
			      : "From:");
	gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 0, 1,
			 GTK_FILL, 0, 2, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("To:")
			      : "To:");
	gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 1, 2,
			 GTK_FILL, 0, 2, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("Newsgroups:")
			      : "Newsgroups:");
	gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 2, 3,
			 GTK_FILL, 0, 2, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("Subject:")
			      : "Subject:");
	gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 3, 4,
			 GTK_FILL, 0, 2, 0);
	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("Cc:")
			      : "Cc:");
	gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 0, 1, 4, 5,
			 GTK_FILL, 0, 2, 0);

	bcc_hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("Bcc:")
			      : "Bcc:");
	gtk_box_pack_end(GTK_BOX(bcc_hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), bcc_hbox, 0, 1, 5, 6,
			 GTK_FILL, 0, 2, 0);

	reply_hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(prefs_common.trans_hdr ? _("Reply-To:")
			      : "Reply-To:");
	gtk_box_pack_end(GTK_BOX(reply_hbox), label, FALSE, FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), reply_hbox, 0, 1, 6, 7,
			 GTK_FILL, 0, 2, 0);

	compose->nserver=compose->newsgroups=NULL;
	if (m_finfo && m_finfo->type == F_NEWSGROUP) {
		if (m_finfo->path != NULL) {
		    gchar** delim=g_strsplit(m_finfo->path,"/",2);
		    if (delim[0]!=NULL && delim[1]!=NULL) {
			compose->newsgroups=g_strdup(delim[1]);
			compose->nserver=g_strdup(delim[0]);
		    };
		    g_strfreev (delim);
		};
	}
	if (cur_account->nntp_for_all) compose->nserver=g_strdup(cur_account->nntp_server);

	/* header entries */
        to_glist = addresbook_recursive_list_items(NULL);
        from_glist = account_list_addresses();

	from_entry = gtk_combo_new();
        if(from_glist) gtk_combo_set_popdown_strings(GTK_COMBO(from_entry),from_glist);
	gtk_table_attach_defaults(GTK_TABLE(table), from_entry, 1, 2, 0, 1);

	to_entry = gtk_combo_new();
        if(to_glist) gtk_combo_set_popdown_strings(GTK_COMBO(to_entry),to_glist);
	gtk_table_attach_defaults(GTK_TABLE(table), to_entry, 1, 2, 1, 2);

	ng_entry = gtk_combo_new();
        if(ng_glist) gtk_combo_set_popdown_strings(GTK_COMBO(ng_entry),ng_glist);
	gtk_table_attach_defaults(GTK_TABLE(table), ng_entry, 1, 2, 2, 3);


	subject_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table), subject_entry, 1, 2, 3, 4);

	cc_entry = gtk_combo_new();
        if(to_glist) gtk_combo_set_popdown_strings(GTK_COMBO(cc_entry),to_glist);
	gtk_table_attach_defaults(GTK_TABLE(table), cc_entry, 1, 2, 4, 5);

	bcc_entry = gtk_combo_new();
        if(to_glist) gtk_combo_set_popdown_strings(GTK_COMBO(bcc_entry),to_glist);
	gtk_table_attach_defaults(GTK_TABLE(table), bcc_entry, 1, 2, 5, 6);

	reply_entry = gtk_entry_new();
	gtk_table_attach_defaults(GTK_TABLE(table), reply_entry, 1, 2, 6, 7);

        if(to_glist) g_list_free(to_glist);
        if(from_glist) g_list_free(from_glist);

	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(from_entry)->entry), "activate",
			   GTK_SIGNAL_FUNC(from_activated), compose);
        gtk_signal_connect(GTK_OBJECT(GTK_COMBO(to_entry)->entry), "activate",
                           GTK_SIGNAL_FUNC(to_activated), compose);

        gtk_signal_connect(GTK_OBJECT(GTK_COMBO(ng_entry)->entry), "activate",
                           GTK_SIGNAL_FUNC(ng_activated), compose);

	gtk_signal_connect(GTK_OBJECT(subject_entry), "activate",
			   GTK_SIGNAL_FUNC(subject_activated), compose);
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(cc_entry)->entry), "activate",
			   GTK_SIGNAL_FUNC(cc_activated), compose);
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(bcc_entry)->entry), "activate",
			   GTK_SIGNAL_FUNC(bcc_activated), compose);
	gtk_signal_connect(GTK_OBJECT(reply_entry), "activate",
			   GTK_SIGNAL_FUNC(replyto_activated), compose);

	gtk_signal_connect(GTK_OBJECT(subject_entry), "grab_focus",
			   GTK_SIGNAL_FUNC(compose_grab_focus_cb), compose);
	gtk_signal_connect(GTK_OBJECT(to_entry), "grab_focus",
			   GTK_SIGNAL_FUNC(compose_grab_focus_cb), compose);
	gtk_signal_connect(GTK_OBJECT(cc_entry), "grab_focus",
			   GTK_SIGNAL_FUNC(compose_grab_focus_cb), compose);
	gtk_signal_connect(GTK_OBJECT(bcc_entry), "grab_focus",
			   GTK_SIGNAL_FUNC(compose_grab_focus_cb), compose);
	gtk_signal_connect(GTK_OBJECT(reply_entry), "grab_focus",
			   GTK_SIGNAL_FUNC(compose_grab_focus_cb), compose);

	/* attachment list */
	attach_scrwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(attach_scrwin),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_ALWAYS);
	gtk_widget_set_usize(attach_scrwin, -1, 80);

	attach_clist = gtk_clist_new_with_titles(N_ATTACH_COLS, titles);
	gtk_clist_set_column_justification(GTK_CLIST(attach_clist), COL_SIZE,
					   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_width(GTK_CLIST(attach_clist), COL_MIMETYPE, 240);
	gtk_clist_set_column_width(GTK_CLIST(attach_clist), COL_SIZE, 64);
	gtk_clist_set_selection_mode(GTK_CLIST(attach_clist),
				     GTK_SELECTION_EXTENDED);
	//gtk_clist_set_reorderable(GTK_CLIST(attach_clist), TRUE);
	for (i = 0; i < N_ATTACH_COLS; i++)
		GTK_WIDGET_UNSET_FLAGS
			(GTK_CLIST(attach_clist)->column[i].button,
			 GTK_CAN_FOCUS);
	gtk_container_add(GTK_CONTAINER(attach_scrwin), attach_clist);

	gtk_signal_connect(GTK_OBJECT(attach_clist), "select_row",
			   GTK_SIGNAL_FUNC(attach_selected), compose);
	gtk_signal_connect(GTK_OBJECT(attach_clist), "button_press_event",
			   GTK_SIGNAL_FUNC(attach_button_pressed), compose);
	gtk_signal_connect(GTK_OBJECT(attach_clist), "key_press_event",
			   GTK_SIGNAL_FUNC(attach_key_pressed), compose);

	/* drag and drop */
	gtk_drag_dest_set(attach_clist,
			  GTK_DEST_DEFAULT_ALL, compose_mime_types, 1,
			  GDK_ACTION_COPY);
	gtk_signal_connect(GTK_OBJECT(attach_clist), "drag_data_received",
			   GTK_SIGNAL_FUNC(compose_attach_drag_received_cb),
			   compose);

	/* pane between attach clist and text */
	paned = gtk_vpaned_new();
	gtk_paned_add1(GTK_PANED(paned), attach_scrwin);
	gtk_widget_ref(paned);
	gtk_widget_show_all(paned);

	/* text widget */
	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	gtk_box_pack_start(GTK_BOX(vbox2), scrolledwin, TRUE, TRUE, 0);
	gtk_widget_set_usize(scrolledwin, prefs_common.compose_width, -1);

	text = gtk_text_new(gtk_scrolled_window_get_hadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)),
			    gtk_scrolled_window_get_vadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)));
	gtk_text_set_editable(GTK_TEXT(text), TRUE);
	gtk_text_set_word_wrap(GTK_TEXT(text), TRUE);
        gtk_text_set_word_wrap(GTK_TEXT(text), TRUE);
        gtkspell_attach(GTK_TEXT(text));



	gtk_container_add(GTK_CONTAINER(scrolledwin), text);

	gtk_signal_connect(GTK_OBJECT(text), "changed",
			   GTK_SIGNAL_FUNC(compose_changed_cb), compose);
	gtk_signal_connect(GTK_OBJECT(text), "grab_focus",
			   GTK_SIGNAL_FUNC(compose_grab_focus_cb), compose);
//	gtk_signal_connect_after(GTK_OBJECT(text), "button_press_event",
//				 GTK_SIGNAL_FUNC(compose_button_press_cb),
//				 compose);

//	gtkspell_attach(GTK_TEXT(text));
#if 0
	gtk_signal_connect_after(GTK_OBJECT(text), "key_press_event",
				 GTK_SIGNAL_FUNC(compose_key_press_cb),
				 compose);
#endif

	/* drag and drop */
	gtk_drag_dest_set(text, GTK_DEST_DEFAULT_ALL, compose_mime_types, 1,
			  GDK_ACTION_COPY);
	gtk_signal_connect(GTK_OBJECT(text), "drag_data_received",
			   GTK_SIGNAL_FUNC(compose_insert_drag_received_cb),
			   compose);

	gtk_widget_show_all(vbox);
	gtk_widget_hide(bcc_hbox);
	gtk_widget_hide(bcc_entry);
	gtk_widget_hide(reply_hbox);
	gtk_widget_hide(reply_entry);

	style = gtk_widget_get_style(text);

	/* workaround for the slow down of GtkText when using Pixmap theme */
	if (style->engine) {
		GtkThemeEngine *engine;

		engine = style->engine;
		style->engine = NULL;
		new_style = gtk_style_copy(style);
		style->engine = engine;
	} else
		new_style = gtk_style_copy(style);

	if (prefs_common.messagefont) {
		font = gdk_fontset_load(prefs_common.messagefont);
		if (font) new_style->font = font;
	}

	gtk_widget_set_style(text, new_style);

	n_entries = sizeof(compose_popup_entries) /
		sizeof(compose_popup_entries[0]);
	popupmenu = menu_create_items(compose_popup_entries, n_entries,
				      "<Compose>", &popupfactory,
				      compose);

	ifactory = gtk_item_factory_from_widget(menubar);
	menu_set_sensitive(ifactory, "/Edit/Undo",	  FALSE);
	menu_set_sensitive(ifactory, "/Edit/Redo",	  FALSE);

	switch (prefs_common.toolbar_style) {
	case TOOLBAR_NONE:
		gtk_widget_hide(handlebox);
		break;
	case TOOLBAR_ICON:
		gtk_toolbar_set_style(GTK_TOOLBAR(compose->toolbar),
				      GTK_TOOLBAR_ICONS);
		break;
	case TOOLBAR_TEXT:
		gtk_toolbar_set_style(GTK_TOOLBAR(compose->toolbar),
				      GTK_TOOLBAR_TEXT);
		break;
	case TOOLBAR_BOTH:
		gtk_toolbar_set_style(GTK_TOOLBAR(compose->toolbar),
				      GTK_TOOLBAR_BOTH);
		break;
	}

	gtk_widget_show(window);



	compose->window        = window;
	compose->vbox	       = vbox;
	compose->menubar       = menubar;
	compose->handlebox     = handlebox;

	compose->vbox2	       = vbox2;

	compose->table_vbox    = table_vbox;
	compose->table	       = table;
	compose->to_entry      = to_entry;

	compose->ng_entry      = ng_entry;

	compose->subject_entry = subject_entry;
	compose->cc_entry      = cc_entry;
	compose->bcc_hbox      = bcc_hbox;
	compose->bcc_entry     = bcc_entry;
	compose->from_entry    = from_entry;
	compose->reply_hbox    = reply_hbox;
	compose->reply_entry   = reply_entry;

	compose->paned	       = paned;

	compose->attach_scrwin = attach_scrwin;
	compose->attach_clist  = attach_clist;

	compose->scrolledwin   = scrolledwin;
	compose->text	       = text;

	compose->focused_editable = NULL;

	compose->popupmenu    = popupmenu;
	compose->popupfactory = popupfactory;

	compose->mode = COMPOSE_NEW;
	compose->replyto    = NULL;
	compose->cc	    = NULL;
	compose->bcc	    = NULL;
	compose->inreplyto  = NULL;
	compose->references = NULL;
	compose->msgid      = NULL;
	compose->to_list    = NULL;
	compose->ng_list    = NULL;
	compose->boundary   = NULL;


	compose->use_bcc     = FALSE;
	compose->use_replyto = FALSE;
	compose->use_attach  = FALSE;

	compose->modified = FALSE;

	compose->exteditor_file    = NULL;
	compose->exteditor_pid     = -1;
	compose->exteditor_readdes = -1;
	compose->exteditor_tag     = -1;

        if(cur_account->address) {
	        if (cur_account->name && *cur_account->name) {
                        somebuf=malloc(strlen(cur_account->name)+strlen(cur_account->address)+4);
                        if(somebuf) {
                                sprintf(somebuf,"%s <%s>",cur_account->name,cur_account->address);
	                        gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(from_entry)->entry), somebuf);
                                free(somebuf);
                        } else {
	                        gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(from_entry)->entry), cur_account->address);
                        }
                } else {
	                gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(from_entry)->entry), cur_account->address);
                }
        }

	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(to_entry)->entry), "");

	if (cur_account->set_autocc && cur_account->auto_cc) {
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(cc_entry)->entry), cur_account->auto_cc);
        } else {
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(cc_entry)->entry), "");
        }
	if (cur_account->set_autobcc && cur_account->auto_bcc) {
		GtkWidget *menuitem;

		menuitem = gtk_item_factory_get_item(ifactory, "/Message/Bcc");
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(bcc_entry)->entry),
				   cur_account->auto_bcc);
		gtk_check_menu_item_set_active
			(GTK_CHECK_MENU_ITEM(menuitem), TRUE);
	} else {
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(bcc_entry)->entry), "");
        }
	if (cur_account->set_autoreplyto && cur_account->auto_replyto) {
		GtkWidget *menuitem;

		menuitem = gtk_item_factory_get_item(ifactory,
						     "/Message/Reply to");
		gtk_entry_set_text(GTK_ENTRY(reply_entry),
				   cur_account->auto_replyto);
		gtk_check_menu_item_set_active
			(GTK_CHECK_MENU_ITEM(menuitem), TRUE);
	}

	addressbook_set_target_compose(compose);

	compose_list = g_list_append(compose_list, compose);


	return compose;
}

#include "pixmaps/linewrap.xpm"
#include "pixmaps/stock_mail_send.xpm"
#include "pixmaps/stock_mail_send_queue.xpm"
#include "pixmaps/stock_mail.xpm"
#include "pixmaps/stock_paste.xpm"
#include "pixmaps/stock_addresbook.xpm"
#include "pixmaps/stock_attach.xpm"
#include "pixmaps/stock_mail_compose.xpm"
#include "pixmaps/stock_mail_checkspell.xpm"

#include "pixmaps/stock_draft.xpm"
#include "pixmaps/stock_mail_put_queue.xpm"
#include "pixmaps/stock_external.xpm"

#define CREATE_TOOLBAR_ICON(xpm_d) \
{ \
	icon = gdk_pixmap_create_from_xpm_d(container->window, &mask, \
					    &container->style->white, \
					    xpm_d); \
	icon_wid = gtk_pixmap_new(icon, mask); \
}

static void compose_toolbar_create(Compose *compose, GtkWidget *container)
{
	GtkWidget *toolbar;
	GdkPixmap *icon;
	GdkBitmap *mask;
	GtkWidget *icon_wid;
	GtkWidget *send_btn;
	GtkWidget *sendl_btn;
	GtkWidget *draft_btn;
	GtkWidget *insert_btn;
	GtkWidget *attach_btn;
	GtkWidget *sig_btn;
	GtkWidget *exteditor_btn;
	GtkWidget *linewrap_btn;
	GtkWidget *checkspell_btn;
	GtkWidget *addrbook_btn;

	toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL,
				  GTK_TOOLBAR_BOTH);
	gtk_container_add(GTK_CONTAINER(container), toolbar);
	gtk_container_set_border_width(GTK_CONTAINER(container), 2);
	gtk_toolbar_set_button_relief(GTK_TOOLBAR(toolbar), GTK_RELIEF_NONE);
	gtk_toolbar_set_space_style(GTK_TOOLBAR(toolbar),
				    GTK_TOOLBAR_SPACE_LINE);

	CREATE_TOOLBAR_ICON(stock_mail_send_xpm);
	send_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					   _("Send"),
					   _("Send message"),
					   "Send",
					   icon_wid, toolbar_send_cb, compose);

	CREATE_TOOLBAR_ICON(stock_mail_put_queue_xpm);
	sendl_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					   _("Send later"),
					   _("Put into queue folder and send later"),
					   "Send later",
					   icon_wid, toolbar_send_later_cb,
					   compose);

	CREATE_TOOLBAR_ICON(stock_draft_xpm);
	draft_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					    _("Draft"),
					    _("Save to draft folder"),
					    "Draft",
					    icon_wid, toolbar_draft_cb,
					    compose);

	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	CREATE_TOOLBAR_ICON(stock_paste_xpm);
	insert_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					     _("Insert"),
					     _("Insert file"),
					     "Insert",
					     icon_wid, toolbar_insert_cb,
					     compose);

	CREATE_TOOLBAR_ICON(stock_attach_xpm);
	attach_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					     _("Attach"),
					     _("Attach file"),
					     "Attach",
					     icon_wid, toolbar_attach_cb,
					     compose);

	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	CREATE_TOOLBAR_ICON(stock_mail_xpm);
	sig_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					  _("Signature"),
					  _("Insert signature"),
					  "Signature",
					  icon_wid, toolbar_sig_cb, compose);

	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	CREATE_TOOLBAR_ICON(stock_external_xpm);
	exteditor_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
						_("Editor"),
						_("Edit with external editor"),
						"Editor",
						icon_wid,
						toolbar_ext_editor_cb,
						compose);

	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	if (prefs_common.auto_ispell) { /* if gtkspell enabled */
	CREATE_TOOLBAR_ICON(stock_mail_checkspell_xpm);
	checkspell_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					       _("Spell checking"),
					       _("Check spell"),
					       "Spell checking",
					       icon_wid, toolbar_checkspell_cb,
					       compose);
	}; /* if not gtkspell */

	CREATE_TOOLBAR_ICON(linewrap_xpm);
	linewrap_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					       _("Linewrap"),
					       _("Wrap long lines"),
					       "Linewrap",
					       icon_wid,
					       toolbar_linewrap_cb,
					       compose);

	gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

	CREATE_TOOLBAR_ICON(stock_addresbook_xpm);
	addrbook_btn = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
					       _("Address"),
					       _("Address book"),
					       "Address",
					       icon_wid, toolbar_address_cb,
					       compose);

	compose->toolbar       = toolbar;
	compose->send_btn      = send_btn;
	compose->sendl_btn     = sendl_btn;
	compose->draft_btn     = draft_btn;
	compose->insert_btn    = insert_btn;
	compose->attach_btn    = attach_btn;
	compose->sig_btn       = sig_btn;
	compose->exteditor_btn = exteditor_btn;
	compose->linewrap_btn  = linewrap_btn;
	if (prefs_common.auto_ispell) { /* if gtkspell enabled */
	compose->checkspell_btn  = checkspell_btn;
	};
	compose->addrbook_btn  = addrbook_btn;

	gtk_widget_show_all(toolbar);
}

static void compose_destroy(Compose *compose)
{
	gint row;
	GtkCList *clist = GTK_CLIST(compose->attach_clist);
	AttachInfo *ainfo;

	slist_free_strings(compose->to_list);
	g_slist_free(compose->to_list);
	slist_free_strings(compose->ng_list);
	g_slist_free(compose->ng_list);

	procmsg_msginfo_free(compose->targetinfo);


	g_free(compose->replyto);
	g_free(compose->cc);
	g_free(compose->bcc);
	g_free(compose->inreplyto);
	g_free(compose->references);
	g_free(compose->msgid);
	g_free(compose->boundary);

	for (row = 0; (ainfo = gtk_clist_get_row_data(clist, row)) != NULL;
	     row++)
		compose_attach_info_free(ainfo);

	if (addressbook_get_target_compose() == compose)
		addressbook_set_target_compose(NULL);

	prefs_common.compose_width = compose->scrolledwin->allocation.width;
	prefs_common.compose_height = compose->window->allocation.height;

	gtk_widget_destroy(compose->paned);

	g_free(compose);

	compose_list = g_list_remove(compose_list, compose);
}

static void compose_attach_info_free(AttachInfo *ainfo)
{
	g_free(ainfo->file);
	g_free(ainfo->content_type);
	g_free(ainfo->name);
	g_free(ainfo);
}

static void compose_attach_remove_selected(Compose *compose)
{
	GtkCList *clist = GTK_CLIST(compose->attach_clist);
	AttachInfo *ainfo;
	gint row;

	while (clist->selection != NULL) {
		row = GPOINTER_TO_INT(clist->selection->data);
		ainfo = gtk_clist_get_row_data(clist, row);
		compose_attach_info_free(ainfo);
		gtk_clist_remove(clist, row);
	}
}

static struct _AttachProperty
{
	GtkWidget *window;
	GtkWidget *mimetype_entry;
	GtkWidget *encoding_optmenu;
	GtkWidget *path_entry;
	GtkWidget *filename_entry;
	GtkWidget *ok_btn;
	GtkWidget *cancel_btn;
} attach_prop;

static void compose_attach_property(Compose *compose)
{
	GtkCList *clist = GTK_CLIST(compose->attach_clist);
	AttachInfo *ainfo;
	gint row;
	gboolean cancelled;
	GtkOptionMenu *optmenu;

	if (!clist->selection) return;
	row = GPOINTER_TO_INT(clist->selection->data);

	ainfo = gtk_clist_get_row_data(clist, row);
	if (!ainfo) return;

	if (!attach_prop.window)
		compose_attach_property_create(&cancelled);
	gtk_widget_grab_focus(attach_prop.ok_btn);
	gtk_widget_show(attach_prop.window);
	manage_window_set_transient(GTK_WINDOW(attach_prop.window));

	optmenu = GTK_OPTION_MENU(attach_prop.encoding_optmenu);
	if (ainfo->encoding == ENC_UNKNOWN)
		gtk_option_menu_set_history(optmenu, ENC_BASE64);
	else
		gtk_option_menu_set_history(optmenu, ainfo->encoding);

	gtk_entry_set_text(GTK_ENTRY(attach_prop.mimetype_entry),
			   ainfo->content_type ? ainfo->content_type : "");
	gtk_entry_set_text(GTK_ENTRY(attach_prop.path_entry),
			   ainfo->file ? ainfo->file : "");
	gtk_entry_set_text(GTK_ENTRY(attach_prop.filename_entry),
			   ainfo->name ? ainfo->name : "");

	for (;;) {
		gchar *text;
		gchar *cnttype = NULL;
		gchar *file = NULL;
		off_t size = 0;
		GtkWidget *menu;
		GtkWidget *menuitem;

		gtk_main();

		if (cancelled == TRUE) {
			gtk_widget_hide(attach_prop.window);
			break;
		}

		text = gtk_entry_get_text(GTK_ENTRY(attach_prop.mimetype_entry));
		if (*text != '\0') {
			gchar *p;

			text = g_strstrip(g_strdup(text));
			if ((p = strchr(text, '/')) && !strchr(p + 1, '/')) {
				cnttype = g_strdup(text);
				g_free(text);
			} else {
				alertpanel_error(_("Invalid MIME type."));
				g_free(text);
				continue;
			}
		}

		menu = gtk_option_menu_get_menu(optmenu);
		menuitem = gtk_menu_get_active(GTK_MENU(menu));
		ainfo->encoding = GPOINTER_TO_INT
			(gtk_object_get_user_data(GTK_OBJECT(menuitem)));

		text = gtk_entry_get_text(GTK_ENTRY(attach_prop.path_entry));
		if (*text != '\0') {
			if (is_file_exist(text) &&
			    (size = get_file_size(text)) > 0)
				file = g_strdup(text);
			else {
				alertpanel_error
					(_("File doesn't exist or is empty."));
				g_free(cnttype);
				continue;
			}
		}

		text = gtk_entry_get_text(GTK_ENTRY(attach_prop.filename_entry));
		if (*text != '\0') {
			g_free(ainfo->name);
			ainfo->name = g_strdup(text);
		}

		if (cnttype) {
			g_free(ainfo->content_type);
			ainfo->content_type = cnttype;
		}
		if (file) {
			g_free(ainfo->file);
			ainfo->file = file;
		}
		if (size)
			ainfo->size = size;

		gtk_clist_set_text(clist, row, COL_MIMETYPE,
				   ainfo->content_type);
		gtk_clist_set_text(clist, row, COL_SIZE,
				   to_human_readable(ainfo->size));
		gtk_clist_set_text(clist, row, COL_NAME, ainfo->name);

		gtk_widget_hide(attach_prop.window);
		break;
	}
}

#define SET_LABEL_AND_ENTRY(str, entry, top) \
{ \
	label = gtk_label_new(str); \
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, top, (top + 1), \
			 GTK_FILL, 0, 0, 0); \
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5); \
 \
	entry = gtk_entry_new(); \
	gtk_table_attach(GTK_TABLE(table), entry, 1, 2, top, (top + 1), \
			 GTK_EXPAND|GTK_SHRINK|GTK_FILL, 0, 0, 0); \
}

static void compose_attach_property_create(gboolean *cancelled)
{
	GtkWidget *window;
	GtkWidget *vbox;
	GtkWidget *table;
	GtkWidget *label;
	GtkWidget *mimetype_entry;
	GtkWidget *hbox;
	GtkWidget *optmenu;
	GtkWidget *optmenu_menu;
	GtkWidget *menuitem;
	GtkWidget *path_entry;
	GtkWidget *filename_entry;
	GtkWidget *hbbox;
	GtkWidget *ok_btn;
	GtkWidget *cancel_btn;

	debug_print("Creating attach_property window...\n");

	window = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(window, 480, -1);
	gtk_container_set_border_width(GTK_CONTAINER(window), 8);
	gtk_window_set_title(GTK_WINDOW(window), _("Property"));
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
	gtk_window_set_modal(GTK_WINDOW(window), TRUE);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			   GTK_SIGNAL_FUNC(attach_property_delete_event),
			   cancelled);
	gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
			   GTK_SIGNAL_FUNC(attach_property_key_pressed),
			   cancelled);

	vbox = gtk_vbox_new(FALSE, 8);
	gtk_container_add(GTK_CONTAINER(window), vbox);

	table = gtk_table_new(4, 2, FALSE);
	gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);
	gtk_table_set_row_spacings(GTK_TABLE(table), 8);
	gtk_table_set_col_spacings(GTK_TABLE(table), 8);

	SET_LABEL_AND_ENTRY(_("MIME type"), mimetype_entry, 0);

	label = gtk_label_new(_("Encoding"));
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
			 GTK_FILL, 0, 0, 0);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), hbox, 1, 2, 1, 2,
			 GTK_EXPAND|GTK_SHRINK|GTK_FILL, 0, 0, 0);

	optmenu = gtk_option_menu_new();
	gtk_box_pack_start(GTK_BOX(hbox), optmenu, FALSE, FALSE, 0);

	optmenu_menu = gtk_menu_new();
	MENUITEM_ADD(optmenu_menu, menuitem, "7bit", ENC_7BIT);
	gtk_widget_set_sensitive(menuitem, FALSE);
	MENUITEM_ADD(optmenu_menu, menuitem, "8bit", ENC_8BIT);
	gtk_widget_set_sensitive(menuitem, FALSE);
	MENUITEM_ADD(optmenu_menu, menuitem, "quoted-printable", ENC_QUOTED_PRINTABLE);
	gtk_widget_set_sensitive(menuitem, FALSE);
	MENUITEM_ADD(optmenu_menu, menuitem, "base64", ENC_BASE64);

	gtk_option_menu_set_menu(GTK_OPTION_MENU(optmenu), optmenu_menu);

	SET_LABEL_AND_ENTRY(_("Path"),      path_entry,     2);
	SET_LABEL_AND_ENTRY(_("File name"), filename_entry, 3);

	gtkut_button_set_create(&hbbox, &ok_btn, _("OK"),
				&cancel_btn, _("Cancel"), NULL, NULL);
	gtk_box_pack_end(GTK_BOX(vbox), hbbox, FALSE, FALSE, 0);
	gtk_widget_grab_default(ok_btn);

	gtk_signal_connect(GTK_OBJECT(ok_btn), "clicked",
			   GTK_SIGNAL_FUNC(attach_property_ok),
			   cancelled);
	gtk_signal_connect(GTK_OBJECT(cancel_btn), "clicked",
			   GTK_SIGNAL_FUNC(attach_property_cancel),
			   cancelled);

	gtk_widget_show_all(vbox);

	attach_prop.window           = window;
	attach_prop.mimetype_entry   = mimetype_entry;
	attach_prop.encoding_optmenu = optmenu;
	attach_prop.path_entry       = path_entry;
	attach_prop.filename_entry   = filename_entry;
	attach_prop.ok_btn           = ok_btn;
	attach_prop.cancel_btn       = cancel_btn;
}

static void attach_property_ok(GtkWidget *widget, gboolean *cancelled)
{
	*cancelled = FALSE;
	gtk_main_quit();
}

static void attach_property_cancel(GtkWidget *widget, gboolean *cancelled)
{
	*cancelled = TRUE;
	gtk_main_quit();
}

static gint attach_property_delete_event(GtkWidget *widget, GdkEventAny *event,
					 gboolean *cancelled)
{
	*cancelled = TRUE;
	gtk_main_quit();

	return TRUE;
}

static void attach_property_key_pressed(GtkWidget *widget, GdkEventKey *event,
					gboolean *cancelled)
{
	if (event && event->keyval == GDK_Escape) {
		*cancelled = TRUE;
		gtk_main_quit();
	}
}

static void compose_exec_ext_editor(Compose *compose)
{
	gchar tmp[64];
	pid_t pid;
	gint pipe_fds[2];

	g_snprintf(tmp, sizeof(tmp), "%s%ctmpmsg.%08x",
		   g_get_tmp_dir(), G_DIR_SEPARATOR, (gint)compose);

	if (pipe(pipe_fds) < 0) {
		perror("pipe");
		return;
	}

	if ((pid = fork()) < 0) {
		perror("fork");
		return;
	}

	if (pid != 0) {
		/* close the write side of the pipe */
		close(pipe_fds[1]);

		compose->exteditor_file    = g_strdup(tmp);
		compose->exteditor_pid     = pid;
		compose->exteditor_readdes = pipe_fds[0];

		compose_set_ext_editor_sensitive(compose, FALSE);

		compose->exteditor_tag =
			gdk_input_add(pipe_fds[0], GDK_INPUT_READ,
				      compose_input_cb, compose);
	} else {	/* process-monitoring process */
		pid_t pid_ed;

		if (setpgid(0, 0))
			perror("setpgid");

		/* close the read side of the pipe */
		close(pipe_fds[0]);

		if (compose_write_body_to_file(compose, tmp) < 0) {
			sock_write(pipe_fds[1], "2\n", 2);
			_exit(1);
		}

		pid_ed = compose_exec_ext_editor_real(tmp);
		if (pid_ed < 0) {
			sock_write(pipe_fds[1], "1\n", 2);
			_exit(1);
		}

		/* wait until editor is terminated */
		waitpid(pid_ed, NULL, 0);

		sock_write(pipe_fds[1], "0\n", 2);

		close(pipe_fds[1]);
		_exit(0);
	}
}

static gint compose_exec_ext_editor_real(const gchar *file)
{
	static gchar *def_cmd = "emacs %s";
	gchar buf[1024];
	gchar *p;
	gchar **cmdline;
	pid_t pid;

	g_return_val_if_fail(file != NULL, -1);

	if ((pid = fork()) < 0) {
		perror("fork");
		return -1;
	}

	if (pid != 0) return pid;

	/* grandchild process */

	if (setpgid(0, getppid()))
		perror("setpgid");

	if (prefs_common.ext_editor_cmd &&
	    (p = strchr(prefs_common.ext_editor_cmd, '%')) &&
	    *(p + 1) == 's' && !strchr(p + 2, '%')) {
		g_snprintf(buf, sizeof(buf), prefs_common.ext_editor_cmd, file);
	} else {
		if (prefs_common.ext_editor_cmd)
			g_warning(_("External editor command line is invalid: `%s'\n"),
				  prefs_common.ext_editor_cmd);
		g_snprintf(buf, sizeof(buf), def_cmd, file);
	}

	cmdline = g_strsplit(buf, " ", 1024);
	execvp(cmdline[0], cmdline);

	perror("execvp");
	g_strfreev(cmdline);

	_exit(1);
}

static gboolean compose_ext_editor_kill(Compose *compose)
{
	pid_t pgid = compose->exteditor_pid * -1;
	gint ret;

	ret = kill(pgid, 0);

	if (ret == 0 || (ret == -1 && EPERM == errno)) {
		AlertValue val;
		gchar *msg;

		msg = g_strdup_printf
			(_("The external editor is still working.\n"
			   "Force terminating the process?\n"
			   "process group id: %d"), -pgid);
		val = alertpanel(_("Notice"), msg, _("Yes"), _("+No"), NULL);
		g_free(msg);

		if (val == G_ALERTDEFAULT) {
			gdk_input_remove(compose->exteditor_tag);
			close(compose->exteditor_readdes);

			if (kill(pgid, SIGTERM) < 0) perror("kill");
			waitpid(compose->exteditor_pid, NULL, 0);

			g_warning(_("Terminated process group id: %d"), -pgid);
			g_warning(_("Temporary file: %s"),
				  compose->exteditor_file);

			compose_set_ext_editor_sensitive(compose, TRUE);

			g_free(compose->exteditor_file);
			compose->exteditor_file    = NULL;
			compose->exteditor_pid     = -1;
			compose->exteditor_readdes = -1;
			compose->exteditor_tag     = -1;
		} else
			return FALSE;
	}

	return TRUE;
}

static void compose_input_cb(gpointer data, gint source,
			     GdkInputCondition condition)
{
	gchar buf[3];
	Compose *compose = (Compose *)data;
	gint i = 0;

	debug_print(_("Compose: input from monitoring process\n"));

	gdk_input_remove(compose->exteditor_tag);

	for (;;) {
		if (read(source, &buf[i], 1) < 1) {
			buf[0] = '3';
			break;
		}
		if (buf[i] == '\n') {
			buf[i] = '\0';
			break;
		}
		i++;
		if (i == sizeof(buf) - 1)
			break;
	}

	waitpid(compose->exteditor_pid, NULL, 0);

	if (buf[0] == '0') {		/* success */
		GtkText *text = GTK_TEXT(compose->text);

		gtk_text_freeze(text);
		gtk_text_set_point(text, 0);
		gtk_text_forward_delete(text, gtk_text_get_length(text));
		compose_insert_file(compose, compose->exteditor_file,FALSE);
		compose_changed_cb(NULL, compose);
		gtk_text_thaw(text);

		if (unlink(compose->exteditor_file) < 0)
			FILE_OP_ERROR(compose->exteditor_file, "unlink");
	} else if (buf[0] == '1') {	/* failed */
		g_warning(_("Couldn't exec external editor\n"));
		if (unlink(compose->exteditor_file) < 0)
			FILE_OP_ERROR(compose->exteditor_file, "unlink");
	} else if (buf[0] == '2') {
		g_warning(_("Couldn't write to file\n"));
	} else if (buf[0] == '3') {
		g_warning(_("Pipe read failed\n"));
	}

	close(source);

	compose_set_ext_editor_sensitive(compose, TRUE);

	g_free(compose->exteditor_file);
	compose->exteditor_file    = NULL;
	compose->exteditor_pid     = -1;
	compose->exteditor_readdes = -1;
	compose->exteditor_tag     = -1;
}

static void compose_set_ext_editor_sensitive(Compose *compose,
					     gboolean sensitive)
{
	GtkItemFactory *ifactory;

	ifactory = gtk_item_factory_from_widget(compose->menubar);

	menu_set_sensitive(ifactory, "/Message/Send", sensitive);
	menu_set_sensitive(ifactory, "/Message/Send later", sensitive);
	menu_set_sensitive(ifactory, "/Message/Save to draft folder",
			   sensitive);
	menu_set_sensitive(ifactory, "/File/Insert file", sensitive);
	menu_set_sensitive(ifactory, "/File/Insert signature", sensitive);
	menu_set_sensitive(ifactory, "/Edit/Edit with external editor",
			   sensitive);

	gtk_widget_set_sensitive(compose->text, sensitive);
	gtk_widget_set_sensitive(compose->send_btn, sensitive);
	gtk_widget_set_sensitive(compose->sendl_btn, sensitive);
	gtk_widget_set_sensitive(compose->draft_btn, sensitive);
	gtk_widget_set_sensitive(compose->insert_btn, sensitive);
	gtk_widget_set_sensitive(compose->sig_btn, sensitive);
	gtk_widget_set_sensitive(compose->exteditor_btn, sensitive);
}

/* callback functions */

static void toolbar_send_cb(GtkWidget *widget, gpointer data)
{
	compose_send_cb(data, 0, NULL);
}

static void toolbar_send_later_cb(GtkWidget *widget, gpointer data)
{
	compose_send_later_cb(data, 0, NULL);
}

static void toolbar_draft_cb(GtkWidget *widget, gpointer data)
{
	compose_draft_cb(data, 0, NULL);
}

static void toolbar_insert_cb(GtkWidget *widget, gpointer data)
{
	compose_insert_file_cb(data, 0, NULL);
}

static void toolbar_attach_cb(GtkWidget *widget, gpointer data)
{
	compose_attach_cb(data, 0, NULL);
}

static void toolbar_sig_cb(GtkWidget *widget, gpointer data)
{
	Compose *compose = (Compose *)data;

	compose_insert_sig(compose);
}

static void toolbar_ext_editor_cb(GtkWidget *widget, gpointer data)
{
	Compose *compose = (Compose *)data;

	compose_exec_ext_editor(compose);
}

static void toolbar_checkspell_cb(GtkWidget *widget, gpointer data)
{
	compose_checkspell_cb(data, 0, NULL);
}

static void toolbar_linewrap_cb(GtkWidget *widget, gpointer data)
{
	Compose *compose = (Compose *)data;

	compose_wrap_line(compose);
}

static void toolbar_address_cb(GtkWidget *widget, gpointer data)
{
	compose_address_cb(data, 0, NULL);
}

static void attach_selected(GtkCList *clist, gint row, gint column,
			    GdkEvent *event, gpointer data)
{
	Compose *compose = (Compose *)data;

	if (event && event->type == GDK_2BUTTON_PRESS)
		compose_attach_property(compose);
}

static void attach_button_pressed(GtkWidget *widget, GdkEventButton *event,
				  gpointer data)
{
	Compose *compose = (Compose *)data;
	GtkCList *clist = GTK_CLIST(compose->attach_clist);
	gint row, column;

	if (!event) return;

	if (event->button == 3) {
		if ((clist->selection && !clist->selection->next) ||
		    !clist->selection) {
			gtk_clist_unselect_all(clist);
			if (gtk_clist_get_selection_info(clist,
							 event->x, event->y,
							 &row, &column)) {
				gtk_clist_select_row(clist, row, column);
				gtkut_clist_set_focus_row(clist, row);
			}
		}
		gtk_menu_popup(GTK_MENU(compose->popupmenu), NULL, NULL,
			       NULL, NULL, event->button, event->time);
	}
}

static void attach_key_pressed(GtkWidget *widget, GdkEventKey *event,
			       gpointer data)
{
	Compose *compose = (Compose *)data;

	if (!event) return;

	switch (event->keyval) {
	case GDK_Delete:
		compose_attach_remove_selected(compose);
		break;
	}
}

static void compose_send_cb(gpointer data, guint action, GtkWidget *widget)
{
	Compose *compose = (Compose *)data;
	gint val;

	val = compose_send(compose);

	if (val == 0) gtk_widget_destroy(compose->window);
}

static void compose_send_later_cb(gpointer data, guint action,
				  GtkWidget *widget)
{
	Compose *compose = (Compose *)data;
	gchar tmp[22];
	gchar *str;

	str = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(compose->to_entry)->entry));
	if (*str == '\0') {
		alertpanel_error(_("Recipient is not specified."));
		return;
	}

	g_snprintf(tmp, 22, "%s%ctmpmsg%d",
		   g_get_tmp_dir(), G_DIR_SEPARATOR, (gint)compose);

	if (prefs_common.linewrap_at_send)
		compose_wrap_line(compose);

	if (compose_write_to_file_for_queue(compose, tmp, FALSE) < 0 ||
	    compose_queue(compose, tmp) < 0) {
		alertpanel_error(_("Can't queue the message."));
		return;
	}

	if (prefs_common.savemsg) {
		if (compose_save_to_outbox(compose, tmp) < 0)
			alertpanel_error
				(_("Can't save the message to outbox."));
	}

	if (unlink(tmp) < 0)
		FILE_OP_ERROR(tmp, "unlink");

	gtk_widget_destroy(compose->window);
}
static void compose_draft_cb(gpointer data, guint action, GtkWidget *widget)
{
	Compose *compose = (Compose *)data;
	gchar *draft_file;
	gint num;

    if (procmsg_msg_exist(compose->targetinfo)) {
		draft_file = (gchar*) procmsg_get_message_file_path(compose->targetinfo);
		if (unlink(draft_file) < 0)
			FILE_OP_ERROR(draft_file, "unlink");
	} else {
		if ((num = procmsg_get_last_message_number(DRAFT_DIR)) < 0)
			return;
		draft_file =
			g_strdup_printf("%s%c%s%c%d", maildir, G_DIR_SEPARATOR,
					DRAFT_DIR, G_DIR_SEPARATOR, num + 1);
		if (is_file_exist(draft_file)) {
			g_free(draft_file);
			return;
		}
	}

	if (compose_write_to_file_for_queue(compose, draft_file, TRUE) < 0) {
		g_free(draft_file);
		return;
	}

	g_free(draft_file);

	folderview_scan_folder_a(DRAFT_DIR, TRUE);

	gtk_widget_destroy(compose->window);
}

static void compose_attach_cb(gpointer data, guint action, GtkWidget *widget)
{
	Compose *compose = (Compose *)data;
	gchar *file,*dir;
        GList *filenamelist;
        gint i;

	filenamelist = filesel_select_file(_("Select file"), NULL, TRUE);

	if (filenamelist) {
                dir=(gchar *)g_list_nth_data(filenamelist,0);
                for(i=1;i<g_list_length(filenamelist);i++) {
                        debug_print("number: '%u' dir: '%s' file: '%s'\n",i,dir,(gchar *)g_list_nth_data(filenamelist,i));
                        file=g_strconcat(dir,(gchar *)g_list_nth_data(filenamelist,i),NULL);
		        compose_attach_append(compose, file, MIME_UNKNOWN);
                        g_free(g_list_nth_data(filenamelist,i));
                        g_free(file);
                }
                g_list_free(filenamelist);
        }
}

static void compose_insert_file_cb(gpointer data, guint action,
				   GtkWidget *widget)
{
	Compose *compose = (Compose *)data;
	gchar *file,*dir;
        GList *filenamelist=NULL;
        gint i;

	filenamelist = filesel_select_file(_("Select file"), NULL, TRUE);

	if (filenamelist) {
                dir=(gchar *)g_list_nth_data(filenamelist,0);
                for(i=1;i<g_list_length(filenamelist);i++) {
                        file=g_strconcat(dir,(gchar *)g_list_nth_data(filenamelist,i),NULL);
		        compose_insert_file(compose, file, FALSE);
                        g_free(g_list_nth_data(filenamelist,i));
                        g_free(file);
                }
                g_list_free(filenamelist);
        }
}

static gint compose_delete_cb(GtkWidget *widget, GdkEventAny *event,
			      gpointer data)
{
	compose_close_cb(data, 0, NULL);
	return TRUE;
}

static void compose_close_cb(gpointer data, guint action, GtkWidget *widget)
{
	Compose *compose = (Compose *)data;
	AlertValue val;

	if (compose->exteditor_tag != -1) {
		if (!compose_ext_editor_kill(compose))
			return;
	}

	if (compose->modified) {
		val = alertpanel(_("Discard message"),
				 _("This message has been modified.\n"
				   "discard it?"),
				 _("Discard"), _("to Draft"), _("Cancel"));

		switch (val) {
		case G_ALERTDEFAULT:
			break;
		case G_ALERTALTERNATE:
			compose_draft_cb(data, 0, NULL);
			return;
		default:
			return;
		}
	}

	gtk_widget_destroy(compose->window);
}

static void compose_checkspell_cb(gpointer data, guint action, GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	gtkspell_check_all(GTK_TEXT(compose->text));

}


static void compose_address_cb(gpointer data, guint action, GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	addressbook_open(compose);
}

static void compose_ext_editor_cb(gpointer data, guint action,
				  GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	compose_exec_ext_editor(compose);
}

static void compose_spellcheck_cb(gpointer data, guint action,
				  GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	gtkspell_check_all(GTK_TEXT(compose->text));
}


static void compose_destroy_cb(GtkWidget *widget, Compose *compose)
{
	compose_destroy(compose);
}

static void compose_cut_cb(Compose *compose)
{
	if (compose->focused_editable &&
	    GTK_WIDGET_HAS_FOCUS(compose->focused_editable))
		gtk_editable_cut_clipboard
			(GTK_EDITABLE(compose->focused_editable));
}

static void compose_copy_cb(Compose *compose)
{
	if (compose->focused_editable &&
	    GTK_WIDGET_HAS_FOCUS(compose->focused_editable))
		gtk_editable_copy_clipboard
			(GTK_EDITABLE(compose->focused_editable));
}

static void compose_paste_cb(Compose *compose)
{
	if (compose->focused_editable &&
	    GTK_WIDGET_HAS_FOCUS(compose->focused_editable))
		gtk_editable_paste_clipboard
			(GTK_EDITABLE(compose->focused_editable));
}

static void compose_allsel_cb(Compose *compose)
{
	if (compose->focused_editable &&
	    GTK_WIDGET_HAS_FOCUS(compose->focused_editable))
		gtk_editable_select_region
			(GTK_EDITABLE(compose->focused_editable), 0, -1);
}

static void compose_grab_focus_cb(GtkWidget *widget, Compose *compose)
{
	if (GTK_IS_EDITABLE(widget))
		compose->focused_editable = widget;
}

static void compose_changed_cb(GtkEditable *editable, Compose *compose)
{
	if (compose->modified == FALSE) {
		compose->modified = TRUE;
	}
}

static void compose_button_press_cb(GtkWidget *widget, GdkEventButton *event,
				    Compose *compose)
{
	gtk_text_set_point(GTK_TEXT(widget),
			   gtk_editable_get_position(GTK_EDITABLE(widget)));
}

#if 0
static void compose_key_press_cb(GtkWidget *widget, GdkEventKey *event,
				 Compose *compose)
{
	gtk_text_set_point(GTK_TEXT(widget),
			   gtk_editable_get_position(GTK_EDITABLE(widget)));
}
#endif

static void compose_toggle_bcc_cb(gpointer data, guint action,
				  GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	if (GTK_CHECK_MENU_ITEM(widget)->active) {
		gtk_widget_show(compose->bcc_hbox);
		gtk_widget_show(compose->bcc_entry);
		compose->use_bcc = TRUE;
	} else {
		gtk_widget_hide(compose->bcc_hbox);
		gtk_widget_hide(compose->bcc_entry);
		gtk_widget_queue_resize(compose->table_vbox);
		compose->use_bcc = FALSE;
	}

	if (addressbook_get_target_compose() == compose)
		addressbook_set_target_compose(compose);
}

static void compose_toggle_replyto_cb(gpointer data, guint action,
				      GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	if (GTK_CHECK_MENU_ITEM(widget)->active) {
		gtk_widget_show(compose->reply_hbox);
		gtk_widget_show(compose->reply_entry);
		compose->use_replyto = TRUE;
	} else {
		gtk_widget_hide(compose->reply_hbox);
		gtk_widget_hide(compose->reply_entry);
		gtk_widget_queue_resize(compose->table_vbox);
		compose->use_replyto = FALSE;
	}
}

static void compose_toggle_attach_cb(gpointer data, guint action,
				     GtkWidget *widget)
{
	Compose *compose = (Compose *)data;

	if (GTK_CHECK_MENU_ITEM(widget)->active) {
		gtk_widget_ref(compose->scrolledwin);

		gtk_container_remove(GTK_CONTAINER(compose->vbox2),
				     compose->scrolledwin);
		gtk_paned_add2(GTK_PANED(compose->paned),
			       compose->scrolledwin);
		gtk_box_pack_start(GTK_BOX(compose->vbox2), compose->paned,
				   TRUE, TRUE, 0);
		gtk_widget_show(compose->paned);

		gtk_widget_unref(compose->scrolledwin);
		gtk_widget_unref(compose->paned);

		compose->use_attach = TRUE;
	} else {
		gtk_widget_ref(compose->paned);
		gtk_widget_ref(compose->scrolledwin);

		gtk_container_remove(GTK_CONTAINER(compose->vbox2),
				     compose->paned);
		gtk_container_remove(GTK_CONTAINER(compose->paned),
				     compose->scrolledwin);
		gtk_box_pack_start(GTK_BOX(compose->vbox2),
				   compose->scrolledwin, TRUE, TRUE, 0);

		gtk_widget_unref(compose->scrolledwin);

		compose->use_attach = FALSE;
	}
}

static void compose_attach_drag_received_cb (GtkWidget		*widget,
					     GdkDragContext	*drag_context,
					     gint		 x,
					     gint		 y,
					     GtkSelectionData	*data,
					     guint		 info,
					     guint		 time,
					     gpointer		 user_data)
{
	Compose *compose = (Compose *)user_data;
	GList *list, *tmp;

	list = uri_list_extract_filenames((const gchar *)data->data);
	for (tmp = list; tmp != NULL; tmp = tmp->next)
		compose_attach_append(compose, (const gchar *)tmp->data,MIME_UNKNOWN);
	list_free_strings(list);
	g_list_free(list);
}

static void compose_insert_drag_received_cb (GtkWidget		*widget,
					     GdkDragContext	*drag_context,
					     gint		 x,
					     gint		 y,
					     GtkSelectionData	*data,
					     guint		 info,
					     guint		 time,
					     gpointer		 user_data)
{
	Compose *compose = (Compose *)user_data;
	GList *list, *tmp;

	list = uri_list_extract_filenames((const gchar *)data->data);
	for (tmp = list; tmp != NULL; tmp = tmp->next)
		compose_insert_file(compose, (const gchar *)tmp->data,FALSE);
	list_free_strings(list);
	g_list_free(list);
}

static void to_activated(GtkWidget *widget, Compose *compose)
{
	gtk_widget_grab_focus(GTK_COMBO(compose->to_entry)->entry); /* FIXME: better add selected address to to_entry->entry? */
}

static void ng_activated(GtkWidget *widget, Compose *compose)
{
	gtk_widget_grab_focus(GTK_COMBO(compose->ng_entry)->entry); /* FIXME: better add selected address to to_entry->entry? */
}

static void from_activated(GtkWidget *widget, Compose *compose)
{
/*	gtk_widget_grab_focus(compose->to_entry); */
	gtk_widget_grab_focus(GTK_COMBO(compose->from_entry)->entry);
}

static void subject_activated(GtkWidget *widget, Compose *compose)
{
	gtk_widget_grab_focus(compose->cc_entry);
}

static void cc_activated(GtkWidget *widget, Compose *compose)
{
/*	if (compose->use_bcc)
		gtk_widget_grab_focus(compose->bcc_entry);
	else if (compose->use_replyto)
		gtk_widget_grab_focus(compose->reply_entry);
	else
		gtk_widget_grab_focus(compose->text);   */
	gtk_widget_grab_focus(GTK_COMBO(compose->cc_entry)->entry);
}

static void bcc_activated(GtkWidget *widget, Compose *compose)
{
/*	if (compose->use_replyto)
		gtk_widget_grab_focus(compose->reply_entry);
	else
		gtk_widget_grab_focus(compose->text);*/
	gtk_widget_grab_focus(GTK_COMBO(compose->bcc_entry)->entry);
}

static void replyto_activated(GtkWidget *widget, Compose *compose)
{
	gtk_widget_grab_focus(compose->text);
}

static void compose_wrap_line(Compose *compose)
{
	GtkText *text = GTK_TEXT(compose->text);
	guint text_len;
	guint line_pos = 0, cur_pos = 0;
	gint line_len = 0, cur_len = 0;
	gint ch_len;
	gchar cbuf[MB_CUR_MAX];

	gtk_text_freeze(text);

	text_len = gtk_text_get_length(text);

	for (; cur_pos < text_len; cur_pos++) {
		if (text->use_wchar)
			ch_len = safe_wctomb
				(cbuf, (wchar_t)GTK_TEXT_INDEX(text, cur_pos));
		else {
			cbuf[0] = GTK_TEXT_INDEX(text, cur_pos);
			ch_len = 1;
		}

		if (ch_len == 1 && *cbuf == '\n') {
			line_pos = cur_pos + 1;
			line_len = cur_len = 0;
			continue;
		}

		if (ch_len < 0) {
			cbuf[0] = '\0';
			ch_len = 1;
		}

		if (ch_len == 1 && isspace(*cbuf)) {
			line_pos = cur_pos + 1;
			line_len = cur_len + ch_len;
		}

		if (cur_len + ch_len > prefs_common.linewrap_len &&
		    line_len > 0) {
			gint tlen;

			if (text->use_wchar)
				tlen = safe_wctomb(cbuf, (wchar_t)GTK_TEXT_INDEX(text, line_pos - 1));
			else {
				cbuf[0] = GTK_TEXT_INDEX(text, line_pos - 1);
				tlen = 1;
			}
			if (tlen == 1 && isspace(*cbuf)) {
				gtk_text_set_point(text, line_pos);
				gtk_text_backward_delete(text, 1);
				text_len--;
				cur_pos--;
				line_pos--;
				cur_len--;
				line_len--;
			}

			gtk_text_set_point(text, line_pos);
			gtk_text_insert(text, NULL, NULL, NULL, "\n", 1);
			text_len++;
			cur_pos++;
			line_pos++;
			cur_len = cur_len - line_len + ch_len;
			line_len = 0;
			continue;
		}

		if (ch_len > 1) {
			line_pos = cur_pos + 1;
			line_len = cur_len + ch_len;
		}
		cur_len += ch_len;
	}

	gtk_text_thaw(text);
}

