/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#ifdef LIBICONV_PLUG
#include <giconv.h>
#else
#include <iconv.h>
#endif

#include "intl.h"
#include "utils.h"
#include "prefs_common.h"

#include "cpconfigparse.h"

#include "conv_iconv.h"

#define RCBUFSIZE	1024

#define NCV	'\0'
#define MAX_ENCLEN	75
#define MAX_LINELEN	76

#define B64LEN(len)	((len) / 3 * 4 + ((len) % 3 ? 4 : 0))
iconv_t conv_iconv_open(gchar* charsetto,gchar* charsetfrm) {
	iconv_t ict;
	if (charsetfrm==NULL || charsetto==NULL) return (iconv_t)(-1);
//	if (!strcmp(charsetfrm,charsetto)) return (iconv_t)(-1);
	ict = iconv_open (charsetto,charsetfrm);
	debug_print(_("Open converting from %s to %s ocured.\n"),charsetfrm,charsetto);
	if (ict == (iconv_t)(-1)) {
		debug_print(_("Strange error while converting from %s to %s ocured.\n"),charsetfrm,charsetto);
	};
	return ict;
};

iconv_t conv_iconv_open_tolocal(gchar* charsetfrm) {
	iconv_t ict;
	if (charsetfrm==NULL) return (iconv_t)(-1);
	ict = iconv_open (curr_charset,charsetfrm);
	debug_print(_("Open converting from %s to local (%s) ocured.\n"),charsetfrm,curr_charset);
	if (ict == (iconv_t)(-1)) {
		debug_print(_("Strange error while converting from %s to local(%s) ocured.\n"),charsetfrm,curr_charset);
	};
	return ict;
};
void conv_iconv_close(iconv_t ict) {
	debug_print(_("Closing converting.\n"));
	if (ict != (iconv_t)(-1)) iconv_close(ict);
};

gint conv_iconv_convert(iconv_t ict,gchar *buf2, size_t sz2, gchar* buf, size_t sz) {
	size_t res,inlen,outlen;
        outlen=sz2;
        inlen=sz;

	if (ict == (iconv_t)(-1)) {
		debug_print(_("Strange error while converting.\n"));
		memcpy(buf2,buf,inlen);
		return -1;
	};
	while((ssize_t)inlen>0) {
                res=iconv (ict, (const char* * ) (&buf), (size_t*) &inlen, &buf2, &outlen);
 		if ((iconv_t)(res)==(iconv_t)(-1)) {
            		switch(errno) {
                    		case EILSEQ:
                            		*buf2++=*buf++;
                                        inlen--;
                                        outlen++;
                                        break;
                        	default:
                            		g_warning("Conversion from failed: %s\n",strerror(errno));
// 	                                memcpy(buf2,buf,inlen);
                                        return -1;
                                        break;
            		}
		}
	};
	return (int) outlen;
}

gint code_convert(gchar *buf2, size_t sz2, gchar* buf, size_t sz,  gchar* charsetfrm,gchar* charsetto) {
	iconv_t ict;
	gint outlen;
	debug_print("code_convert\n");
	ict = conv_iconv_open (charsetto,charsetfrm);
	outlen = conv_iconv_convert(ict, buf2, sz2, buf, sz);
	conv_iconv_close(ict);
	return outlen;
};
gint code_convert_to_current(gchar *buf2, size_t sz2, gchar* buf, size_t sz, gchar* charsetfrm) {
    debug_print("code_convert_tolocal\n");
    return code_convert(buf2,sz2,buf,sz,charsetfrm,curr_charset);
};

void conv_mb_alnum(gchar *str)
{
	static guchar char_tbl[] = {
		/* 0xa0 - 0xaf */
		NCV, ' ', NCV, NCV, ',', '.', NCV, ':',
		';', '?', '!', NCV, NCV, NCV, NCV, NCV,
		/* 0xb0 - 0xbf */
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV,
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV,
		/* 0xc0 - 0xcf */
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV,
		NCV, NCV, '(', ')', NCV, NCV, '[', ']',
		/* 0xd0 - 0xdf */
		'{', '}', NCV, NCV, NCV, NCV, NCV, NCV,
		NCV, NCV, NCV, NCV, '+', '-', NCV, NCV,
		/* 0xe0 - 0xef */
		NCV, '=', NCV, '<', '>', NCV, NCV, NCV,
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV
	};

	register guchar *p = str;
	register gint len;

	len = strlen(str);

	while (len > 1) {
		if (*p == 0xa3) {
			register guchar ch = *(p + 1);

			if (ch >= 0xb0 && ch <= 0xfa) {
				/* [a-zA-Z] */
				*p = ch & 0x7f;
				p++;
				len--;
				memmove(p, p + 1, len);
				len--;
			} else  {
				p += 2;
				len -= 2;
			}
		} else if (*p == 0xa1) {
			register guchar ch = *(p + 1);

			if (ch >= 0xa0 && ch <= 0xef &&
			    NCV != char_tbl[ch - 0xa0]) {
				*p = char_tbl[ch - 0xa0];
				p++;
				len--;
				memmove(p, p + 1, len);
				len--;
			} else {
				p += 2;
				len -= 2;
			}
		} else {
			p++;
			len--;
		}
	}
}

void conv_unmime_header(gchar *outbuf, gint outlen, const gchar *str,
			const gchar *charset)
{
	gchar *buf, *cset, *icset;
	gint a;

	Xstrdup_a(buf, str, {return;});
	cset=(gchar*)UnMimeHeader(buf);
	if (cset!=NULL) {
	    icset=get_iconv_charset(cset);
	    free(cset);
	    a=code_convert_to_current(outbuf, outlen, buf,strlen(buf)+1, icset);
	    if (a>0) outbuf[a]=0;
	} else {
	  if(charset) {
	      icset=get_iconv_charset((gchar*) charset);
	      a=code_convert_to_current(outbuf, outlen, buf,strlen(buf)+1, icset);
	      if (a>0) outbuf[a]=0;
	  } else {
	      strncpy2(outbuf, buf, strlen(buf)+1);
	  }
	}
}

void conv_encode_header_misc(gchar *dest, gint len, const gchar *src,
                        gint header_len) {
    gchar *buf;
    const gchar *out_codeset;
    const gchar *out_iconv_codeset;
    gint hdlen;
    gint baselen=(2+strlen(curr_charset)+3+B64LEN( strlen(src))+2);
    gint i;
    out_iconv_codeset = get_iconv_charset(prefs_common.outgoing_charset);
    if (out_iconv_codeset==NULL) {
      out_codeset=g_strdup(curr_charset);
      out_iconv_codeset=g_strdup(curr_charset);
    } else {
      out_codeset=prefs_common.outgoing_charset;
    }

    hdlen=g_snprintf(dest,baselen,"=?%s?B?",out_codeset);
    buf = g_strdup(src);

    code_convert( buf, strlen(buf), (gchar*)src, strlen(src), (gchar*)curr_charset, (gchar*)out_iconv_codeset);
    to64frombits(dest+hdlen, buf, strlen(buf));
    strcat(dest,"?=");
    g_free(buf);
    return;
};


void conv_encode_header_1(gchar *dest, gint len, const gchar *src,
                        gint header_len) {
    gchar *buf;
    gchar *h;
    gchar *header;
    const gchar *out_codeset;
    const gchar *out_iconv_codeset;
    gint hdlen;
    gboolean collect=TRUE;
    gint baselen=(2+strlen(curr_charset)+3+B64LEN( strlen(src))+2);
    gint i;
    out_iconv_codeset = get_iconv_charset(prefs_common.outgoing_charset);
    if (out_iconv_codeset==NULL) {
      out_codeset=g_strdup(curr_charset);
      out_iconv_codeset=g_strdup(curr_charset);
    } else {
      out_codeset=prefs_common.outgoing_charset;
    }

    header=g_strdup_printf("=?%s?B?",out_codeset);
    hdlen=strlen(header);

    buf = g_strdup(src);
    code_convert( buf, strlen(buf), (gchar*)src, strlen(src), (gchar*)curr_charset, (gchar*)out_iconv_codeset);
    h=buf;
    dest[0]=0;
    for (i=0;i<strlen(buf);i++) {
        if (buf[i]=='<' && collect) {
                        strcat(dest,header);
                        to64frombits(dest+strlen(dest),h,i-(h-buf));
                        strcat((char*)dest,"?=");
                        h=buf+i;
                        collect=FALSE;
                        continue;
        };
        if (buf[i]=='>' && !collect) {
                        strncat(dest,h,i-(h-buf)+1);
                        h=buf+i;
                        collect=TRUE;
                        continue;
        };
    };
        if (collect) {
                        strcat(dest,header);
                        to64frombits(dest+strlen(dest),h,i-(h-buf));
                        strcat(dest,"?=");
                        return;
        };
        if (!collect) {
                        strncat(dest,h,(size_t)i-(h-buf));
                        return;
        };

    return;
};

void conv_encode_header(gchar *dest, gint len, const gchar *src,
                        gint header_len) {
    gchar *buf;
    gchar *h;
    gchar *header;
    const gchar *out_codeset;
    const gchar *out_iconv_codeset;
    gint hdlen;
    gboolean collect=TRUE;
    gint baselen=(2+strlen(curr_charset)+3+B64LEN( strlen(src))+2);
    gint i;
    out_iconv_codeset = get_iconv_charset(prefs_common.outgoing_charset);
    if (out_iconv_codeset==NULL) {
      out_codeset=g_strdup(curr_charset);
      out_iconv_codeset=g_strdup(curr_charset);
    } else {
      out_codeset=prefs_common.outgoing_charset;
    }

    header=g_strdup_printf("=?%s?B?",out_codeset);
    hdlen=strlen(header);

    buf = g_strdup(src);
    code_convert( buf, strlen(buf), (gchar*)src, strlen(src), (gchar*)curr_charset, (gchar*)out_iconv_codeset);
    h=buf;
    dest[0]=0;
    for (i=0;i<strlen(buf);i++) {
        if (isascii(buf[i]) && collect) {
                        strcat(dest,header);
                        to64frombits(dest+strlen(dest),h,i-(h-buf));
                        strcat((char*)dest,"?=");
                        h=buf+i;
                        collect=FALSE;
                        continue;
        };
        if (!isascii(buf[i]) && !collect) {
                        strncat(dest,h,i-(h-buf)+1);
                        h=buf+i;
                        collect=TRUE;
                        continue;
        };
    };
        if (collect) {
                        strcat(dest,header);
                        to64frombits(dest+strlen(dest),h,i-(h-buf));
                        strcat(dest,"?=");
                        return;
        };
        if (!collect) {
                        strncat(dest,h,(size_t)i-(h-buf));
                        return;
        };

    return;
};

