/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#ifdef LIBICONV_PLUG
#include <giconv.h>
#else
#include <iconv.h>
#endif

#include "intl.h"
#include "utils.h"
#include "prefs_common.h"

#include "cpconfigparse.h"

#define RCBUFSIZE	1024

#define NCV	'\0'
#define MAX_ENCLEN	75
#define MAX_LINELEN	76

#define B64LEN(len)	((len) / 3 * 4 + ((len) % 3 ? 4 : 0))



CodePages       *known_code_pages=NULL;

gchar		*curr_locale=NULL;
gchar		*curr_charset=NULL;

gint cp_parse(char *cp_file_name) {
	FILE    *fp;
	gchar	*rcpath;
	gchar	buf[RCBUFSIZE];
	gchar	**split_str;
	gint	i;
	gint	ret_val = 1;

	g_return_val_if_fail(cp_file_name != NULL, 0);

	cp_clear();

	debug_print(_("Reading code pages configuration...\n"));

	/* open file */
	rcpath = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, cp_file_name, NULL);
	if ((fp = fopen(rcpath, "r")) != NULL) {
		while (fgets(buf, sizeof(buf), fp) != NULL) {
			strretchomp(buf);
			split_str = g_strsplit(buf, ":", 3);
			cp_append(split_str[0], split_str[1], split_str[2]);
			g_strfreev(split_str);
		}
		fclose(fp) != 0;
	} else {
		if (ENOENT != errno) FILE_OP_ERROR(rcpath, "fopen");
		ret_val = 0;
	}

	g_free(rcpath);

        return ret_val;
}

/* saves charsets to file */
gint cp_save(char *cp_file_name) {
	FILE    *fp;
	gchar   *rcpath;
	GString *charset_list;
	gchar	*iconv_cp;
	gchar	*cp_name;
	gint	cp_num = 0;
	gint	ret_val = 1;

	g_return_val_if_fail(cp_file_name != NULL, 0);

	debug_print(_("Saving code pages configuration...\n"));

	/* open file */
	rcpath = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, cp_file_name, NULL);
	if ((fp = fopen(rcpath, "w")) == NULL) {
		if (ENOENT != errno) FILE_OP_ERROR(rcpath, "fopen");
		g_free(rcpath);
		return 0;
	};

	/* save charsets */
	charset_list = g_string_sized_new(32);
	while ((cp_num = cp_get(cp_num, charset_list, &iconv_cp, &cp_name)) != -1) {
		g_string_sprintfa(charset_list, ":%s:%s\n", iconv_cp, cp_name);
		if (fputs(charset_list->str, fp) == EOF) {
			FILE_OP_ERROR(rcpath, "fputs");
			ret_val = 0;
			break;
		}
	}

	g_string_free(charset_list, TRUE);
		
	if ((fclose(fp)) != 0) {
		FILE_OP_ERROR(rcpath, "fclose");
		ret_val = 0;
	};
	g_free(rcpath);
	return ret_val;
}

gint cp_create(char *cp_file_name) {
	FILE            *fp;
	gchar           *rcpath;

	g_return_val_if_fail(cp_file_name != NULL, 0);

	debug_print(_("Create code pages configuration...\n"));

	rcpath = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, cp_file_name, NULL);
	unlink(rcpath);
	if ((fp = fopen(rcpath, "w")) != NULL) {
		if (ENOENT != errno) FILE_OP_ERROR(rcpath, "fopen");
		fprintf(fp,"US-ASCII,ASCII,EN,US,GB-ASCII:ASCII:");
		fprintf(fp,_("7-bit ascii\n"));
		fprintf(fp,"KOI8-R,KOI8,UNIX-KOI8-RUSSIAN:KOI8-R:");
		fprintf(fp,_("Cyrillic KOI8-R\n"));
		fprintf(fp,"CP1251,WINDOWS-1251,CP-1251:CP1251:");
		fprintf(fp,_("Cyrillic CP1251\n"));
		fprintf(fp,"ISO-8859-5,ISO_8859-5,ISO-8859_5,ISO_8859_5,8859_5,8859-5:ISO_8859-5:");
		fprintf(fp,_("Cyrillic ISO_8859-5\n"));
		fprintf(fp,"CP866,IBM-866,IBM866,CP-866,DOS,CSIBM866,866:CP866:");
		fprintf(fp,_("Cyrillic IBM-866\n"));
		fprintf(fp,"UTF-8,UTF8,UNICODE,UNICODE-8:UTF-8:");
		fprintf(fp,_("Unicode UTF-8"));
/*
		fprintf(fp,"::");
		fprintf(fp,_("\n"));
*/
		fclose(fp);
		g_free(rcpath);
		return 0;
	}
	g_free(rcpath);
}

gint	cp_clear(void)
{
        gint i;

        if(known_code_pages) {
                for (i = 0; known_code_pages[i].charset; i++) {
                        g_free(known_code_pages[i].charset);
                        g_free(known_code_pages[i].iconv_charset);
			g_free(known_code_pages[i].charset_name);
                }
                g_free(known_code_pages);
		known_code_pages = NULL;
        }
}

/* store charset, iconv charset and charset name in gchar*
   charset_list is comma separated charsets
   search starts from cp_num
   Returns: 
   the new cp_num is returned or
   -1 if no more charsets available (contents of gchar* are undefined in this case)
*/
gint cp_get (gint cp_num, GString *charset_list, gchar **iconv_charset, gchar **charset_name) {

	if (!known_code_pages || known_code_pages[cp_num].charset == NULL)
		return -1;

	g_string_assign(charset_list, known_code_pages[cp_num].charset);
	*iconv_charset = known_code_pages[cp_num].iconv_charset;
	*charset_name = known_code_pages[cp_num].charset_name;
	cp_num++;

	while (known_code_pages[cp_num].charset != NULL &&
	       g_strcasecmp(known_code_pages[cp_num].iconv_charset, *iconv_charset) == 0) {
		g_string_sprintfa(charset_list, ",%s", known_code_pages[cp_num].charset);
		cp_num++;
	};

	return cp_num;
}

gchar *get_iconv_charset(gchar* mail_charset) {
        return get_iconv_charset_from(mail_charset, known_code_pages);
}

gchar *get_iconv_charset_from(gchar* mail_charset, CodePages *cpages) {
        gint i=0;

        if((!known_code_pages) || (!mail_charset)) {
                return NULL;
        }

        while((cpages[i].charset) && strcasecmp(cpages[i].charset,mail_charset)) {
                i++;
        }
        if(cpages[i].charset) {
                return g_strdup(cpages[i].iconv_charset);
        }
        return NULL;
};

gint insert_cp_in_menu(GtkItemFactory *ifactory, GtkItemFactoryCallback callback,
                       gchar *path, gpointer callback_data) {
        GtkItemFactoryEntry entry;
	GString *charset_list;
	gchar	*iconv_cp;
	gchar	*cp_name;
        guint	cp_num = 0;

	g_return_val_if_fail(ifactory != NULL, 0);

        debug_print(_("Starting menu altering: %s\n"),path);
        entry.accelerator = NULL;
        entry.callback = callback;
        entry.item_type = "/Message/Code set/Auto detect";
	charset_list = g_string_sized_new(32);

	while ((cp_num = cp_get(cp_num, charset_list, &iconv_cp, &cp_name)) != -1) {
                entry.path = g_strconcat(path, cp_name, NULL);
		entry.callback_action = cp_num-1;

		debug_print(_("Inserting in menu: %s with callback: %u (%s)\n"),
			    entry.path, cp_num, iconv_cp);
		gtk_item_factory_create_item(ifactory, &entry, callback_data, 1);

                g_free(entry.path);
	}

	g_string_free(charset_list, TRUE);

        return 0;
};

/* append new charset to CodePages 
   mail_charsets is comma separated list of mail charsets
*/
void cp_append(gchar *mail_charsets, gchar *iconv_charset, gchar *charset_name)
{
	gchar **split_charset;
	gint cp_num;
	gint i;

	g_return_if_fail(mail_charsets != NULL);
	g_return_if_fail(iconv_charset != NULL);
	g_return_if_fail(charset_name != NULL);

	/* if empty create new */
	if (!known_code_pages) {
		if ((known_code_pages = g_malloc(sizeof(CodePages))) == NULL) {
			log_error("Memory allocation error");
			return;
		};
		known_code_pages[0].charset = NULL;
	}

	/* split charsets */
	split_charset = g_strsplit(mail_charsets, ",", 0);
	for (i = 0; split_charset[i] != NULL; i++) {
		/* check for duplicate charsets */
		cp_num = 0;
		while(known_code_pages[cp_num].charset != NULL &&
		      strcasecmp(known_code_pages[cp_num].charset, split_charset[i]) != 0) {
			cp_num++;
		}
		if (known_code_pages[cp_num].charset != NULL) {
			log_warning(_("Ignoring duplicate code page %s (%s)"), 
				    split_charset[i], charset_name);
			continue;
		}

		/* allocate memory and save new values */
		debug_print(_("Adding code page: %s iconv code page: %s name: %s\n"),
			    split_charset[i], iconv_charset, charset_name);

		if ((known_code_pages = g_realloc(known_code_pages,
						  sizeof(CodePages)*(cp_num+2))) == NULL) {
			log_error("Memory allocation error");
			return;
		}
		known_code_pages[cp_num].charset = g_strdup(split_charset[i]);
		known_code_pages[cp_num].iconv_charset = g_strdup(iconv_charset);
		known_code_pages[cp_num].charset_name = g_strdup(charset_name);

		known_code_pages[cp_num+1].charset = NULL;
	}

	g_strfreev(split_charset);
}
