/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkfilesel.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "filesel.h"
#include "manage_window.h"
#include "gtkutils.h"

static GtkWidget *filesel;
static gboolean filesel_ack;
GList *selected_files=NULL;

static void filesel_create(const gchar *title, gboolean multiple);
static void filesel_ok_cb(GtkWidget *widget, gpointer data);
static void filesel_cancel_cb(GtkWidget *widget, gpointer data);
static void key_pressed(GtkWidget *widget, GdkEventKey *event, gpointer data);

void selection_made(GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data)
{
        gchar *text;

        gtk_clist_get_text(GTK_CLIST(clist), row, 0, &text);
        if(!g_list_find(selected_files,text)) {
                selected_files=g_list_append(selected_files,text);
        }

        return;
}

void selection_clean(GtkWidget *clist, gint row, gint column, GdkEventButton *event, gpointer data)
{
        gchar *text;

        gtk_clist_get_text(GTK_CLIST(clist), row, 0, &text);
        selected_files=g_list_remove(selected_files,text);

        return;
}

GList *filesel_select_file(const gchar *title, const gchar *file, gboolean multiple)
{
	static gchar *cwd = NULL;
        static gchar *filename = NULL;
        GList *filelist=NULL;
        gint i;

        if(selected_files) {
                g_list_free(selected_files);
                selected_files=NULL;
        }

	filesel_create(title, multiple);

	manage_window_set_transient(GTK_WINDOW(filesel));

	if (filename) {
		g_free(filename);
		filename = NULL;
	}

	if (!cwd)
		cwd = g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, NULL);

	gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), cwd);

	if (file)
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel),
						file);

	gtk_widget_show(filesel);

	gtk_main();

	if (filesel_ack) {
		gchar *str;
		gchar *dir=NULL;

		str = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
		if (str && str[0] != '\0') {
                        filename=g_strdup(str);
			dir = g_dirname(str);
			g_free(cwd);
			cwd = g_strconcat(dir, G_DIR_SEPARATOR_S, NULL);
			g_free(dir);
		}
                if(!multiple) {
		        filelist=g_list_append(filelist, g_strdup(str));
                } else {
                        filelist=g_list_append(filelist,cwd);
                        if(selected_files) {
                                for(i=0;i<g_list_length(selected_files);i++) {
                                        filelist=g_list_append(filelist,g_strdup(g_list_nth_data(selected_files,i)));
                                }
                        }
                }
	}

	manage_window_focus_out(filesel, NULL, NULL);
	gtk_widget_destroy(filesel);
	GTK_EVENTS_FLUSH();

	return filelist;
}

static void filesel_create(const gchar *title, gboolean multiple)
{
	filesel = gtk_file_selection_new(title);
        if(multiple) {
                gtk_clist_set_selection_mode(
                        GTK_CLIST(GTK_FILE_SELECTION(filesel)->file_list),
                        GTK_SELECTION_EXTENDED);
                gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->file_list),
                        "select_row",
                        GTK_SIGNAL_FUNC(selection_made),
                        NULL);
                gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->file_list),
                        "unselect_row",
                        GTK_SIGNAL_FUNC(selection_clean),
                        NULL);
        }

	gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button),
			   "clicked", GTK_SIGNAL_FUNC(filesel_ok_cb),
			   NULL);
	gtk_signal_connect
		(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button),
		 "clicked", GTK_SIGNAL_FUNC(filesel_cancel_cb),
		 NULL);
	gtk_signal_connect(GTK_OBJECT(filesel), "delete_event",
			   GTK_SIGNAL_FUNC(filesel_cancel_cb),
			   NULL);
	gtk_signal_connect(GTK_OBJECT(filesel), "key_press_event",
			   GTK_SIGNAL_FUNC(key_pressed), NULL);
	gtk_signal_connect(GTK_OBJECT(filesel), "focus_in_event",
			   GTK_SIGNAL_FUNC(manage_window_focus_in), NULL);
	gtk_signal_connect(GTK_OBJECT(filesel), "focus_out_event",
			   GTK_SIGNAL_FUNC(manage_window_focus_out), NULL);

	gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
}

static void filesel_ok_cb(GtkWidget *widget, gpointer data)
{
	filesel_ack = TRUE;
	gtk_main_quit();
}

static void filesel_cancel_cb(GtkWidget *widget, gpointer data)
{
	filesel_ack = FALSE;
	gtk_main_quit();
}

static void key_pressed(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	if (event && event->keyval == GDK_Escape)
		filesel_cancel_cb(NULL, NULL);
}
