/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __FOLDERVIEW_H__
#define __FOLDERVIEW_H__

#include <glib.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkctree.h>
#include <time.h>

typedef struct _FolderInfo	FolderInfo;
typedef struct _FolderView	FolderView;

#include "mainwindow.h"
#include "summaryview.h"

typedef enum
{
	F_TOPMAILFOLDER,
	F_MHFOLDER,
	F_TOPNEWSFOLDER,
	F_NEWSSERVER,
	F_NEWSGROUP
} FolderType;
typedef enum
{
	F_MH_NORMAL,
	F_MH_INBOX,
	F_MH_OUTBOX,
	F_MH_DRAFT,
	F_MH_QUEUE,
	F_MH_TRASH
} MHFolderType;

struct _FolderInfo
{
	FolderType type;
	MHFolderType mh_type;
	gchar *path;
	time_t mtime;

	gint new;
	gint unread;
	gint total;
};

struct _FolderView
{
	GtkWidget *scrolledwin;
	GtkWidget *ctree;
	GtkWidget *mail_popup;
	GtkWidget *news_popup;

	GtkItemFactory *mail_factory;
	GtkItemFactory *news_factory;

	GtkCTreeNode *mailbox;
	GtkCTreeNode *news;

	GtkCTreeNode *selected;
	GtkCTreeNode *opened;

	gboolean open_folder;

	GdkColor color_new;
	GdkColor color_normal;

	MainWindow   *mainwin;
	SummaryView  *summaryview;
};

FolderView *folderview_create		(void);
void folderview_init			(FolderView	*folderview);
void folderview_set			(FolderView	*folderview);
void folderview_select			(FolderView	*folderview,
					 const gchar	*folder);
void folderview_unselect		(FolderView	*folderview);
void folderview_select_next_unread	(FolderView	*folderview);
void folderview_write_cache		(FolderView	*folderview);
void folderview_update_msg_num		(FolderView	*folderview,
					 GtkCTreeNode	*row,
					 gint		 new,
					 gint		 unread,
					 gint		 total);
void folderview_scan_folder		(FolderView	*folderview,
					 const gchar	*folder,
					 gboolean	 update_summary);
void folderview_scan_folder_a		(const gchar	*folder,
					 gboolean	 update_summary);
void folderview_scan_folder_foreach_a	(GHashTable	*folder_table);

void folderview_update_all		(FolderView	*folderview);

gint folderview_compare_path		(gconstpointer	 a,
					 gconstpointer	 b);

#endif /* __FOLDERVIEW_H__ */
