/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gtk/gtk.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkstyle.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtktext.h>
#include <gtk/gtkpixmap.h>
#include <gtk/gtkcontainer.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#if HAVE_LIBCOMPFACE
#  include <compface.h>
#endif

#include "intl.h"
#include "main.h"
#include "headerview.h"
#include "prefs_common.h"
#include "gtkutils.h"
#include "utils.h"
#include "addressbook.h"

#include "pixmaps/addresbook_add.xpm"

static GdkFont *normalfont;
static GdkFont *boldfont;
static GdkPixmap *addresbookaddxpm;
static GdkBitmap *addresbookaddxpmmask;

gint addresbook_add_sender_cb(GtkWidget *widget,HeaderView *headerview);
#if HAVE_LIBCOMPFACE
#define XPM_XFACE_HEIGHT	(HEIGHT + 3)  /* 3 = 1 header + 2 colors */

static gchar *xpm_xface[XPM_XFACE_HEIGHT];

static void headerview_show_xface	(HeaderView	*headerview,
					 MsgInfo	*msginfo);
static gint create_xpm_from_xface	(gchar		*xpm[],
					 const gchar	*xface);
#endif

HeaderView *headerview_create(void)
{
	GtkWidget *hbox;
	GtkWidget *text;
        HeaderView *headerview;

	debug_print(_("Creating header view...\n"));
	headerview = g_new0(HeaderView, 1);

	hbox = gtk_hbox_new(FALSE, 0);

	text = gtk_text_new(NULL, NULL);
	GTK_WIDGET_UNSET_FLAGS(text, GTK_CAN_FOCUS);
	gtk_text_set_word_wrap(GTK_TEXT(text), FALSE);
	gtk_text_set_line_wrap(GTK_TEXT(text), FALSE);

	gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 0);

	headerview->hbox = hbox;
	headerview->text = text;
	headerview->image = NULL;
        headerview->from = NULL;
	gtk_widget_show_all(hbox);

	return headerview;
}

void headerview_init(HeaderView *headerview)
{
	GtkStyle *style;
        GtkWidget *pixmap;
        GtkTooltips *tooltips;
	//gint height;

	//if (!normalfont)
	//	normalfont = gdk_fontset_load(NORMAL_FONT);
	if (!boldfont)
		boldfont = gdk_fontset_load(BOLD_FONT);

	style = gtk_style_copy(gtk_widget_get_style(headerview->text));
	//style = gtk_widget_get_style(headerview->text);
	style->base[GTK_STATE_NORMAL] = style->bg[GTK_STATE_NORMAL];
	//style->font = normalfont;
	gtk_widget_set_style(headerview->text, style);

	//height = gtkut_get_font_height(style->font);
	gtk_widget_set_usize(headerview->text, -1, DEFAULT_HEADERVIEW_HEIGHT);
	gtk_widget_queue_resize(headerview->hbox);

        headerview->button=gtk_button_new();
        gtk_box_pack_start(GTK_BOX(headerview->hbox), headerview->button, FALSE, FALSE, 4);
        gtk_box_reorder_child(GTK_BOX(headerview->hbox), headerview->button, 0);
        gtk_button_set_relief(GTK_BUTTON(headerview->button), GTK_RELIEF_NONE);
        GTK_WIDGET_UNSET_FLAGS (headerview->button, GTK_CAN_FOCUS);
	gtk_widget_hide(headerview->button);
        tooltips=gtk_tooltips_new();
        gtk_tooltips_set_tip(tooltips,headerview->button,_("Add Sender to Address book"),NULL);

        PIXMAP_CREATE(headerview->button, addresbookaddxpm, addresbookaddxpmmask, addresbook_add_xpm);
        pixmap = gtk_pixmap_new(addresbookaddxpm, addresbookaddxpmmask);
        gtk_container_add (GTK_CONTAINER (headerview->button), pixmap);
	gtk_widget_show(pixmap);
        gtk_signal_connect (GTK_OBJECT(headerview->button), "clicked", GTK_SIGNAL_FUNC (addresbook_add_sender_cb), headerview);

#if HAVE_LIBCOMPFACE
	{
		gint i;

		for (i = 0; i < XPM_XFACE_HEIGHT; i++) {
			xpm_xface[i] = g_malloc(WIDTH + 1);
			*xpm_xface[i] = '\0';
		}
	}
#endif

#if 0
	_("From: ");
	_("To: ");
	_("Newsgroups: ");
	_("Subject: ");
#endif
}

#define TR(str)	(prefs_common.trans_hdr ? gettext(str) : str)

void headerview_show(HeaderView *headerview, MsgInfo *msginfo)
{
	GtkText *text = GTK_TEXT(headerview->text);

	headerview_clear(headerview);
	gtk_widget_show(headerview->button);
	gtk_text_freeze(text);

	gtk_text_insert(text, boldfont  , NULL, NULL, TR("From: "), -1);
        if (msginfo->from) {
            gtk_text_insert(text, normalfont, NULL, NULL, msginfo->from, -1);
            headerview->from = g_strdup(msginfo->from);
        } else {
            gtk_text_insert(text, normalfont, NULL, NULL, _("(No From)"), -1);
        }

        if (msginfo->to) {
		gtk_text_insert(text, NULL      , NULL, NULL, "  " , 2);
		gtk_text_insert(text, boldfont  , NULL, NULL, TR("To: ") , -1);
		gtk_text_insert(text, normalfont, NULL, NULL, msginfo->to , -1);
	}
	if (msginfo->newsgroups) {
		gtk_text_insert(text, NULL, NULL, NULL, "  ", 2);
		gtk_text_insert(text, boldfont, NULL, NULL, TR("Newsgroups: "), -1);
		gtk_text_insert(text, normalfont, NULL, NULL,
				msginfo->newsgroups, -1);
	}
	gtk_text_insert(text, NULL      , NULL, NULL, "\n", 1);

	gtk_text_insert(text, boldfont  , NULL, NULL, TR("Subject: "), -1);
	gtk_text_insert(text, normalfont, NULL, NULL,
			msginfo->subject ? msginfo->subject :
			_("(No Subject)"), -1);
	gtk_text_thaw(text);

#if HAVE_LIBCOMPFACE
	headerview_show_xface(headerview, msginfo);
#endif
}

#if HAVE_LIBCOMPFACE
static void headerview_show_xface(HeaderView *headerview, MsgInfo *msginfo)
{
	gchar xface[2048];
	GdkPixmap *pixmap;
	GdkBitmap *mask;
	GtkWidget *hbox = headerview->hbox;

	if (!msginfo->xface) {
		if (headerview->image &&
		    GTK_WIDGET_VISIBLE(headerview->image)) {
			gtk_widget_hide(headerview->image);
                        gtk_widget_queue_resize(hbox);
                }
                return;
	}

	strncpy(xface, msginfo->xface, sizeof(xface));

	if (uncompface(xface) < 0) {
		g_warning("uncompface failed\n");
		if (headerview->image)
			gtk_widget_hide(headerview->image);
		return;
	}

	create_xpm_from_xface(xpm_xface, xface);

	pixmap = gdk_pixmap_create_from_xpm_d
		(hbox->window, &mask, &hbox->style->white, xpm_xface);

	if (!headerview->image) {
		GtkWidget *image;

		image = gtk_pixmap_new(pixmap, mask);
		gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
		gtk_widget_show(image);
		headerview->image = image;
	} else {
		gtk_pixmap_set(GTK_PIXMAP(headerview->image), pixmap, mask);
		gtk_widget_show(headerview->image);
	}

	gdk_pixmap_unref(pixmap);
}
#endif

void headerview_clear(HeaderView *headerview)
{
	GtkText *text = GTK_TEXT(headerview->text);

	gtk_widget_hide(headerview->button);
	gtk_text_freeze(text);
	gtk_text_backward_delete(text, gtk_text_get_length(text));
	gtk_text_thaw(text);

	if (headerview->image && GTK_WIDGET_VISIBLE(headerview->image)) {
		gtk_widget_hide(headerview->image);
                gtk_widget_queue_resize(headerview->hbox);
        }
        g_free(headerview->from);
        headerview->from = NULL;
}

#if HAVE_LIBCOMPFACE
static gint create_xpm_from_xface(gchar *xpm[], const gchar *xface)
{
	static gchar *bit_pattern[] = {
		"....",
		"...#",
		"..#.",
		"..##",
		".#..",
		".#.#",
		".##.",
		".###",
		"#...",
		"#..#",
		"#.#.",
		"#.##",
		"##..",
		"##.#",
		"###.",
		"####"
	};

	static gchar *xface_header = "48 48 2 1";
	static gchar *xface_black  = "# c #000000";
	static gchar *xface_white  = ". c #ffffff";

	gint i, line = 0;
	const guchar *p;
	gchar buf[WIDTH * 4 + 1];  /* 4 = strlen("0x0000") */

	p = xface;

	strcpy(xpm[line++], xface_header);
	strcpy(xpm[line++], xface_black);
	strcpy(xpm[line++], xface_white);

	for (i = 0; i < HEIGHT; i++) {
		gint col;

		buf[0] = '\0';

		for (col = 0; col < 3; col++) {
			gint figure;

			p += 2;  /* skip '0x' */

			for (figure = 0; figure < 4; figure++) {
				gint n = 0;

				if ('0' <= *p && *p <= '9') {
					n = *p - '0';
				} else if ('a' <= *p && *p <= 'f') {
					n = *p - 'a' + 10;
				} else if ('A' <= *p && *p <= 'F') {
					n = *p - 'A' + 10;
				}

				strcat(buf, bit_pattern[n]);
				p++;  /* skip ',' */
			}

			p++;  /* skip '\n' */
		}

		strcpy(xpm[line++], buf);
		p++;
	}

	return 0;
}
#endif

gint addresbook_add_sender_cb(GtkWidget *widget,HeaderView *headerview)
{
    if(headerview->from)
        addresbook_add_sender(headerview->from);
}
