/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "intl.h"
#include "mainwindow.h"
#include "prefs_common.h"
#include "procmime.h"
#include "htmlview.h"
#include "utils.h"
#include <glib.h>
#include <textview.h>
/* INCLUDE PART		*/

#ifdef USE_MOZILLA
#include <mozilla/gtkmozembed.h>
#endif /* USE_MOZILLA */

#ifdef USE_CSCHTML
#include <cschtml/cschtml.h>
#endif /* USE_CSCHTML */

/* DEPENDENS PART	*/
/* DECLARATIONS		*/
#ifdef USE_MOZILLA
gboolean mozilla_preference_set(const char *preference_name, const char *new_value);
void htmlview_set_preferences (GtkMozEmbed *embed);
#endif /* USE_MOZILLA */

/* DEPENDENS PART	*/
/* FUNCTIONS		*/
#ifdef USE_MOZILLA
void htmlview_brsr_destroy (GtkMozEmbed *embed, HtmlView *htmlview)
{
	gtk_widget_unref(GTK_WIDGET(htmlview->html));
	gtk_widget_destroy(GTK_WIDGET(htmlview->html));
	htmlview->html=NULL;
}

void htmlview_set_preferences (GtkMozEmbed *embed)
{
	mozilla_preference_set ("intl.charset.detector", "ruprob");
}

#endif /* USE_MOZILLA */

#ifdef USE_CSCHTML
static void
htmlview_on_link_clicked (CscHTML *html, const gchar *url, gpointer data)
{
	textview_open_uri(url);
}
#endif

/* MAIN INTERFACE	*/

HtmlView *htmlview_create(void)
{
	HtmlView *htmlview;
#ifdef USE_MOZILLA
	GtkMozEmbed *mozwin;
	gchar *mozpat;
	htmlview = g_new0(HtmlView, 1);
	if((mozpat=g_getenv("MOZILLA_FIVE_HOME"))!=NULL) {
			gtk_moz_embed_set_comp_path(mozpat);
			g_free(mozpat);
	}
	mozwin=(GtkMozEmbed*)gtk_moz_embed_new();
	htmlview_set_preferences(mozwin);
	gtk_widget_ref(GTK_WIDGET(mozwin));
	htmlview->html=mozwin;
#endif /* USE_MOZILLA */

#ifdef USE_CSCHTML
	GtkWidget *scrolledwin;
	GtkWidget *html_widget;
	debug_print(_("Creating html view...\n"));
	htmlview = g_new0(HtmlView, 1);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER,
				       GTK_POLICY_ALWAYS);
	gtk_widget_set_usize(scrolledwin, prefs_common.mainview_width, -1);

	html_widget = csc_html_new();
	gtk_container_add (GTK_CONTAINER (scrolledwin), html_widget);
	gtk_widget_show_all(scrolledwin);

	htmlview->scrolledwin  = scrolledwin;
	htmlview->html_widget  = html_widget;
#endif /* USE_CSCHTML */
	return htmlview;
}

void htmlview_init(HtmlView *htmlview)
{
#ifdef USE_MOZILLA
#endif /* USE_MOZILLA */
}


gint htmlview_show_html(HtmlView *htmlview, MimeInfo *mimeinfo,
			  const gchar *file)
{
#ifdef USE_MOZILLA
	const char urlstart[]="file://";
	gchar *namebuf;
    if (htmlview->messageview->mainwin)
	    main_window_cursor_wait(htmlview->messageview->mainwin);
	namebuf=g_strconcat(urlstart,file,NULL);
	gtk_moz_embed_load_url(GTK_MOZ_EMBED(htmlview->html),namebuf);
	g_free(namebuf);
	if (htmlview->messageview->mainwin)
		main_window_cursor_normal(htmlview->messageview->mainwin);
	return 0;
#endif /* USE_MOZILLA */

#ifdef USE_CSCHTML
        CscHTMLStream *handle;
	CscHTML *html;
	FILE *fp;
	char *buf, *c, *gbuf, *charset;
	int bufsize;
	const int maxbuf = 1024;
	gint ict;
	gboolean conv_err=FALSE;

	html = CSC_HTML(htmlview->html_widget);
	if (!html) {
		g_warning(_("Bad htmlview parameters."));
		return;
	}
	fp = fopen(file, "rt");
	if (!fp) {
		g_warning(_("Can't open HTML attachment."));
		return;
	}
	buf = malloc(maxbuf);
	if (!buf) {
		g_warning(_("Can't allocate buffer for HTML."));
		return;
	}

	gbuf = malloc(maxbuf*4);
	if (!gbuf) {
		g_warning(_("Can't allocate buffer for converting HTML."));
		return;
	}

	if (prefs_common.force_charset) {
		charset = g_strdup(prefs_common.force_charset);
		debug_print("changed to fc2 for CscHTML %s\n",charset);
	} else if (mimeinfo->charset) {
		charset = (gchar*) get_iconv_charset(mimeinfo->charset);
		debug_print("changed to mi %s for CscHTML from %s\n",charset, mimeinfo->charset);
	} else charset=(gchar*) NULL;

	csc_html_load_empty(html);
	csc_html_allow_font_switch(html, FALSE);
	csc_html_set_default_font_face(html, prefs_common.html_charset);

	handle = csc_html_begin(html);
	ict=conv_iconv_open_tolocal(charset);
	while (!feof(fp)) {
	    if (fgets(buf, maxbuf-1, fp)!=NULL) {
		gint retval=conv_iconv_convert(ict,gbuf, maxbuf*4,   buf, strlen(buf)+1);
		if (retval!=-1)	csc_html_write(html, handle, gbuf, strlen(gbuf));
		else {
			conv_err=TRUE;
			csc_html_end(html, handle, CSC_HTML_STREAM_OK);
			csc_html_load_empty(html);
			handle = csc_html_begin(html);
			csc_html_write(html, handle, "<html><body><h2><center>", strlen("<html><body><h2><center>"));
			csc_html_write(html, handle, _("Incorrect charset"), strlen(_("Incorrect charset")));
			csc_html_write(html, handle, "</center></h2></body></html>", strlen("</center></h2></body></html>"));
			break;
		};
	    }
	}
	conv_iconv_close(ict);
	free(charset);
	csc_html_end(html, handle, CSC_HTML_STREAM_OK);
	free(gbuf);
	free(buf);
	gtk_signal_connect (GTK_OBJECT (html), "link_clicked",
			    GTK_SIGNAL_FUNC (htmlview_on_link_clicked), NULL);
	if (conv_err) return 1; else return 0;
#endif
	return 0;
}

void htmlview_destroy(HtmlView *htmlview)
{
	g_free(htmlview);
}
