/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gtk/gtk.h>

#if HAVE_GDK_IMLIB
#  include <gdk_imlib.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#if HAVE_LOCALE_H
#  include <locale.h>
#endif
#if USE_THREADS
#  include <pthread.h>
#endif
#ifdef GNOME_SUPPORT
#  include <gnome.h>
#endif
#ifdef HAVE_LIBPOPT
#  include <popt.h>
#endif

#include "intl.h"
#include "main.h"
#include "mainwindow.h"
#include "folderview.h"
#include "summaryview.h"
#include "prefs_common.h"
#include "prefs_account.h"
#include "account.h"
#include "procmsg.h"
#include "inc.h"
#include "import.h"
#include "alertpanel.h"
#include "addressbook.h"
#include "compose.h"
#include "utils.h"
#include "cpconfigparse.h"

extern int gtkspell_start(unsigned char *path, char *args[]);
extern void gtkspell_stop();
extern int gtkspell_running();
static gchar* get_maildir();
static void socket_input_cb(gpointer data, gint source, GdkInputCondition condition);
static gint socket_start(void);
static void parse_compose_command(gchar* cline);
gchar *prog_version;
gpointer wmwin_main;
gchar *maildir;
gboolean debug_mode = FALSE;
guint timeout_func = 0;
#if USE_THREADS
pthread_t main_thread_id;
#endif

static gint lock_socket = -1;
static gint lock_socket_tag = 0;

cmdl_mail_info minfo = { FALSE, NULL, FALSE, NULL, FALSE, NULL };

#ifdef HAVE_LIBPOPT

struct poptOption poptoptions[] = {
	{ "debug", 'd', 0, NULL, 'd', N_("Enable debug"), NULL},
	{ "mailto", 'm', POPT_ARG_STRING, &minfo.mailto, 'm', N_("Open compose window and fill \"To:\""), N_("ADDR")},
	{ "subject", 's', POPT_ARG_STRING, &minfo.subject, 's', N_("Open compose window and fill \"Subject:\""), N_("SUBJ")},
	{ "file", 'f', POPT_ARG_STRING, &minfo.fname, 'f', N_("Open compose window and fill message's body from file"), N_("FILE")},
	{ NULL, 0, 0, NULL, 0, NULL, NULL }
};
#endif

#if 0
/* for gettext */
_("File `%s' already exists.\n"
  "Can't create folder.")
#endif

int main(int argc, char *argv[])
{
	gchar *userrc;
	MainWindow *mainwin;
	FolderView *folderview;
	int opt;
#ifdef HAVE_LIBPOPT
        poptContext poptctx;
#endif

#if USE_THREADS
	g_thread_init(NULL);
	if (!g_thread_supported())
		g_error(_("g_thread is not supported by glib.\n"));
	main_thread_id = pthread_self();
#endif

	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	curr_locale=gtk_set_locale();
	userrc=strchr(curr_locale,'.');
	if (userrc!=NULL) {
		gchar** trk=g_strsplit(++userrc,";",1);
		curr_charset=g_strdup(trk[0]);
		g_strfreev(trk);
	}
	else	 curr_charset=g_strdup("ASCII");
	userrc=NULL;
	debug_print("Current charset : %s.\n",curr_charset);

#ifdef GNOME_SUPPORT
#ifdef HAVE_LIBPOPT
	gnome_init_with_popt_table(PACKAGE, VERSION,
				   argc, argv, poptoptions, 0, NULL);
#else
	gnome_init(PACKAGE, VERSION, argc, argv);
#endif
#else
	gtk_init(&argc, &argv);
#endif

#if USE_THREADS
	gdk_threads_enter();
#endif

#if HAVE_GDK_IMLIB
	gdk_imlib_init();
	gtk_widget_push_visual(gdk_imlib_get_visual());
	gtk_widget_push_colormap(gdk_imlib_get_colormap());
#endif

#ifdef HAVE_LIBPOPT
	/* parse command line option */
	/* FIXME: no error checking is done for now */
	poptctx = poptGetContext(PACKAGE, argc, (char **) argv,
				 (const struct poptOption*) poptoptions, 0);
	while ((opt = poptGetNextOpt(poptctx)) > 0)
		switch (opt) {
		case 'd':
			debug_mode = TRUE;
			break;

		case 'm':
			minfo.ext_mailto = TRUE;
			break;

		case 's':
			minfo.ext_subject = TRUE;
			break;

		case 'f':
			minfo.ext_body = TRUE;
			break;
	       	};
	poptFreeContext(poptctx);
#else
        for (opt=1;opt<argc;opt++) {
            if (argv[opt]==NULL) break;
            if (argv[opt][0]!='-') continue;
            if (!strcmp(argv[opt],"--debug") || !strcmp(argv[opt],"-d" )) {
                debug_mode = TRUE;
                continue;
            } else if (!strcmp(argv[opt],"--mailto") || !strcmp(argv[opt],"-m" )) {
		if ((opt+1)<argc) {
            	    minfo.ext_mailto = TRUE;
		    minfo.mailto= g_strdup(argv[opt+1]);
		    opt++;
		} else { fprintf(stderr,_("No mailto address\n"));}
                continue;
            } else if (!strcmp(argv[opt],"--subject") || !strcmp(argv[opt],"-m" )) {
		if ((opt+1)<argc) {
            	    minfo.ext_subject = TRUE;
		    minfo.subject= g_strdup(argv[opt+1]);
		    opt++;
		} else { fprintf(stderr,_("No subject\n"));}
                continue;
            } else if (!strcmp(argv[opt],"--file") || !strcmp(argv[opt],"-f" )) {
		if ((opt+1)<argc) {
            	    minfo.ext_body = TRUE;
		    minfo.fname= g_strdup(argv[opt+1]);
		    opt++;
		} else { fprintf(stderr,_("No file\n"));}
                continue;
            };
        };
#endif

	CHDIR_RETURN_VAL_IF_FAIL(g_get_home_dir(), 1);
	MAKE_DIR_IF_NOT_EXIST(RC_DIR);

	lock_socket = socket_start();
	if (lock_socket < 0) return 0;

	/* read configs */
	userrc = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, MENU_RC, NULL);
	gtk_item_factory_parse_rc(userrc);
	g_free(userrc);

	prog_version = PROG_VERSION_CODENAME;
	maildir = get_maildir();

	CHDIR_RETURN_VAL_IF_FAIL(g_get_home_dir(), 1);

	/* backup if old rc file exists */
	if (is_file_exist(RC_DIR)) {
		if (rename(RC_DIR, RC_DIR ".bak") < 0)
			FILE_OP_ERROR(RC_DIR, "rename");
	}
	MAKE_DIR_IF_NOT_EXIST(get_news_cache_dir());
	MAKE_DIR_IF_NOT_EXIST(get_mime_tmp_dir());

	if (is_file_exist(RC_DIR "/assortrc") &&
	    !is_file_exist(RC_DIR "/filterrc")) {
		if (rename(RC_DIR "/assortrc", RC_DIR "/filterrc") < 0)
			FILE_OP_ERROR(RC_DIR "/assortrc", "rename");
	}

	MAKE_DIR_IF_NOT_EXIST(maildir);
	CHDIR_RETURN_VAL_IF_FAIL(maildir, 1);
	if (!is_dir_exist(INBOX_DIR)) {
    		FILE	*fp;
	        gchar	*path;
		MAKE_DIR_IF_NOT_EXIST(INBOX_DIR);
		path = g_strconcat(INBOX_DIR, G_DIR_SEPARATOR_S, "1", NULL);
	        if ((fp = fopen(path, "w")) != NULL) {
		    fprintf(fp,_(
		    "From: stuphead-hello <black@asplinux.ru>\n" \
		    "Date: Wed, 7 Mar 2001 05:21:07 +0400\n" \
		    "Subject: Wellcome\n" \
		    "X-Mailer: stuphead\n" \
		    "Organization: ASPLinux\n" \
		    "Mime-Version: 1.0\n" \
		    "Content-Type: text/plain; charset=US-ASCII\n" \
		    "Content-Transfer-Encoding: 7bit\n" \
		    "Status:\n\n"));

		    fprintf(fp,_(
		    "Wellcome to Stuphead version %s.\n"),prog_version);
		    fprintf(fp,_(
		    "If you  want to  support us  in our works  please just\n" \
		    "reply to this message.\n" \
		    "------------------------------------------------------\n" \
		    "If you want to know more about last Stuphead updates\n" \
		    "and other information, please visit our site.\n" \
		    "\n" \
		    "http://stuphead.antex.ru/\n" \
		    "TIPS: Start up netscape and then click to URL.\n" \
		    "------------------------------------------------------\n" \
		    "If you want to write bug report please fill form below\n" \
		    "TIPS: Press 'Insert' button to do it more accurately\n" \
		    "\n" \
		    "Operating system:\n" \
		    "Linux		[ ] FreeBSD		[ ] OpenBSD		[ ]\n" \
		    "Solaris		[ ] Other		[ ] _______________\n" \
		    "\n" \
		    "Architecture	_____________________________________\n" \
		    "Kernel version	_____________________________________\n" \
		    "Gtk+ version	_____________________________________\n" \
		    "X-Window 		_____________________________________\n" \
		    "System locale	_____________________________________\n" \
		    "\n" \
		    "BUG: ________________________________________________\n" \
		    "_____________________________________________________\n" \
		    "_____________________________________________________\n" \
		    "_____________________________________________________\n" \
		    "_____________________________________________________\n" \
		    "_____________________________________________________\n" \
		    "\n" \
		    "Your name: __________________________________________\n" \
		    "\n" \
		    "-----------------------------------------------------\n" \
		    "Thank you for your attention.\n" \
		    "					Stuphead development team.\n" \
		    "\n"));
		    fclose(fp);
	    	    g_free(path);
		};
	    };
	MAKE_DIR_IF_NOT_EXIST(OUTBOX_DIR);
	MAKE_DIR_IF_NOT_EXIST(QUEUE_DIR);
	MAKE_DIR_IF_NOT_EXIST(DRAFT_DIR);
	MAKE_DIR_IF_NOT_EXIST(TRASH_DIR);

	srandom((gint)time(NULL));

	prefs_common_read_config();
	prefs_common_save_config();

	/* check and create unix domain socket */
        prefs_common_spell_start(prefs_common.ispell_cmd);

	mainwin = main_window_create
		(prefs_common.sep_folder | prefs_common.sep_msg << 1);
	folderview = mainwin->folderview;

	lock_socket_tag = gdk_input_add(lock_socket,
					GDK_INPUT_READ | GDK_INPUT_EXCEPTION,
					socket_input_cb,
					mainwin);

	account_read_config_all();
	if (!cur_account) {
	    account_edit_open();
	    account_add();
	}
	account_save_config_all();

	folderview_set(folderview);

	if (prefs_common.chk_on_startup) {
		inc_mail(mainwin);
	} else {
		gtk_ctree_select(GTK_CTREE(folderview->ctree),
				 folderview->mailbox);
		gtk_widget_grab_focus(folderview->ctree);
	}
	wmwin_main=(gpointer) mainwin;
	if (prefs_common.autochk_newmail) {
		timeout_func=gtk_timeout_add(prefs_common.autochk_itv*1000*60,(GtkFunction) inc_all_account_mail_by_timer,wmwin_main);
	};

	/* ignore SIGPIPE signal for preventing sudden death of program */
	signal(SIGPIPE, SIG_IGN);

	if (minfo.ext_mailto || minfo.ext_subject || minfo.ext_body)
	    compose_new();

	gtk_main();
#if USE_THREADS
	gdk_threads_leave();
#endif

	return 0;
}

void app_will_exit(GtkWidget *widget, gpointer data)
{
	gchar *filename;
	MainWindow *mainwin = data;

	if (compose_get_compose_list()) {
		if (alertpanel(_("Notice"),
			       _("Composing message exists.\n"
				 "Really quit?"),
			       _("OK"), _("Cancel"), NULL) != G_ALERTDEFAULT)
			return;
	}

        if (prefs_common.clean_on_exit) if (!trash_is_empty()) main_window_clean_trash(mainwin, prefs_common.ask_on_clean);
        if (prefs_common.queue_on_exit && !queue_is_empty()) {
            if (alertpanel(_("Notice"),
                           _("Queue not empty.\n"
                             "Send queue ?"),
                           _("OK"), _("Cancel"), NULL) == G_ALERTDEFAULT)
                send_queue_cb(mainwin, 0, NULL);;
        };

	/* save all state before exiting */
	folderview_write_cache(mainwin->folderview);
	summary_write_cache(mainwin->summaryview);

	main_window_get_size(mainwin);
	main_window_get_position(mainwin);
	prefs_common_save_config();
	account_save_config_all();
	addressbook_export_to_file();

	filename = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, MENU_RC, NULL);
	gtk_item_factory_dump_rc(filename, NULL, TRUE);
	g_free(filename);

	/* delete temporary files */
	gtkspell_stop();
	remove_all_files(get_mime_tmp_dir());
	/* delete socket */
	unlink(get_stuphead_socket_name());

	gtk_main_quit();
}

static gchar*
get_maildir()
{
    gchar* mh_profile;
    char line[BUFFSIZE+1];
    FILE* in;
    char* maildir;

    mh_profile = g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, ".mh_profile",
			      NULL);

    if (!is_file_exist( mh_profile ) )
      goto end;

    in = fopen( mh_profile, "r" );
    if( in == NULL )
      goto end;

    while( fgets( (char*)line, BUFFSIZE, in ) ) {
      char* p;
      if( strncmp( (char*)line, "Path:", 5 ) != 0 )
	continue;

      p = strchr((char*) line, '\r' );
      if( p )
	*p = '\0';
      p = strchr((char*) line, '\n' );
      if( p )
	*p = '\0';

      maildir = strchr( (char*)line, ':' );
      maildir++;
      while( *maildir != '\0' && *maildir == ' ' )
	maildir++;
      if( *maildir == '\0' )
	goto end;
      break;
    }

    fclose(in);
    g_free( mh_profile );

    return g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, maildir,
	                      NULL);

end:
    g_free( mh_profile );
    return g_strconcat(g_get_home_dir(), G_DIR_SEPARATOR_S, MAIL_DIR,
	                      NULL);
}

int safe_wctomb(char *s, wchar_t wc) {
    int result=wctomb(s,wc);
    if (result==-1) debug_print("conversion failed for wide to mb\n");
    return result;
};

static gint socket_start(void)
{
	gint sock;
	gchar *path;

	path = get_stuphead_socket_name();
	sock = sock_connect_unix(path);
	if (sock < 0) {
		unlink(path);
		return sock_open_unix(path);
	}

	if (minfo.ext_mailto || minfo.ext_subject || minfo.ext_body) {
		sock_printf(sock, "COMP");
    		if (minfo.ext_mailto && minfo.mailto!=NULL)
			sock_printf(sock, "m%s\001", minfo.mailto);
		if (minfo.ext_subject && minfo.subject!=NULL)
			sock_printf(sock, "s%s\001", minfo.subject);
		if (minfo.ext_body && minfo.fname!=NULL)
			sock_printf(sock, "f%s\001", minfo.fname);
		sock_printf(sock,"\n");
	} else sock_printf(sock, "WNUP\n");
	sock_close(sock);
	return -1;
}
static void parse_compose_command(gchar* cline) {
	gchar *cur=cline;
	gchar *del=NULL;
	enum {CM_N,CM_M,CM_S,CM_F};
	guint parsestate=CM_N;

	while (cur[0]!='\0' && cur[0]!='\n') {
	    cur++;
	    if (parsestate==CM_N) {
		switch(cur[0]) {
		    case 'm': parsestate=CM_M; del=cur+1;break;
		    case 'f': parsestate=CM_F; del=cur+1;break;
		    case 's': parsestate=CM_S; del=cur+1;break;
		}
	    } else
		if ((cur[0]=='\0' || cur[0]=='\001' || cur[0]=='\n') && del!=NULL) {
		    switch(parsestate) {
			case CM_M:
				minfo.ext_mailto=TRUE;
				minfo.mailto=g_strndup(del,cur-del);
				break;
			case CM_S:
				minfo.ext_subject=TRUE;
				minfo.subject=g_strndup(del,cur-del);
				break;
			case CM_F:
				minfo.ext_body=TRUE;
				minfo.fname=g_strndup(del,cur-del);
				break;
		    }
		    parsestate=CM_N;
		}
	}
};
static void socket_input_cb(gpointer data, gint source, GdkInputCondition condition)
{
	MainWindow *mainwin = (MainWindow *)data;
	gint sock;
	gchar buf[BUFFSIZE];

	sock = sock_accept(source);
	sock_read(sock, buf, sizeof(buf));
	sock_close(sock);

	if        (!strncmp(buf, "WNUP",4)){
		gdk_window_raise(mainwin->window->window);

	} else if (!strncmp(buf, "COMP",4)){
		//main_window_popup(mainwin);
		parse_compose_command(buf);
		compose_new();
	}
}
