/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * stuphead -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#define MANUAL_HTML_INDEX	"stuphead.html"

#include <glib.h>
#include <string.h>

#include "main.h"
#include "prefs_common.h"
#include "manual.h"
#include "utils.h"

static void manual_open_uri(const gchar *uri);

void manual_open(const gchar* lang)
{
	gchar *file_uri;

	g_return_if_fail(lang != NULL);

	file_uri = g_strconcat("file://", MANUALDIR,
			       G_DIR_SEPARATOR_S, lang, G_DIR_SEPARATOR_S,
			       MANUAL_HTML_INDEX, NULL);
	debug_print("Opening manual: %s\n", file_uri);
	manual_open_uri(file_uri);
	g_free(file_uri);
}

static void manual_open_uri(const gchar *uri)
{
	static gchar *default_cmdline = "netscape -remote openURL(%s,raise)";
	gchar buf[1024];
	gchar *p;

	if (prefs_common.uri_cmd &&
	    (p = strchr(prefs_common.uri_cmd, '%')) && *(p + 1) == 's' &&
	    !strchr(p + 2, '%'))
		g_snprintf(buf, sizeof(buf), prefs_common.uri_cmd, uri);
	else {
		if (prefs_common.uri_cmd)
			g_warning(_("Open URI command line is invalid: `%s'"),
				  prefs_common.uri_cmd);
		g_snprintf(buf, sizeof(buf), default_cmdline, uri);
	}

	execute_command_line(buf);
}
