/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkhbbox.h>
#include <gtk/gtk.h>

#include "intl.h"
#include "inputdialog.h"
#include "manage_window.h"
#include "gtkutils.h"
#include "utils.h"
#include <stdio.h>

#define INPUT_DIALOG_WIDTH	420
#define INPUT_DIALOG_HEIGHT 	400

static gboolean ack;

static GtkWidget *dialog;
static GtkWidget *msg_label;
static GtkWidget *selector;
static GtkWidget *entry;
static GtkWidget *ok_button;
static GtkWidget *scrolled_window;

static void select_dialog_create	(void);
static void select_dialog_set	(const gchar	*title,
				 const gchar	*message,
				 const GSList 	*list,
				 const gchar	*default_string);

static void ok_clicked		(GtkWidget	*widget,
				 gpointer	 data);
static void cancel_clicked	(GtkWidget	*widget,
				 gpointer	 data);
static void key_pressed		(GtkWidget	*widget,
				 GdkEventKey	*event,
				 gpointer	 data);
static void entry_activated	(GtkEditable	*editable);
static void selector_cb		(GtkWidget *widget,gint row, gint column, GdkEventButton *event,gpointer data);

gchar *select_dialog(const gchar *title, const gchar *message,
		    const GSList 	*list,
		    const gchar *default_string)
{
	gchar *str;
	int i;
	if (dialog && GTK_WIDGET_VISIBLE(dialog)) return NULL;

	if (!dialog)
		select_dialog_create();

	select_dialog_set(title, message, list, default_string);
	gtk_widget_show(dialog);
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	manage_window_set_transient(GTK_WINDOW(dialog));

	gtk_main();

        manage_window_focus_out(dialog, NULL, NULL);
	gtk_widget_hide(dialog);
	gtk_entry_set_visibility(GTK_ENTRY(entry), TRUE);
        gtk_clist_clear(GTK_CLIST(selector));
	if (ack) {
		str = gtk_editable_get_chars(GTK_EDITABLE(entry), 0, -1);
		if (str && *str == '\0')
			str = NULL;
	} else
		str = NULL;

	GTK_EVENTS_FLUSH();

	if (str!=NULL) { for (i=0;str[i]>32;i++);
		    str[i]=0;
		    return g_strdup(str);
	} else return NULL;

}

gchar *select_dialog_with_invisible(const gchar *title, const gchar *message,
				    const GSList 	*list,
				   const gchar *default_string)
{
	if (dialog && GTK_WIDGET_VISIBLE(dialog)) return NULL;

	if (!dialog)
		select_dialog_create();

	gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);

	return select_dialog(title, message, list,default_string);
}

static void select_dialog_create(void)
{
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *confirm_area;
	GtkWidget *cancel_button;

	dialog = gtk_dialog_new();
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
	gtk_widget_set_usize(dialog, INPUT_DIALOG_WIDTH, INPUT_DIALOG_HEIGHT);
	gtk_container_set_border_width
		(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 5);
	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
	gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
			   GTK_SIGNAL_FUNC(cancel_clicked), NULL);
	gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event",
			   GTK_SIGNAL_FUNC(key_pressed), NULL);
	gtk_signal_connect(GTK_OBJECT(dialog), "focus_in_event",
			   GTK_SIGNAL_FUNC(manage_window_focus_in), NULL);
	gtk_signal_connect(GTK_OBJECT(dialog), "focus_out_event",
			   GTK_SIGNAL_FUNC(manage_window_focus_out), NULL);

	gtk_widget_realize(dialog);

	vbox = gtk_vbox_new(FALSE, 8);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), vbox);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	msg_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), msg_label, FALSE, FALSE, 0);
	gtk_label_set_justify(GTK_LABEL(msg_label), GTK_JUSTIFY_LEFT);

	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
                                	GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled_window, TRUE, TRUE, 0);
	gtk_widget_show (scrolled_window);

	selector= gtk_clist_new(1);
	gtk_clist_set_shadow_type (GTK_CLIST(selector), GTK_SHADOW_OUT);
	gtk_clist_set_column_width (GTK_CLIST(selector), 0, 250);
	gtk_container_add(GTK_CONTAINER(scrolled_window), selector);
	//gtk_box_pack_start(GTK_BOX(vbox), selector, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(selector), "select_row", GTK_SIGNAL_FUNC(selector_cb),NULL);

	entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(vbox), entry, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(entry), "activate",
			   GTK_SIGNAL_FUNC(entry_activated), NULL);


	gtkut_button_set_create(&confirm_area,
				&ok_button,     _("OK"),
				&cancel_button, _("Cancel"),
				NULL, NULL);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
			  confirm_area);
	gtk_widget_grab_default(ok_button);

	gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
			   GTK_SIGNAL_FUNC(ok_clicked), NULL);
	gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
			   GTK_SIGNAL_FUNC(cancel_clicked), NULL);


	gtk_widget_show_all(GTK_DIALOG(dialog)->vbox);
}


void my_slist_insert(gpointer data,gpointer user_data) {
	gchar* mn[2];
	mn[0]=g_strdup(data);
	mn[1]=NULL;
	gtk_clist_prepend( GTK_CLIST(selector), mn);
};

static void select_dialog_set(const gchar *title, const gchar *message,
			     const GSList 	*list,
			     const gchar *default_string)
{
	gtk_window_set_title(GTK_WINDOW(dialog), title);
	gtk_label_set_text(GTK_LABEL(msg_label), message);
	if (default_string && *default_string)
		gtk_entry_set_text(GTK_ENTRY(entry), default_string);
	else
		gtk_entry_set_text(GTK_ENTRY(entry), "");
	gtk_entry_set_position(GTK_ENTRY(entry), 0);
	gtk_entry_select_region(GTK_ENTRY(entry), 0, -1);
	gtk_clist_clear ((GtkCList*) selector);
	g_slist_foreach ((GSList*)list,(GFunc) my_slist_insert, NULL);


	gtk_widget_grab_focus(ok_button);
	gtk_widget_grab_focus(entry);
}

static void ok_clicked(GtkWidget *widget, gpointer data)
{
	ack = TRUE;
	gtk_main_quit();
}

static void cancel_clicked(GtkWidget *widget, gpointer data)
{
	ack = FALSE;
	gtk_main_quit();
}

static void key_pressed(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	if (event && event->keyval == GDK_Escape) {
		ack = FALSE;
		gtk_main_quit();
	};
}

static void entry_activated(GtkEditable *editable)
{
	ack = TRUE;
	gtk_main_quit();
}


static void selector_cb(GtkWidget *widget,gint row, gint column, GdkEventButton *event,gpointer data)
{
        gchar *text;
        gtk_clist_get_text(GTK_CLIST(selector), row, column, &text);
	gtk_entry_set_text((GtkEntry*) entry, text);
}
