/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999,2000 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkbindings.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkpixmap.h>
#include <gtk/gtkctree.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtksignal.h>
#include <gtk/gtktext.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkmenuitem.h>
#include <gtk/gtkitemfactory.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkstyle.h>
#include <gtk/gtkarrow.h>
#include <gtk/gtkeventbox.h>
#include <gtk/gtkstatusbar.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>

#include "main.h"
#include "menu.h"
#include "mainwindow.h"
#include "folderview.h"
#include "summaryview.h"
#include "messageview.h"
#include "foldersel.h"
#include "procmsg.h"
#include "procheader.h"
#include "headerwindow.h"
#include "prefs_common.h"
#include "account.h"
#include "compose.h"
#include "utils.h"
#include "gtkutils.h"
#include "filesel.h"
#include "alertpanel.h"
#include "inputdialog.h"
#include "news.h"
#include "statusbar.h"
#include "filter.h"

#include "pixmaps/dir-open.xpm"
#include "pixmaps/new.xpm"
#include "pixmaps/important.xpm"
#include "pixmaps/unread.xpm"
#include "pixmaps/clip.xpm"
#include "viewsource.h"

#define STATUSBAR_PUSH(mainwin, str) \
{ \
	gtk_statusbar_push(GTK_STATUSBAR(mainwin->statusbar), \
			   mainwin->summaryview_cid, str); \
	gtkut_widget_wait_for_draw(mainwin->statusbar); \
}

#define STATUSBAR_POP(mainwin) \
{ \
	gtk_statusbar_pop(GTK_STATUSBAR(mainwin->statusbar), \
			  mainwin->summaryview_cid); \
}

static GdkFont *smallfont;

static GdkPixmap *folderxpm;
static GdkBitmap *folderxpmmask;
//static GdkPixmap *regularxpm;
//static GdkBitmap *regularxpmmask;
static GdkPixmap *newxpm;
static GdkBitmap *newxpmmask;
static GdkPixmap *unreadxpm;
static GdkBitmap *unreadxpmmask;
static GdkPixmap *importantxpm;
static GdkBitmap *importantxpmmask;
static GdkPixmap *clipxpm;
static GdkBitmap *clipxpmmask;

static void summary_free_msginfo_func	(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);
static void summary_set_marks_func	(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);
static void summary_write_cache_func	(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);

static void summary_set_popup_sensitive	(SummaryView		*summaryview);

static GtkCTreeNode *summary_find_next_unread_msg
					(SummaryView		*summaryview,
					 GtkCTreeNode		*current_node);
#if 0
static GtkCTreeNode *summary_find_prev_unread_msg
					(SummaryView		*summaryview,
					 GtkCTreeNode		*current_node);
#endif

static void summary_update_status	(SummaryView		*summaryview);

/* display functions */
static void summary_status_show		(SummaryView		*summaryview);
static void summary_set_ctree_from_list	(SummaryView		*summaryview,
					 GSList			*mlist);
static void summary_set_header		(gchar			*text[],
					 MsgInfo		*msginfo);
static void summary_display_msg		(SummaryView		*summaryview,
					 GtkCTreeNode		*row,
					 gboolean		 new_window);
static void summary_toggle_view		(SummaryView		*summaryview);

/* message handling */
static void summary_mark_row		(SummaryView		*summaryview,
					 GtkCTreeNode		*row);
static void summary_delete_row		(SummaryView		*summaryview,
					 GtkCTreeNode		*row);
static void summary_unmark_row		(SummaryView		*summaryview,
					 GtkCTreeNode		*row);
static void summary_move_row_to		(SummaryView		*summaryview,
					 GtkCTreeNode		*row,
					 const gchar		*to_folder);

static void summary_delete_duplicated_func
					(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 SummaryView		*summaryview);

static void summary_execute_move	(SummaryView		*summaryview);
static void summary_execute_move_func	(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);
static void summary_execute_delete	(SummaryView		*summaryview);
static void summary_execute_delete_func	(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);
static void summary_execute_copy	(SummaryView		*summaryview);
static void summary_execute_copy_func	(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);

/* thread functions */
static void summary_thread_func			(GtkCTree	*ctree,
						 GtkCTreeNode	*node,
						 gpointer	 data);
static void summary_unthread_func		(GtkCTree	*ctree,
						 GtkCTreeNode	*node,
						 gpointer	 data);
static void summary_unthread_for_exec		(SummaryView	*summaryview);
static void summary_unthread_for_exec_func	(GtkCTree	*ctree,
						 GtkCTreeNode	*node,
						 gpointer	 data);

static void summary_filter_func		(GtkCTree		*ctree,
					 GtkCTreeNode		*node,
					 gpointer		 data);

/* callback functions */
static void summary_toggle_pressed	(GtkWidget		*eventbox,
					 GdkEventButton		*event,
					 SummaryView		*summaryview);
static void summary_button_pressed	(GtkWidget		*ctree,
					 GdkEventButton		*event,
					 SummaryView		*summaryview);
static void summary_button_released	(GtkWidget		*ctree,
					 GdkEventButton		*event,
					 SummaryView		*summaryview);
static void summary_key_pressed		(GtkWidget		*ctree,
					 GdkEventKey		*event,
					 SummaryView		*summaryview);
static void summary_row_selected	(GtkCList		*clist,
					 gint			 row,
					 gint			 column,
					 GdkEvent		*event,
					 gpointer		 data);
static void summary_selected		(GtkCTree		*ctree,
					 GtkCTreeNode		*row,
					 gint			 column,
					 SummaryView		*summaryview);
static void summary_col_resized		(GtkCList		*clist,
					 gint			 column,
					 gint			 width,
					 SummaryView		*summaryview);
static void summary_reply_cb		(SummaryView		*summaryview,
					 guint			 action,
					 GtkWidget		*widget);

static void summary_num_clicked		(GtkWidget		*button,
					 SummaryView		*summaryview);
static void summary_size_clicked	(GtkWidget		*button,
					 SummaryView		*summaryview);
static void summary_date_clicked	(GtkWidget		*button,
					 SummaryView		*summaryview);
static void summary_from_clicked	(GtkWidget		*button,
					 SummaryView		*summaryview);
static void summary_subject_clicked	(GtkWidget		*button,
					 SummaryView		*summaryview);

/* custom compare functions for sorting */

static gint summary_cmp_by_num		(GtkCList		*clist,
					 gconstpointer		 ptr1,
					 gconstpointer		 ptr2);
static gint summary_cmp_by_size		(GtkCList		*clist,
					 gconstpointer		 ptr1,
					 gconstpointer		 ptr2);
static gint summary_cmp_by_date		(GtkCList		*clist,
					 gconstpointer		 ptr1,
					 gconstpointer		 ptr2);
static gint summary_cmp_by_from		(GtkCList		*clist,
					 gconstpointer		 ptr1,
					 gconstpointer		 ptr2);
static gint summary_cmp_by_subject	(GtkCList		*clist,
					 gconstpointer		 ptr1,
					 gconstpointer		 ptr2);

void collapse_thread(SummaryView *summaryview);

static GtkItemFactoryEntry summary_popup_entries[] =
{
	{N_("/Collapse"),		NULL, collapse_thread,	0, NULL},
	{N_("/---"),			NULL, NULL,		0, "<Separator>"},
	{N_("/M_ove..."),		NULL, summary_move_to,	0, NULL},
	{N_("/_Copy..."),		NULL, summary_copy_to, 	0, NULL},
	{N_("/_Delete"),		NULL, summary_delete,	0, NULL},
	{N_("/E_xecute"),		NULL, summary_execute,	0, NULL},
	{N_("/---"),			NULL, NULL,		0, "<Separator>"},
	{N_("/_Reply"),			NULL, summary_reply_cb,	0, NULL},
	{N_("/Reply with _quotation"),	NULL, summary_reply_cb,	1, NULL},
	{N_("/_Forward"),		NULL, summary_reply_cb, 2, NULL},
	{N_("/_Forward as attachment"),	NULL, summary_reply_cb, 3, NULL},
        {N_("/---"),			NULL, NULL,		0, "<Separator>"},
	{N_("/Open in new _window"),	NULL, summary_open_msg,	0, NULL},
	{N_("/Re_edit"),		NULL, summary_reedit,   0, NULL},
	{N_("/---"),			NULL, NULL,		0, "<Separator>"},
	{N_("/_Save as..."),		NULL, summary_save_as,	0, NULL},
	{N_("/_Print..."),		NULL, summary_print,	0, NULL},
	{N_("/---"),			NULL, NULL,		0, "<Separator>"},
	{N_("/_Mark"),			NULL, NULL,		0, "<Branch>"},
	{N_("/_Mark/_Mark"),		NULL, summary_mark,	0, NULL},
	{N_("/_Mark/_Unmark"),		NULL, summary_unmark,	0, NULL},
	{N_("/_Mark/---"),		NULL, NULL,		0, "<Separator>"},
	{N_("/_Mark/Mark as unr_ead"),	NULL, summary_mark_as_unread, 0, NULL},
        {N_("/_Mark/Mark as _important"), NULL, summary_mark_as_important,0, NULL},
	{N_("/_Mark/Make it as _being read"),
					NULL, summary_mark_as_read, 0, NULL},
	{N_("/Select _all"),		NULL, summary_select_all, 0, NULL},
	{N_("/U_nselect all"),		NULL, summary_unselect_all, 0, NULL}
};

SummaryView *summary_create(void)
{
	SummaryView *summaryview;
	gchar *titles[N_SUMMARY_COLS] = {_("M"), _("U")};
	GtkWidget *vbox;
	GtkWidget *scrolledwin;
	GtkWidget *ctree;
	GtkWidget *hbox;
	GtkWidget *statlabel_folder;
	GtkWidget *statlabel_select;
	GtkWidget *statlabel_msgs;
	GtkWidget *toggle_eventbox;
	GtkWidget *toggle_arrow;
	GtkWidget *popupmenu;
	GtkItemFactory *popupfactory;
	GtkBindingSet *binding_set;
	gint n_entries;
	gint i;

	debug_print(_("Creating summary view...\n"));
	summaryview = g_new0(SummaryView, 1);

	vbox = gtk_vbox_new(FALSE, 2);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_box_pack_start(GTK_BOX(vbox), scrolledwin, TRUE, TRUE, 0);
	gtk_widget_set_usize(vbox,
			     prefs_common.summaryview_width,
			     prefs_common.summaryview_height);

	if (prefs_common.trans_hdr) {
		titles[S_COL_NUMBER]  = _("No.");
		titles[S_COL_DATE]    = _("Date");
		titles[S_COL_FROM]    = _("From");
		titles[S_COL_SUBJECT] = _("Subject");
	} else {
		titles[S_COL_NUMBER]  = "No.";
		titles[S_COL_DATE]    = "Date";
		titles[S_COL_FROM]    = "From";
		titles[S_COL_SUBJECT] = "Subject";
	}
	titles[S_COL_SIZE] = _("Size");

	ctree = gtk_ctree_new_with_titles(N_SUMMARY_COLS, S_COL_SUBJECT, titles);
	gtk_scrolled_window_set_vadjustment(GTK_SCROLLED_WINDOW(scrolledwin),
					    GTK_CLIST(ctree)->vadjustment);
	gtk_container_add(GTK_CONTAINER(scrolledwin), ctree);
	gtk_clist_set_selection_mode(GTK_CLIST(ctree), GTK_SELECTION_EXTENDED);
	gtk_clist_set_column_justification(GTK_CLIST(ctree), S_COL_MARK,
					   GTK_JUSTIFY_CENTER);
	gtk_clist_set_column_justification(GTK_CLIST(ctree), S_COL_UNREAD,
					   GTK_JUSTIFY_CENTER);
	gtk_clist_set_column_justification(GTK_CLIST(ctree), S_COL_MIME,
					   GTK_JUSTIFY_CENTER);
	gtk_clist_set_column_justification(GTK_CLIST(ctree), S_COL_NUMBER,
					   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification(GTK_CLIST(ctree), S_COL_SIZE,
					   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_MARK,
				   prefs_common.summary_col_mark);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_UNREAD,
				   prefs_common.summary_col_unread);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_MIME,
				   prefs_common.summary_col_mime);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_NUMBER,
				   prefs_common.summary_col_number);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_SIZE,
				   prefs_common.summary_col_size);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_DATE,
				   prefs_common.summary_col_date);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_FROM,
				   prefs_common.summary_col_from);
	gtk_clist_set_column_width(GTK_CLIST(ctree), S_COL_SUBJECT,
				   prefs_common.summary_col_subject);
	gtk_ctree_set_line_style(GTK_CTREE(ctree), GTK_CTREE_LINES_DOTTED);
	gtk_ctree_set_expander_style(GTK_CTREE(ctree),
				     GTK_CTREE_EXPANDER_SQUARE);
#if 0
	gtk_ctree_set_line_style(GTK_CTREE(ctree), GTK_CTREE_LINES_NONE);
	gtk_ctree_set_expander_style(GTK_CTREE(ctree),
				     GTK_CTREE_EXPANDER_TRIANGLE);
#endif
	gtk_ctree_set_indent(GTK_CTREE(ctree), 18);
	//gtk_clist_set_reorderable(GTK_CLIST(ctree), TRUE);
	gtk_object_set_user_data(GTK_OBJECT(ctree), summaryview);
	/* don't expand tree with double-click */
	GTK_CLIST(ctree)->button_actions[0] &= ~GTK_BUTTON_EXPANDS;


	/* don't let title buttons take key focus */
	for (i = 0; i < N_SUMMARY_COLS; i++)
		GTK_WIDGET_UNSET_FLAGS(GTK_CLIST(ctree)->column[i].button,
				       GTK_CAN_FOCUS);

	/* connect signal to the buttons for sorting */
	gtk_signal_connect
		(GTK_OBJECT(GTK_CLIST(ctree)->column[S_COL_NUMBER].button),
		 "clicked",
		 GTK_SIGNAL_FUNC(summary_num_clicked),
		 summaryview);
	gtk_signal_connect
		(GTK_OBJECT(GTK_CLIST(ctree)->column[S_COL_SIZE].button),
		 "clicked",
		 GTK_SIGNAL_FUNC(summary_size_clicked),
		 summaryview);
	gtk_signal_connect
		(GTK_OBJECT(GTK_CLIST(ctree)->column[S_COL_DATE].button),
		 "clicked",
		 GTK_SIGNAL_FUNC(summary_date_clicked),
		 summaryview);
	gtk_signal_connect
		(GTK_OBJECT(GTK_CLIST(ctree)->column[S_COL_FROM].button),
		 "clicked",
		 GTK_SIGNAL_FUNC(summary_from_clicked),
		 summaryview);
	gtk_signal_connect
		(GTK_OBJECT(GTK_CLIST(ctree)->column[S_COL_SUBJECT].button),
		 "clicked",
		 GTK_SIGNAL_FUNC(summary_subject_clicked),
		 summaryview);

	/* create status label */
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_end(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	statlabel_folder = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), statlabel_folder, FALSE, FALSE, 2);
	statlabel_select = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), statlabel_select, FALSE, FALSE, 16);

	/* toggle view button */
	toggle_eventbox = gtk_event_box_new();
	gtk_box_pack_end(GTK_BOX(hbox), toggle_eventbox, FALSE, FALSE, 4);
	toggle_arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(toggle_eventbox), toggle_arrow);

	statlabel_msgs = gtk_label_new("");
	gtk_box_pack_end(GTK_BOX(hbox), statlabel_msgs, FALSE, FALSE, 4);

	/* create popup menu */
	n_entries = sizeof(summary_popup_entries) /
		sizeof(summary_popup_entries[0]);
	popupmenu = menu_create_items(summary_popup_entries, n_entries,
				      "<SummaryView>", &popupfactory,
				      summaryview);

	/* bind keys */
	binding_set = gtk_binding_set_by_class
		(GTK_CLIST_CLASS(GTK_OBJECT(ctree)->klass));

	gtk_binding_entry_add_signal(binding_set, GDK_n, GDK_CONTROL_MASK,
				     "scroll_vertical", 2,
				     GTK_TYPE_ENUM, GTK_SCROLL_STEP_FORWARD,
				     GTK_TYPE_FLOAT, 0.0);
	gtk_binding_entry_add_signal(binding_set, GDK_p, GDK_CONTROL_MASK,
				     "scroll_vertical", 2,
				     GTK_TYPE_ENUM, GTK_SCROLL_STEP_BACKWARD,
				     GTK_TYPE_FLOAT, 0.0);
	gtk_binding_entry_clear(binding_set, GDK_space, 0);

	/* connect signals */
	gtk_signal_connect(GTK_OBJECT(ctree), "select_row",
			   GTK_SIGNAL_FUNC(summary_row_selected), summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "tree_select_row",
			   GTK_SIGNAL_FUNC(summary_selected), summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "button_press_event",
			   GTK_SIGNAL_FUNC(summary_button_pressed),
			   summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "button_release_event",
			   GTK_SIGNAL_FUNC(summary_button_released),
			   summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "key_press_event",
			   GTK_SIGNAL_FUNC(summary_key_pressed), summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "resize_column",
			   GTK_SIGNAL_FUNC(summary_col_resized), summaryview);

	gtk_signal_connect(GTK_OBJECT(toggle_eventbox), "button_press_event",
			   GTK_SIGNAL_FUNC(summary_toggle_pressed),
			   summaryview);

	summaryview->vbox = vbox;
	summaryview->scrolledwin = scrolledwin;
	summaryview->ctree = ctree;
	summaryview->hbox = hbox;
	summaryview->statlabel_folder = statlabel_folder;
	summaryview->statlabel_select = statlabel_select;
	summaryview->statlabel_msgs = statlabel_msgs;
	summaryview->toggle_eventbox = toggle_eventbox;
	summaryview->toggle_arrow = toggle_arrow;
	summaryview->popupmenu = popupmenu;
	summaryview->popupfactory = popupfactory;
	summaryview->msg_is_toggled_on = TRUE;
	summaryview->sort_mode = SORT_BY_NONE;
	summaryview->sort_type = GTK_SORT_ASCENDING;

	summary_change_display_item(summaryview);

	gtk_widget_show_all(vbox);

	return summaryview;
}

void summary_init(SummaryView *summaryview)
{
	GtkStyle *style;
	GtkWidget *pixmap;

	//PIXMAP_CREATE(summaryview->ctree, regularxpm, regularxpmmask,
	//		regular_xpm);
	PIXMAP_CREATE(summaryview->ctree, newxpm, newxpmmask, new_xpm);
	PIXMAP_CREATE(summaryview->ctree, unreadxpm, unreadxpmmask, unread_xpm);
	PIXMAP_CREATE(summaryview->ctree, importantxpm, importantxpmmask, important_xpm);
	PIXMAP_CREATE(summaryview->ctree, clipxpm, clipxpmmask, clip_xpm);
	PIXMAP_CREATE(summaryview->hbox, folderxpm, folderxpmmask,
		      DIRECTORY_OPEN_XPM);

	pixmap = gtk_pixmap_new(clipxpm, clipxpmmask);
	gtk_clist_set_column_widget(GTK_CLIST(summaryview->ctree),
				    S_COL_MIME, pixmap);
	gtk_widget_show(pixmap);

	if (!smallfont)
		smallfont = gdk_fontset_load(SMALL_FONT);

	style = gtk_style_copy(gtk_widget_get_style
				(summaryview->statlabel_folder));
	if (smallfont) style->font = smallfont;
	gtk_widget_set_style(summaryview->statlabel_folder, style);
	gtk_widget_set_style(summaryview->statlabel_select, style);
	gtk_widget_set_style(summaryview->statlabel_msgs, style);

	pixmap = gtk_pixmap_new(folderxpm, folderxpmmask);
	gtk_box_pack_start(GTK_BOX(summaryview->hbox), pixmap, FALSE, FALSE, 4);
	gtk_box_reorder_child(GTK_BOX(summaryview->hbox), pixmap, 0);
	gtk_widget_show(pixmap);

	summary_clear_list(summaryview);
	summary_set_popup_sensitive(summaryview);
}

gboolean summary_show(SummaryView *summaryview, FolderInfo *finfo,
		      gboolean update_cache)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCTreeNode *node;
	GSList *mlist = NULL;
	gchar *buf;

	STATUSBAR_POP(summaryview->mainwin);
	m_finfo=finfo;

#if 0
	/* process the marks if any */
	if (summaryview->deleted > 0 || summaryview->moved > 0) {
		AlertValue val;

		val = alertpanel(_("Process mark"),
				 _("Some marks are left. Process it?"),
				 _("Yes"), _("No"), _("Cancel"));
		if (G_ALERTDEFAULT == val)
			summary_execute(summaryview);
		else if (G_ALERTALTERNATE == val)
			summary_write_cache(summaryview);
		else
			return FALSE;
	} else
#endif
		summary_write_cache(summaryview);

	gtk_clist_freeze(GTK_CLIST(ctree));

	summary_clear_list(summaryview);
	summary_set_popup_sensitive(summaryview);
	messageview_clear(summaryview->messageview);

	if (finfo && finfo->type == F_NEWSGROUP) {
		gchar *server;

		summaryview->mode = SUMMARY_NEWS;
		server = g_dirname(finfo->path);
		debug_print("server = %s, group = %s\n",
			    server, g_basename(finfo->path));
		buf = g_strdup_printf(_("Connecting to %s ..."), server);
		STATUSBAR_PUSH(summaryview->mainwin, buf);
		g_free(buf);

		main_window_cursor_wait(summaryview->mainwin);
		summaryview->news_session = news_session_get(server);
		STATUSBAR_POP(summaryview->mainwin);
		g_free(server);
		main_window_cursor_normal(summaryview->mainwin);
	}

	if (!finfo || !finfo->path || (summaryview->mode == SUMMARY_MAIL &&
	     (change_dir(maildir) < 0 || !is_dir_exist(finfo->path)))) {
		debug_print(_("empty folder\n\n"));
		summary_clear_all(summaryview);
		gtk_clist_thaw(GTK_CLIST(ctree));
		return TRUE;
	}

	summaryview->cur_folder = g_strdup(finfo->path);

	gtk_signal_disconnect_by_data(GTK_OBJECT(ctree), summaryview);

	buf = g_strdup_printf(_("Scanning folder (%s)..."), finfo->path);
	debug_print("%s\n", buf);
	STATUSBAR_PUSH(summaryview->mainwin, buf);
	g_free(buf);

	main_window_cursor_wait(summaryview->mainwin);

	/* get messages list */
	if (summaryview->mode == SUMMARY_MAIL) {
		struct stat s;
		gboolean scan_new = TRUE;

		if (stat(finfo->path, &s) < 0) {
			FILE_OP_ERROR(finfo->path, "stat");
		} else {
			if (finfo->mtime == s.st_mtime) {
				debug_print("Folder is not modified.\n");
				scan_new = FALSE;
			} else
				finfo->mtime = s.st_mtime;
		}

		mlist = procmsg_get_msgs(finfo->path, !update_cache, scan_new);
	} else
		mlist = news_get_article_info(summaryview->news_session,
					      finfo->path, !update_cache);

	STATUSBAR_POP(summaryview->mainwin);

	/* set ctree and hash table from the msginfo list
	   creating thread, and count the number of messages */
	summary_set_ctree_from_list(summaryview, mlist);

	g_slist_free(mlist);

	summary_write_cache(summaryview);

	gtk_signal_connect(GTK_OBJECT(ctree), "select_row",
			   GTK_SIGNAL_FUNC(summary_row_selected), summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "tree_select_row",
			   GTK_SIGNAL_FUNC(summary_selected), summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "button_press_event",
			   GTK_SIGNAL_FUNC(summary_button_pressed),
			   summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "button_release_event",
			   GTK_SIGNAL_FUNC(summary_button_released),
			   summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "key_press_event",
			   GTK_SIGNAL_FUNC(summary_key_pressed), summaryview);
	gtk_signal_connect(GTK_OBJECT(ctree), "resize_column",
			   GTK_SIGNAL_FUNC(summary_col_resized), summaryview);

	gtk_clist_thaw(GTK_CLIST(ctree));

	/* select first unread message */
	node = summary_find_next_unread_msg(summaryview, NULL);
	if (node == NULL && GTK_CLIST(ctree)->row_list != NULL)
		node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list_end);
	if (node) {
		gtk_ctree_node_moveto(ctree, node, S_COL_MARK, 0.5, 0);
		gtk_widget_grab_focus(GTK_WIDGET(ctree));
		if (prefs_common.open_unread_on_enter)
			summaryview->display_msg = TRUE;
		gtk_ctree_select(ctree, node);
		gtkut_ctree_set_focus_row(ctree, node);
	}

	summary_status_show(summaryview);

	summary_set_popup_sensitive(summaryview);

	main_window_set_toolbar_sensitive
		(summaryview->mainwin, summaryview->selected ? TRUE : FALSE);

	debug_print("\n");
	STATUSBAR_PUSH(summaryview->mainwin, _("done."));

	main_window_cursor_normal(summaryview->mainwin);

	return TRUE;
}

void summary_clear_list(SummaryView *summaryview)
{
	gtk_clist_freeze(GTK_CLIST(summaryview->ctree));

	gtk_ctree_pre_recursive(GTK_CTREE(summaryview->ctree),
				NULL, summary_free_msginfo_func, NULL);

	summaryview->mode = SUMMARY_MAIL;

	g_free(summaryview->cur_folder);
	summaryview->cur_folder   = NULL;
	summaryview->news_session = NULL;

	summaryview->display_msg = FALSE;

	summaryview->selected  = NULL;
	summaryview->displayed = NULL;
	summaryview->newmsgs   = summaryview->unread     = 0;
	summaryview->messages  = summaryview->total_size = 0;
	summaryview->deleted   = summaryview->moved      = 0;
	if (summaryview->msgid_table) {
		g_hash_table_destroy(summaryview->msgid_table);
		summaryview->msgid_table = NULL;
	}
	summaryview->mlist = NULL;
	if (summaryview->folder_table) {
		g_hash_table_destroy(summaryview->folder_table);
		summaryview->folder_table = NULL;
	}
	summaryview->sort_mode = SORT_BY_NONE;
	summaryview->sort_type = GTK_SORT_ASCENDING;

	gtk_clist_clear(GTK_CLIST(summaryview->ctree));

	gtk_clist_thaw(GTK_CLIST(summaryview->ctree));
}

void summary_clear_all(SummaryView *summaryview)
{
	summary_clear_list(summaryview);
	summary_set_popup_sensitive(summaryview);
	main_window_set_toolbar_sensitive(summaryview->mainwin, FALSE);
	summary_status_show(summaryview);
}

static void summary_set_popup_sensitive(SummaryView *summaryview)
{
	GtkItemFactory *ifactory = summaryview->popupfactory;
	GtkCList *clist = GTK_CLIST(summaryview->ctree);
	gboolean sens;

	if (!clist->row_list) {
		GtkWidget *submenu;
		submenu = gtk_item_factory_get_widget
			(summaryview->popupfactory, "/Mark");
		menu_set_insensitive_all(GTK_MENU_SHELL(submenu));
		menu_set_insensitive_all
			(GTK_MENU_SHELL(summaryview->popupmenu));
		return;
	}

	if (path_cmp(summaryview->cur_folder, TRASH_DIR) != 0)
		menu_set_sensitive(ifactory, "/Delete", TRUE);

	menu_set_sensitive(ifactory, "/Collapse",    TRUE);
	menu_set_sensitive(ifactory, "/Move...",    TRUE);
	menu_set_sensitive(ifactory, "/Copy...",    TRUE);
	menu_set_sensitive(ifactory, "/Execute",    TRUE);
        //sens = clist->selection && clist->selection->next ? FALSE : TRUE;
        menu_set_sensitive(ifactory, "/Reply",                TRUE);
        menu_set_sensitive(ifactory, "/Reply with quotation", TRUE);
	menu_set_sensitive(ifactory, "/Forward",              TRUE);
        menu_set_sensitive(ifactory, "/Forward as attachment",TRUE);

        menu_set_sensitive(ifactory, "/Open in new window", TRUE);
        if (path_cmp(summaryview->cur_folder, DRAFT_DIR) == 0 || path_cmp(summaryview->cur_folder, OUTBOX_DIR) == 0)
            menu_set_sensitive(ifactory, "/Reedit", TRUE);


	menu_set_sensitive(ifactory, "/Save as...", TRUE);
	menu_set_sensitive(ifactory, "/Print...",   TRUE);

	menu_set_sensitive(ifactory, "/Mark", TRUE);

	menu_set_sensitive(ifactory, "/Mark/Mark",   TRUE);
	menu_set_sensitive(ifactory, "/Mark/Unmark", TRUE);

        menu_set_sensitive(ifactory, "/Mark/Mark as unread",        TRUE);
        menu_set_sensitive(ifactory, "/Mark/Mark as important",     TRUE);
	menu_set_sensitive(ifactory, "/Mark/Make it as being read", TRUE);

	menu_set_sensitive(ifactory, "/Select all",   TRUE);
}
void summary_view_source(SummaryView * summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;
	gchar *filename;
	FILE * f;

	ViewSourceWindow * viewsrcwin;

	if (!summaryview->selected) return;

	viewsrcwin = viewsource_window_create();

	msginfo = gtk_ctree_node_get_row_data(ctree, summaryview->selected);

	filename = procmsg_get_message_file_path(msginfo);
	if (filename)
	  {
	    f = fopen(filename, "r");
	    if (f)
	      {
		char s[BUFFSIZE];

		while (fgets(s, BUFFSIZE, f))
		  viewsource_window_append(viewsrcwin, s);
		fclose(f);
	      }
	    g_free(filename);
	  }

	viewsource_window_show(viewsrcwin);
}

void summary_select_next_unread(SummaryView *summaryview)
{
	GtkCTreeNode *node;
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCList *clist = GTK_CLIST(summaryview->ctree);

	node = summary_find_next_unread_msg(summaryview,
					    summaryview->selected);

	if (node) {
		gtk_clist_unselect_all(clist);
		gtk_ctree_select(ctree, node);
		gtkut_ctree_set_focus_row(ctree, node);
		if (summaryview->displayed == node)
			summaryview->displayed = NULL;
		summary_display_msg(summaryview, node, FALSE);
	} else {
		AlertValue val;
                if (!prefs_common.change_folder) {
		val = alertpanel(_("No unread message"),
				 _("No unread message found.\n"
				   "Go to next folder?"),
                                 _("Yes"), _("No"), NULL);
                } else val=G_ALERTDEFAULT;
		if (val == G_ALERTDEFAULT) {
			if (gtk_signal_n_emissions_by_name
				(GTK_OBJECT(ctree), "key_press_event") > 0)
					gtk_signal_emit_stop_by_name
						(GTK_OBJECT(ctree),
						 "key_press_event");
			folderview_select_next_unread(summaryview->folderview);
		}
	}
}

static GtkCTreeNode *summary_find_next_unread_msg(SummaryView *summaryview,
						  GtkCTreeNode *current_node)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCTreeNode *node;
	MsgInfo *msginfo;

	if (current_node)
		node = current_node;
		//node = GTK_CTREE_NODE_NEXT(current_node);
	else
		node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list);

	for (; node != NULL; node = GTK_CTREE_NODE_NEXT(node)) {
		msginfo = gtk_ctree_node_get_row_data(ctree, node);
		if (MSG_IS_UNREAD(msginfo->flags)) break;
	}

	return node;
}

#if 0
static GtkCTreeNode *summary_find_prev_unread_msg(SummaryView *summaryview,
						  GtkCTreeNode *current_node)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCTreeNode *node;
	MsgInfo *msginfo;

	if (current_node)
		node = current_node;
		//node = GTK_CTREE_NODE_PREV(current_node);
	else
		node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list_end);

	for (; node != NULL; node = GTK_CTREE_NODE_PREV(node)) {
		msginfo = gtk_ctree_node_get_row_data(ctree, node);
		if (MSG_IS_UNREAD(msginfo->flags)) break;
	}

	return node;
}
#endif

void summary_attract_by_subject(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCList *clist = GTK_CLIST(ctree);
	GtkCTreeNode *src_node;
	GtkCTreeNode *dst_node, *sibling;
	GtkCTreeNode *tmp;
	MsgInfo *src_msginfo, *dst_msginfo;

	debug_print(_("Attracting messages by subject..."));
	STATUSBAR_PUSH(summaryview->mainwin,
		       _("Attracting messages by subject..."));

	main_window_cursor_wait(summaryview->mainwin);
	gtk_clist_freeze(clist);

	for (src_node = GTK_CTREE_NODE(clist->row_list);
	     src_node != NULL;
	     src_node = tmp) {
		tmp = GTK_CTREE_ROW(src_node)->sibling;
		src_msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(src_node);
		if (!src_msginfo) continue;
		if (!src_msginfo->subject) continue;

		/* find attracting node */
		for (dst_node = GTK_CTREE_NODE_PREV(src_node);
		     dst_node != NULL;
		     dst_node = GTK_CTREE_NODE_PREV(dst_node)) {
			/* ignore child node */
			if (GTK_CTREE_ROW(dst_node)->level > 1) continue;

			dst_msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(dst_node);
			if (!dst_msginfo->subject) continue;

			/* if the time difference is more than 30 days,
			   skip the comparison */
			if (ABS(src_msginfo->date_t - dst_msginfo->date_t)
			    > 2592000)
				continue;

			if (!subject_compare(src_msginfo->subject,
					     dst_msginfo->subject)) {
				sibling = GTK_CTREE_ROW(dst_node)->sibling;
				if (src_node != sibling)
					gtk_ctree_move(ctree, src_node,
						       NULL, sibling);
				break;
			}
		}
	}

	gtk_ctree_node_moveto(ctree, summaryview->selected, -1, 0.5, 0);

	gtk_clist_thaw(clist);

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);

	main_window_cursor_normal(summaryview->mainwin);
}

static void summary_free_msginfo_func(GtkCTree *ctree, GtkCTreeNode *node,
				      gpointer data)
{
	MsgInfo *msginfo = gtk_ctree_node_get_row_data(ctree, node);

	if (msginfo)
		procmsg_msginfo_free(msginfo);
}

static void summary_set_marks_func(GtkCTree *ctree, GtkCTreeNode *node,
				   gpointer data)
{
	SummaryView *summaryview = data;
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(ctree, node);

	if (MSG_IS_NEW(msginfo->flags)) {
		gtk_ctree_node_set_pixmap(ctree, node, S_COL_UNREAD,
					  newxpm, newxpmmask);
		summaryview->newmsgs++;
		if (MSG_IS_UNREAD(msginfo->flags))
			summaryview->unread++;
	} else if (MSG_IS_UNREAD(msginfo->flags)) {
		gtk_ctree_node_set_pixmap(ctree, node, S_COL_UNREAD,
					  unreadxpm, unreadxpmmask);
		summaryview->unread++;
	} else if (MSG_IS_IMPORTANT(msginfo->flags)) {
		gtk_ctree_node_set_pixmap(ctree, node, S_COL_UNREAD,
					  importantxpm, importantxpmmask);
        }
        if (MSG_IS_DELETED(msginfo->flags)) {
            gtk_ctree_node_set_text(ctree, node, S_COL_MARK, _("D"));
            gtk_ctree_node_set_foreground(ctree, node, &summaryview->color_dim);
            summaryview->deleted++;
        };
        if (MSG_IS_MOVE(msginfo->flags)) {
            gtk_ctree_node_set_text(ctree, node, S_COL_MARK, _("o"));
            gtk_ctree_node_set_foreground
                (ctree, node, &summaryview->color_marked);
            summaryview->moved++;
        };


	if (MSG_IS_MIME(msginfo->flags))
		gtk_ctree_node_set_pixmap(ctree, node, S_COL_MIME,
					  clipxpm, clipxpmmask);

	summaryview->messages++;
	summaryview->total_size += msginfo->size;
}

static void summary_update_status(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCTreeNode *node;
	MsgInfo *msginfo;

	summaryview->newmsgs = summaryview->unread =
	summaryview->messages = summaryview->total_size =
	summaryview->deleted = summaryview->moved = 0;

	for (node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list);
	     node != NULL; node = GTK_CTREE_NODE_NEXT(node)) {
		msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);

		if (MSG_IS_NEW(msginfo->flags))
			summaryview->newmsgs++;
		if (MSG_IS_UNREAD(msginfo->flags))
			summaryview->unread++;
		if (MSG_IS_DELETED(msginfo->flags))
			summaryview->deleted++;
		if (MSG_IS_MOVE(msginfo->flags))
			summaryview->moved++;
		summaryview->messages++;
		summaryview->total_size += msginfo->size;
	}
}

static void summary_status_show(SummaryView *summaryview)
{
	gchar *buf1, *buf2, *buf3;
	GList *rowlist, *cur;
	guint n_selected = 0;
	off_t sel_size = 0;
	MsgInfo *msginfo;

	if (!summaryview->cur_folder) {
		gtk_label_set(GTK_LABEL(summaryview->statlabel_folder), "");
		gtk_label_set(GTK_LABEL(summaryview->statlabel_select), "");
		gtk_label_set(GTK_LABEL(summaryview->statlabel_msgs),   "");
		return;
	}

	rowlist = GTK_CLIST(summaryview->ctree)->selection;
	for (cur = rowlist; cur != NULL; cur = cur->next) {
		msginfo = gtk_ctree_node_get_row_data
			(GTK_CTREE(summaryview->ctree),
			 GTK_CTREE_NODE(cur->data));
		sel_size += msginfo->size;
		n_selected++;
	}

	gtk_label_set(GTK_LABEL(summaryview->statlabel_folder),
		      summaryview->mode == SUMMARY_NEWS
		      ? g_basename(summaryview->cur_folder)
		      : summaryview->cur_folder);

	if (summaryview->deleted)
		buf1 = g_strdup_printf(_("%d deleted"), summaryview->deleted);
	else
		buf1 = g_strdup("");
	if (summaryview->moved)
		buf2 = g_strdup_printf(_("%d moved"), summaryview->moved);
	else
		buf2 = g_strdup("");

	buf3 = g_strdup_printf("%s%s%s%s",
			       summaryview->deleted || summaryview->moved
						? "    " : "", buf1,
			       summaryview->deleted && summaryview->moved
						? _(", ") : "", buf2);

	g_free(buf1);
	g_free(buf2);
	if (n_selected)
		buf2 = g_strdup_printf(" (%s)", to_human_readable(sel_size));
	else
		buf2 = g_strdup("");
	buf1 = g_strdup_printf("%s%s%s%s",
			      n_selected ? itos(n_selected) : "",
			      n_selected ? _(" item(s) selected") : "",
			      buf2, buf3);
	gtk_label_set(GTK_LABEL(summaryview->statlabel_select), buf1);
	g_free(buf1);
	g_free(buf2);
	g_free(buf3);

	if (summaryview->mode == SUMMARY_MAIL) {
		buf1 = g_strdup_printf(_("%d new, %d unread, %d total (%s)"),
				       summaryview->newmsgs,
				       summaryview->unread,
				       summaryview->messages,
				       to_human_readable(summaryview->total_size));
	} else {
		buf1 = g_strdup_printf(_("%d new, %d unread, %d total"),
				       summaryview->newmsgs,
				       summaryview->unread,
				       summaryview->messages);
	}
	gtk_label_set(GTK_LABEL(summaryview->statlabel_msgs), buf1);
	g_free(buf1);

	folderview_update_msg_num(summaryview->folderview,
				  summaryview->folderview->opened,
				  summaryview->newmsgs,
				  summaryview->unread,
				  summaryview->messages);
}

void summary_sort(SummaryView *summaryview, SummarySortType type)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCList *clist = GTK_CLIST(summaryview->ctree);
	GtkCListCompareFunc cmp_func;

	switch (type) {
	case SORT_BY_NUMBER:
		cmp_func = (GtkCListCompareFunc)summary_cmp_by_num;
		break;
	case SORT_BY_SIZE:
		cmp_func = (GtkCListCompareFunc)summary_cmp_by_size;
		break;
	case SORT_BY_DATE:
		cmp_func = (GtkCListCompareFunc)summary_cmp_by_date;
		break;
	case SORT_BY_FROM:
		cmp_func = (GtkCListCompareFunc)summary_cmp_by_from;
		break;
	case SORT_BY_SUBJECT:
		cmp_func = (GtkCListCompareFunc)summary_cmp_by_subject;
		break;
	default:
		return;
	}

	debug_print(_("Sorting summary..."));
	STATUSBAR_PUSH(summaryview->mainwin, _("Sorting summary..."));

	main_window_cursor_wait(summaryview->mainwin);

	gtk_clist_set_compare_func(clist, cmp_func);

	/* toggle sort type if the same column is selected */
	if (summaryview->sort_mode == type)
		summaryview->sort_type =
			summaryview->sort_type == GTK_SORT_ASCENDING
			? GTK_SORT_DESCENDING : GTK_SORT_ASCENDING;
	else
		summaryview->sort_type = GTK_SORT_ASCENDING;
	gtk_clist_set_sort_type(clist, summaryview->sort_type);
	summaryview->sort_mode = type;

	gtk_ctree_sort_node(ctree, NULL);

	gtk_ctree_node_moveto(ctree, summaryview->selected, -1, 0.5, 0);
	//gtkut_ctree_set_focus_row(ctree, summaryview->selected);

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);

	main_window_cursor_normal(summaryview->mainwin);
}

static void summary_set_ctree_from_list(SummaryView *summaryview,
					GSList *mlist)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;
	GtkCTreeNode *node, *parent;
	gchar *text[N_SUMMARY_COLS];
	GHashTable *msgid_table;
        gint subj_width=0,curr_width;
        GtkStyle *style;
        GtkCTreeRow *row;

	if (!mlist) return;

        style=gtk_widget_get_style(summaryview->ctree);

	debug_print(_("\tSetting summary from message data..."));
	STATUSBAR_PUSH(summaryview->mainwin,
		       _("Setting summary from message data..."));

	msgid_table = g_hash_table_new(g_str_hash, g_str_equal);
	summaryview->msgid_table = msgid_table;

	if (prefs_common.enable_thread) {
		for (; mlist != NULL; mlist = mlist->next) {
			msginfo = (MsgInfo *)mlist->data;
			parent = NULL;

			summary_set_header(text, msginfo);

			/* search parent node for threading */
			if (msginfo->inreplyto && *msginfo->inreplyto)
				parent = g_hash_table_lookup
					(msgid_table, msginfo->inreplyto);

			node = gtk_ctree_insert_node
				(ctree, parent, NULL, text, 2,
				 NULL, NULL, NULL, NULL, FALSE, TRUE);
			GTKUT_CTREE_NODE_SET_ROW_DATA(node, msginfo);
			summary_set_marks_func(ctree, node, summaryview);

			/* preserve previous node if the message is
			   duplicated */
			if (msginfo->msgid && *msginfo->msgid &&
			    g_hash_table_lookup(msgid_table, msginfo->msgid)
			    == NULL)
				g_hash_table_insert(msgid_table,
						    msginfo->msgid, node);

                        if(msginfo->subject) {
                                row=GTK_CTREE_ROW(node);
                                curr_width=gdk_string_width(style->font,msginfo->subject)+row->level*ctree->tree_indent;
                                if(curr_width>subj_width) {
                                        subj_width=curr_width;
                                }
                        }
		}

		/* complete the thread */
		summary_thread_build(summaryview);
	} else {
		for (; mlist != NULL; mlist = mlist->next) {
			msginfo = (MsgInfo *)mlist->data;

                        if(msginfo->subject && (curr_width=gdk_string_width(style->font,msginfo->subject))>subj_width) {
                                subj_width=curr_width;
                        }

			summary_set_header(text, msginfo);

			node = gtk_ctree_insert_node
				(ctree, NULL, NULL, text, 2,
				 NULL, NULL, NULL, NULL, FALSE, TRUE);
			GTKUT_CTREE_NODE_SET_ROW_DATA(node, msginfo);
			summary_set_marks_func(ctree, node, summaryview);

			if (msginfo->msgid && *msginfo->msgid &&
			    g_hash_table_lookup(msgid_table, msginfo->msgid)
			    == NULL)
				g_hash_table_insert(msgid_table,
						    msginfo->msgid, node);
                        if(msginfo->subject) {
                                row=GTK_CTREE_ROW(node);
                                curr_width=gdk_string_width(style->font,msginfo->subject)+row->level*ctree->tree_indent;
                                if(curr_width>subj_width) {
                                        subj_width=curr_width;
                                }
                        }
		}
	}

	gtk_widget_set_usize(summaryview->vbox,
			     prefs_common.summaryview_width+subj_width-prefs_common.summary_col_subject,
			     prefs_common.summaryview_height);
	gtk_clist_set_column_width(GTK_CLIST(summaryview->ctree), S_COL_SUBJECT,
				   subj_width);
	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);
	if (debug_mode)
		debug_print("\tmsgid hash table size = %d\n",
			    g_hash_table_size(msgid_table));
}

struct wcachefp
{
	FILE *cache_fp;
	FILE *mark_fp;
};

gint summary_write_cache(SummaryView *summaryview)
{
	struct wcachefp fps;
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	gint ver = CACHE_VERSION;
	gchar *buf;
	gchar *cachefile, *markfile;

	if (!summaryview->cur_folder) return -1;

	if (summaryview->mode == SUMMARY_NEWS) {
		gchar *path, *server;

		server = g_dirname(summaryview->cur_folder);
		path = g_strconcat(get_news_cache_dir(), G_DIR_SEPARATOR_S,
				   server, NULL);
		g_free(server);
		MAKE_DIR_IF_NOT_EXIST(path);
		g_free(path);

		path = g_strconcat(get_news_cache_dir(), G_DIR_SEPARATOR_S,
				   summaryview->cur_folder, NULL);
		MAKE_DIR_IF_NOT_EXIST(path);
		cachefile = g_strconcat(path, G_DIR_SEPARATOR_S, CACHE_FILE,
					NULL);
		g_free(path);
	} else
		cachefile = g_strconcat(maildir, G_DIR_SEPARATOR_S,
					summaryview->cur_folder,
					G_DIR_SEPARATOR_S,
					CACHE_FILE, NULL);

	if ((fps.cache_fp = fopen(cachefile, "w")) == NULL) {
		FILE_OP_ERROR(cachefile, "fopen");
		g_free(cachefile);
		return -1;
	}
	if (chmod(cachefile, S_IRUSR|S_IWUSR) < 0)
		FILE_OP_ERROR(cachefile, "chmod");
	g_free(cachefile);

	markfile = g_strconcat(summaryview->mode == SUMMARY_MAIL ? maildir :
			       get_news_cache_dir(), G_DIR_SEPARATOR_S,
			       summaryview->cur_folder, G_DIR_SEPARATOR_S,
			       MARK_FILE, NULL);

	if ((fps.mark_fp = fopen(markfile, "w")) == NULL) {
		FILE_OP_ERROR(markfile, "fopen");
		fclose(fps.cache_fp);
		g_free(markfile);
		return -1;
	}
	if (chmod(markfile, S_IRUSR|S_IWUSR) < 0)
		FILE_OP_ERROR(markfile, "chmod");
	g_free(markfile);

	buf = g_strdup_printf(_("Writing summary cache (%s)..."),
			      summaryview->cur_folder);
	debug_print(buf);
	STATUSBAR_PUSH(summaryview->mainwin, buf);
	g_free(buf);

	WRITE_CACHE_DATA_INT(ver, fps.cache_fp);
	ver = MARK_VERSION;
	WRITE_CACHE_DATA_INT(ver, fps.mark_fp);

	gtk_ctree_pre_recursive(ctree, NULL, summary_write_cache_func, &fps);

	fclose(fps.cache_fp);
	fclose(fps.mark_fp);

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);

	return 0;
}

static void summary_write_cache_func(GtkCTree *ctree, GtkCTreeNode *node,
				     gpointer data)
{
	struct wcachefp *fps = data;
	MsgInfo *msginfo = gtk_ctree_node_get_row_data(ctree, node);

	if (msginfo == NULL) return;

	procmsg_write_cache(msginfo, fps->cache_fp);
	procmsg_write_flags(msginfo, fps->mark_fp);
}

static void summary_set_header(gchar *text[], MsgInfo *msginfo)
{
	static gchar date_modified[80];
	static gchar *to = NULL;

	text[S_COL_MARK]   = NULL;
	text[S_COL_UNREAD] = NULL;
	text[S_COL_MIME] = NULL;
	text[S_COL_NUMBER] = itos(msginfo->msgnum);
	text[S_COL_SIZE]   = to_human_readable(msginfo->size);

	if (msginfo->date_t) {
		procheader_date_get_localtime(date_modified,
					      sizeof(date_modified),
					      msginfo->date_t);
		text[S_COL_DATE] = date_modified;
	} else if (msginfo->date)
		text[S_COL_DATE] = msginfo->date;
	else
		text[S_COL_DATE] = _("(No Date)");

	text[S_COL_FROM] = msginfo->fromname ? msginfo->fromname :
		_("(No From)");
	if (prefs_common.swap_from && msginfo->from && msginfo->to &&
	    cur_account && cur_account->address) {
		gchar *from;

		Xalloca(from, strlen(msginfo->from) + 1, return);
		strcpy(from, msginfo->from);
		extract_address(from);
		if (!strcmp(from, cur_account->address)) {
			g_free(to);
			to = g_strconcat("-->", msginfo->to, NULL);
			text[S_COL_FROM] = to;
		}
	}

	text[S_COL_SUBJECT] = msginfo->subject ? msginfo->subject :
		_("(No Subject)");
}

static void summary_display_msg(SummaryView *summaryview, GtkCTreeNode *row,
				gboolean new_window)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;
	gchar *filename;
	static gboolean lock = FALSE;

	if (!new_window && summaryview->displayed == row) return;
	g_return_if_fail(row != NULL);

	if (lock) return;
	lock = TRUE;

	STATUSBAR_POP(summaryview->mainwin);

	msginfo = gtk_ctree_node_get_row_data(ctree, row);

	if (summaryview->mode == SUMMARY_MAIL) {
		filename = g_strconcat(summaryview->cur_folder,
				       G_DIR_SEPARATOR_S,
				       itos(msginfo->msgnum), NULL);
	} else if (summaryview->mode == SUMMARY_NEWS) {
		gint ok;

		filename = g_strconcat(get_news_cache_dir(), G_DIR_SEPARATOR_S,
				       summaryview->cur_folder,
				       G_DIR_SEPARATOR_S,
				       itos(msginfo->msgnum), NULL);
		if (!is_file_exist(filename)) {
			if (!summaryview->news_session) {
				g_warning(_("news session is not established\n"));
				g_free(filename);
				lock = FALSE;
				return;
			}

			debug_print(_("getting article %d...\n"),
				    msginfo->msgnum);
			ok = news_get_article(summaryview->news_session,
					      msginfo->msgnum,
					      filename);
			statusbar_pop_all();
			if (ok < 0) {
				g_warning(_("can't read article %d\n"),
					  msginfo->msgnum);
				g_free(filename);
				lock = FALSE;
				return;
			}
			debug_print(_("done.\n"));
		} else
			debug_print(_("article %d has been already cached.\n"),
				    msginfo->msgnum);
	} else {
		lock = FALSE;
		return;
	}
	g_free(filename);

	if (MSG_IS_NEW(msginfo->flags))
		summaryview->newmsgs--;
	if (MSG_IS_UNREAD(msginfo->flags))
		summaryview->unread--;
	if (MSG_IS_NEW(msginfo->flags) || MSG_IS_UNREAD(msginfo->flags)) {
		MSG_UNSET_FLAGS(msginfo->flags, MSG_NEW | MSG_UNREAD);
		gtk_ctree_node_set_text(ctree, row, S_COL_UNREAD, NULL);
		gtk_clist_thaw(GTK_CLIST(ctree));
		summary_status_show(summaryview);
	}

	if (new_window) {
		MessageView *msgview;

		msgview = messageview_create_with_new_window();
		messageview_show(msgview, msginfo);
	} else {
		summaryview->displayed = row;
		if (!summaryview->msg_is_toggled_on)
			summary_toggle_view(summaryview);
		messageview_show(summaryview->messageview, msginfo);
		GTK_EVENTS_FLUSH();
		if (summaryview->messageview->type == MVIEW_TEXT)
			gtk_widget_grab_focus(summaryview->ctree);
		if (ctree!=NULL && row!=NULL)
		    gtkut_ctree_node_move_if_on_the_edge(ctree, row);
//		    if (!gtk_ctree_node_is_visible(ctree, row))
//			gtk_ctree_node_moveto(ctree, row, S_COL_MARK, 0.5, 0);

	}

	if (GTK_WIDGET_VISIBLE(summaryview->headerwin->window))
		header_window_show(summaryview->headerwin, msginfo);
	lock = FALSE;
}

void summary_redisplay_msg(SummaryView *summaryview)
{
	GtkCTreeNode *node;

	if (summaryview->displayed) {
		node = summaryview->displayed;
		summaryview->displayed = NULL;
		summary_display_msg(summaryview, node, FALSE);
	}
}

void summary_open_msg(SummaryView *summaryview)
{
	if (!summaryview->selected) return;

	summary_display_msg(summaryview, summaryview->selected, TRUE);
}
void summary_reedit(SummaryView *summaryview)
{
	MsgInfo *msginfo;

	if (!summaryview->selected) return;
        if (path_cmp(summaryview->cur_folder, DRAFT_DIR) != 0 &&
            path_cmp(summaryview->cur_folder, OUTBOX_DIR) != 0) return;

	msginfo = gtk_ctree_node_get_row_data(GTK_CTREE(summaryview->ctree),
					      summaryview->selected);
	if (!msginfo) return;

	compose_reedit(msginfo);
}

void summary_step(SummaryView *summaryview, GtkScrollType type)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);

	gtk_signal_emit_by_name(GTK_OBJECT(ctree), "scroll_vertical",
				type, 0.0);

	if (summaryview->msg_is_toggled_on)
		summary_display_msg(summaryview, summaryview->selected, FALSE);
}

static void summary_toggle_view(SummaryView *summaryview)
{
	MainWindow *mainwin = summaryview->mainwin;
	union CompositeWin *cwin = &mainwin->win;
	GtkWidget *vpaned = NULL;
	GtkWidget *container = NULL;

	switch (mainwin->type) {
	case SEPARATE_NONE:
		vpaned = cwin->sep_none.vpaned;
		container = cwin->sep_none.hpaned;
		break;
	case SEPARATE_FOLDER:
		vpaned = cwin->sep_folder.vpaned;
		container = mainwin->vbox_body;
		break;
	case SEPARATE_MESSAGE:
	case SEPARATE_BOTH:
		return;
	}

	if (vpaned->parent != NULL) {
		summaryview->msg_is_toggled_on = FALSE;
		summaryview->displayed = NULL;
		gtk_widget_ref(vpaned);
                gtk_container_remove(GTK_CONTAINER(container), vpaned);
                gtk_widget_reparent(GTK_WIDGET_PTR(summaryview), container);
		gtk_arrow_set(GTK_ARROW(summaryview->toggle_arrow),
			      GTK_ARROW_UP, GTK_SHADOW_OUT);
	} else {
		summaryview->msg_is_toggled_on = TRUE;
		gtk_widget_reparent(GTK_WIDGET_PTR(summaryview), vpaned);
		gtk_container_add(GTK_CONTAINER(container), vpaned);
		gtk_widget_unref(vpaned);
		gtk_arrow_set(GTK_ARROW(summaryview->toggle_arrow),
			      GTK_ARROW_DOWN, GTK_SHADOW_OUT);
	}

	gtk_widget_grab_focus(summaryview->ctree);
}

static void summary_mark_row(SummaryView *summaryview, GtkCTreeNode *row)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(ctree, row);
	g_free(msginfo->to_folder);
	msginfo->to_folder = NULL;
	if (MSG_IS_DELETED(msginfo->flags))
		summaryview->deleted--;
	if (MSG_IS_MOVE(msginfo->flags))
		summaryview->moved--;
	MSG_UNSET_FLAGS(msginfo->flags, MSG_DELETED | MSG_MOVE | MSG_COPY);
	MSG_SET_FLAGS(msginfo->flags, MSG_MARKED);
	gtk_ctree_node_set_text(ctree, row, S_COL_MARK, "*");
	gtk_ctree_node_set_foreground
		(ctree, row, &summaryview->color_marked);
	debug_print(_("Message %d is marked\n"), msginfo->msgnum);
}

void summary_mark(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GList *cur;

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL; cur = cur->next)
		summary_mark_row(summaryview, (GtkCTreeNode *)cur->data);

	summary_step(summaryview, GTK_SCROLL_STEP_FORWARD);
}

void summary_mark_as_read(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GList *cur;
	GtkCTreeNode *row;
	MsgInfo *msginfo;

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL;
	     cur = cur->next) {
		row = cur->data;
		msginfo = gtk_ctree_node_get_row_data(ctree, row);
		if (MSG_IS_NEW(msginfo->flags))
			summaryview->newmsgs--;
		if (MSG_IS_UNREAD(msginfo->flags))
			summaryview->unread--;
		if (MSG_IS_NEW(msginfo->flags)    ||
		    MSG_IS_UNREAD(msginfo->flags) ||
		    MSG_IS_IMPORTANT(msginfo->flags)) {
                    MSG_UNSET_FLAGS(msginfo->flags, MSG_NEW | MSG_UNREAD | MSG_IMPORTANT);
			gtk_ctree_node_set_text(ctree, row, S_COL_UNREAD, NULL);
			debug_print(_("Message %d is marked as being read\n"),
				    msginfo->msgnum);
		}
	}

	summary_status_show(summaryview);
}

void summary_mark_as_unread(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GList *cur;
	GtkCTreeNode *row;
	MsgInfo *msginfo;

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL;
	     cur = cur->next) {
		row = cur->data;
		msginfo = gtk_ctree_node_get_row_data(ctree, row);
		if (MSG_IS_DELETED(msginfo->flags)) {
			g_free(msginfo->to_folder);
			msginfo->to_folder = NULL;
			MSG_UNSET_FLAGS(msginfo->flags, MSG_DELETED);
			summaryview->deleted--;
		}
		if (!MSG_IS_UNREAD(msginfo->flags)) {
                        MSG_SET_FLAGS(msginfo->flags, MSG_UNREAD);
			MSG_UNSET_FLAGS(msginfo->flags, MSG_IMPORTANT);
			gtk_ctree_node_set_pixmap(ctree, row, S_COL_UNREAD,
						  unreadxpm, unreadxpmmask);
			summaryview->unread++;
			gtk_ctree_node_set_foreground
				(ctree, row, &summaryview->color_normal);
			debug_print(_("Message %d is marked as unread\n"),
				    msginfo->msgnum);
		}
	}

	summary_status_show(summaryview);
}

void summary_mark_as_important(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GList *cur;
	GtkCTreeNode *row;
	MsgInfo *msginfo;

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL;
	     cur = cur->next) {
		row = cur->data;
		msginfo = gtk_ctree_node_get_row_data(ctree, row);
		if (MSG_IS_DELETED(msginfo->flags)) {
			g_free(msginfo->to_folder);
			msginfo->to_folder = NULL;
			MSG_UNSET_FLAGS(msginfo->flags, MSG_DELETED);
			summaryview->deleted--;
		}
		if (!MSG_IS_IMPORTANT(msginfo->flags)) {
			MSG_SET_FLAGS(msginfo->flags, MSG_IMPORTANT);
			gtk_ctree_node_set_pixmap(ctree, row, S_COL_UNREAD,
						  importantxpm, importantxpmmask);
			gtk_ctree_node_set_foreground
				(ctree, row, &summaryview->color_normal);
			debug_print(_("Message %d is marked as important\n"),
				    msginfo->msgnum);
		}
	}

	summary_status_show(summaryview);
}



static void summary_delete_row(SummaryView *summaryview, GtkCTreeNode *row)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(ctree, row);

	if (MSG_IS_DELETED(msginfo->flags)) return;

	g_free(msginfo->to_folder);
	msginfo->to_folder = NULL;
	if (MSG_IS_MOVE(msginfo->flags))
		summaryview->moved--;
	MSG_UNSET_FLAGS(msginfo->flags,
			MSG_MARKED |
			MSG_MOVE |
			MSG_COPY);
	MSG_SET_FLAGS(msginfo->flags, MSG_DELETED);
	summaryview->deleted++;

	gtk_ctree_node_set_text(ctree, row, S_COL_MARK, _("D"));
	gtk_ctree_node_set_foreground(ctree, row, &summaryview->color_dim);
	debug_print(_("Message %s/%d is set to delete\n"),
		    msginfo->folder, msginfo->msgnum);
}

void summary_delete(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GList *cur;

	/* if current folder is trash, don't delete */
	if (!path_cmp(summaryview->cur_folder, TRASH_DIR)) {
		alertpanel_notice(_("Current folder is Trash."));
		return;
	}

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL; cur = cur->next)
		summary_delete_row(summaryview, GTK_CTREE_NODE(cur->data));

	summary_step(summaryview, GTK_SCROLL_STEP_FORWARD);
        if(prefs_common.no_execute) {
                summary_execute(summaryview);
        } else {
	        summary_status_show(summaryview);
        }
}

void summary_delete_duplicated(SummaryView *summaryview)
{
	if (!path_cmp(summaryview->cur_folder, TRASH_DIR)) return;

	main_window_cursor_wait(summaryview->mainwin);
	debug_print(_("Deleting duplicated messages..."));
	STATUSBAR_PUSH(summaryview->mainwin,
		       _("Deleting duplicated messages..."));

	gtk_ctree_pre_recursive(GTK_CTREE(summaryview->ctree), NULL,
				GTK_CTREE_FUNC(summary_delete_duplicated_func),
				summaryview);

	summary_status_show(summaryview);

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);
	main_window_cursor_normal(summaryview->mainwin);
}

static void summary_delete_duplicated_func(GtkCTree *ctree, GtkCTreeNode *node,
					   SummaryView *summaryview)
{
	GtkCTreeNode *found;
	MsgInfo *msginfo = GTK_CTREE_ROW(node)->row.data;

	if (!msginfo->msgid || !*msginfo->msgid) return;

	found = g_hash_table_lookup(summaryview->msgid_table, msginfo->msgid);

	if (found && found != node)
		summary_delete_row(summaryview, node);
}

static void summary_unmark_row(SummaryView *summaryview, GtkCTreeNode *row)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(ctree, row);
	g_free(msginfo->to_folder);
	msginfo->to_folder = NULL;
	if (MSG_IS_DELETED(msginfo->flags))
		summaryview->deleted--;
	if (MSG_IS_MOVE(msginfo->flags))
		summaryview->moved--;
	MSG_UNSET_FLAGS(msginfo->flags,
			MSG_MARKED |
			MSG_DELETED |
			MSG_IMPORTANT |
			MSG_MOVE |
			MSG_COPY);
	gtk_ctree_node_set_text(ctree, row, S_COL_MARK, NULL);
	gtk_ctree_node_set_foreground
		(ctree, row, &summaryview->color_normal);
	debug_print(_("Message %s/%d is unmarked\n"),
		    msginfo->folder, msginfo->msgnum);
}

void summary_unmark(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GList *cur;

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL;
	     cur = cur->next)
		summary_unmark_row(summaryview, GTK_CTREE_NODE(cur->data));

	summary_status_show(summaryview);
}

static void summary_move_row_to(SummaryView *summaryview, GtkCTreeNode *row,
				const gchar *to_folder)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(ctree, row);
	g_free(msginfo->to_folder);
	msginfo->to_folder = g_strdup(to_folder);
	if (MSG_IS_DELETED(msginfo->flags))
		summaryview->deleted--;
	if (MSG_IS_COPY(msginfo->flags))
		summaryview->copied--;
	MSG_UNSET_FLAGS(msginfo->flags,
			MSG_MARKED | MSG_DELETED | MSG_COPY);
	if (!MSG_IS_MOVE(msginfo->flags)) {
		MSG_SET_FLAGS(msginfo->flags, MSG_MOVE);
		summaryview->moved++;
	}
	gtk_ctree_node_set_text(ctree, row, S_COL_MARK, _("o"));
	gtk_ctree_node_set_foreground
		(ctree, row, &summaryview->color_marked);
	debug_print(_("Message %d is set to move to %s\n"),
		    msginfo->msgnum, to_folder);
}

void summary_move_to(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	gchar *to_folder;
	GList *cur;

	to_folder = foldersel_folder_sel(NULL);

	if (!to_folder) return;
	if (!path_cmp(summaryview->cur_folder, to_folder)) {
		alertpanel_notice(_("Destination is same as current folder."));
		return;
	}

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL; cur = cur->next)
		summary_move_row_to
			(summaryview, (GtkCTreeNode *)cur->data, to_folder);

	summary_step(summaryview, GTK_SCROLL_STEP_FORWARD);
        if(prefs_common.no_execute) {
                summary_execute(summaryview);
        } else {
	        summary_status_show(summaryview);
        }
}

static void summary_copy_row_to(SummaryView *summaryview, GtkCTreeNode *row,
				const gchar *to_folder)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(ctree, row);
	g_free(msginfo->to_folder);
	msginfo->to_folder = g_strdup(to_folder);
	if (MSG_IS_DELETED(msginfo->flags))
		summaryview->deleted--;
	if (MSG_IS_MOVE(msginfo->flags))
		summaryview->moved--;
	MSG_UNSET_FLAGS(msginfo->flags,
			MSG_MARKED | MSG_DELETED | MSG_MOVE);
	if (!MSG_IS_COPY(msginfo->flags)) {
		MSG_SET_FLAGS(msginfo->flags, MSG_COPY);
		summaryview->copied++;
	}
	gtk_ctree_node_set_text(ctree, row, S_COL_MARK, _("c"));
	gtk_ctree_node_set_foreground
		(ctree, row, &summaryview->color_marked);
	debug_print(_("Message %d is set to copy to %s\n"),
		    msginfo->msgnum, to_folder);
}

void summary_copy_to(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	gchar *to_folder;
	GList *cur;

	to_folder = foldersel_folder_sel(NULL);

	if (!to_folder) return;
	if (!path_cmp(summaryview->cur_folder, to_folder)) {
		alertpanel_notice(_("Destination is same as current folder."));
		return;
	}

	for (cur = GTK_CLIST(ctree)->selection; cur != NULL; cur = cur->next)
		summary_copy_row_to
			(summaryview, (GtkCTreeNode *)cur->data, to_folder);

	summary_step(summaryview, GTK_SCROLL_STEP_FORWARD);
        if(prefs_common.no_execute) {
                summary_execute(summaryview);
        } else {
	        summary_status_show(summaryview);
        }
}

void collapse_thread(SummaryView *summaryview)
{
	gtk_ctree_toggle_expansion(GTK_CTREE(summaryview->ctree),summaryview->selected);
}

void summary_select_all(SummaryView *summaryview)
{
	gtk_clist_select_all(GTK_CLIST(summaryview->ctree));
}

void summary_unselect_all(SummaryView *summaryview)
{
	gtk_clist_unselect_all(GTK_CLIST(summaryview->ctree));
}

void summary_save_as(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;
	gchar *src, *dest;
        GList *destlist=NULL;

	if (!summaryview->selected) return;
	msginfo = gtk_ctree_node_get_row_data(ctree, summaryview->selected);
	if (!msginfo) return;

	destlist = filesel_select_file(_("Save as"), NULL, FALSE);
	if (!destlist) return;
        dest=(gchar *)destlist->data;
	if (is_file_exist(dest)) {
		AlertValue aval;

		aval = alertpanel(_("Overwrite"),
				  _("Overwrite existing file?"),
				  _("OK"), _("Cancel"), NULL);
		if (G_ALERTDEFAULT != aval) return;
	}

	src = g_strdup_printf("%s%c%s%c%d",
			      MSG_IS_NEWS(msginfo->flags)
			      ? get_news_cache_dir() : maildir,
			      G_DIR_SEPARATOR,
			      msginfo->folder, G_DIR_SEPARATOR,
			      msginfo->msgnum);
	copy_file(src, dest);
	g_free(src);
}

void summary_print(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCList *clist = GTK_CLIST(summaryview->ctree);
	MsgInfo *msginfo;
	GList *cur;
	gchar *cmdline;
	gchar *p;

	if (clist->selection == NULL) return;

	cmdline = input_dialog(_("Print"),
			       _("Enter the print command line:\n"
				 "(`%s' will be replaced with file name)"),
			       prefs_common.print_cmd);
	if (!cmdline) return;
	if (!(p = strchr(cmdline, '%')) || *(p + 1) != 's' ||
	    strchr(p + 2, '%')) {
		alertpanel_error(_("Print command line is invalid:\n`%s'"),
				 cmdline);
		return;
	}

	for (cur = clist->selection; cur != NULL; cur = cur->next) {
		msginfo = gtk_ctree_node_get_row_data
			(ctree, GTK_CTREE_NODE(cur->data));
		if (msginfo) procmsg_print_message(msginfo, cmdline);
	}
}

void summary_execute(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCList *clist = GTK_CLIST(summaryview->ctree);
	GtkCTreeNode *node, *next;

	gtk_clist_freeze(clist);

	if (prefs_common.enable_thread)
		summary_unthread_for_exec(summaryview);

	summary_execute_copy(summaryview);
	summary_execute_move(summaryview);
	summary_execute_delete(summaryview);

	node = GTK_CTREE_NODE(clist->row_list);
	while (node != NULL) {
		next = GTK_CTREE_NODE_NEXT(node);
		if (gtk_ctree_node_get_row_data(ctree, node) == NULL) {
			if (node == summaryview->displayed) {
				messageview_clear(summaryview->messageview);
				summaryview->displayed = NULL;
			}
			if (GTK_CTREE_ROW(node)->children != NULL)
				g_warning("summary_execute(): children != NULL\n");
			else
				gtk_ctree_remove_node(ctree, node);
		}
		node = next;
	}

	if (prefs_common.enable_thread)
		summary_thread_build(summaryview);

	summaryview->selected = clist->selection ?
		GTK_CTREE_NODE(clist->selection->data) : NULL;

	if (!GTK_CLIST(summaryview->ctree)->row_list) {
		menu_set_insensitive_all
			(GTK_MENU_SHELL(summaryview->popupmenu));
		gtk_widget_grab_focus(summaryview->folderview->ctree);
	} else
		gtk_widget_grab_focus(summaryview->ctree);

	summary_update_status(summaryview);
	summary_status_show(summaryview);

	summary_write_cache(summaryview);

	gtk_ctree_node_moveto(ctree, summaryview->selected, -1, 0.5, 0);

	gtk_clist_thaw(clist);
}

static void summary_execute_move(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GSList *cur;

	summaryview->folder_table = g_hash_table_new(g_str_hash, g_str_equal);

	/* search moving messages and execute */
	gtk_ctree_pre_recursive(ctree, NULL, summary_execute_move_func,
				summaryview);

	if (summaryview->mlist) {
		procmsg_move_messages(summaryview->cur_folder,
				      summaryview->mlist);

		folderview_scan_folder_foreach_a(summaryview->folder_table);
		for (cur = summaryview->mlist; cur != NULL; cur = cur->next)
			procmsg_msginfo_free((MsgInfo *)cur->data);

		g_slist_free(summaryview->mlist);
		summaryview->mlist = NULL;
	}

	g_hash_table_destroy(summaryview->folder_table);
	summaryview->folder_table = NULL;
}

static void summary_execute_move_func(GtkCTree *ctree, GtkCTreeNode *node,
				      gpointer data)
{
	SummaryView *summaryview = data;
	MsgInfo *msginfo;

	msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);

	if (msginfo && MSG_IS_MOVE(msginfo->flags) && msginfo->to_folder) {
		gint val;

		if ((val = GPOINTER_TO_INT(g_hash_table_lookup
					   (summaryview->folder_table,
					    msginfo->to_folder))))
			g_hash_table_insert(summaryview->folder_table,
					    msginfo->to_folder,
					    GINT_TO_POINTER(val + 1));
		else
			g_hash_table_insert(summaryview->folder_table,
					    g_strdup(msginfo->to_folder),
					    GINT_TO_POINTER(1));

		summaryview->mlist =
			g_slist_append(summaryview->mlist, msginfo);
		gtk_ctree_node_set_row_data(ctree, node, NULL);

		if (msginfo->msgid && *msginfo->msgid &&
		    node == g_hash_table_lookup(summaryview->msgid_table,
						msginfo->msgid))
			g_hash_table_remove(summaryview->msgid_table,
					    msginfo->msgid);
	}
}

static void summary_execute_delete(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GSList *cur;

	if (!path_cmp(summaryview->cur_folder, TRASH_DIR)) return;

	/* search deleting messages and execute */
	gtk_ctree_pre_recursive
		(ctree, NULL, summary_execute_delete_func, summaryview);

	if (!summaryview->mlist) return;

	procmsg_move_messages_with_dest(TRASH_DIR, summaryview->cur_folder,
					summaryview->mlist);
        for (cur = summaryview->mlist; cur != NULL; cur = cur->next)
		procmsg_msginfo_free((MsgInfo *)cur->data);

	g_slist_free(summaryview->mlist);
	summaryview->mlist = NULL;

	folderview_scan_folder_a(TRASH_DIR, FALSE);
}

static void summary_execute_delete_func(GtkCTree *ctree, GtkCTreeNode *node,
					gpointer data)
{
	SummaryView *summaryview = data;
	MsgInfo *msginfo;

	msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);


	if (msginfo && MSG_IS_DELETED(msginfo->flags)) {
		summaryview->mlist =
			g_slist_append(summaryview->mlist, msginfo);
		gtk_ctree_node_set_row_data(ctree, node, NULL);

		if (msginfo->msgid && *msginfo->msgid &&
		    node == g_hash_table_lookup(summaryview->msgid_table,
                                                msginfo->msgid)) {
                    g_hash_table_remove(summaryview->msgid_table,
                                        msginfo->msgid);
                };
        }
}

static void summary_execute_copy(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GSList *cur;

	summaryview->folder_table = g_hash_table_new(g_str_hash, g_str_equal);

	/* search copying messages and execute */
	gtk_ctree_pre_recursive(ctree, NULL, summary_execute_copy_func,
				summaryview);

	if (summaryview->mlist) {
		procmsg_copy_messages(summaryview->cur_folder,
				      summaryview->mlist);

		folderview_scan_folder_foreach_a(summaryview->folder_table);
//		for (cur = summaryview->mlist; cur != NULL; cur = cur->next)
//			procmsg_msginfo_free((MsgInfo *)cur->data);

		g_slist_free(summaryview->mlist);
		summaryview->mlist = NULL;
	}

	g_hash_table_destroy(summaryview->folder_table);
	summaryview->folder_table = NULL;
}

static void summary_execute_copy_func(GtkCTree *ctree, GtkCTreeNode *node,
				      gpointer data)
{
	SummaryView *summaryview = data;
	MsgInfo *msginfo;

	msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);

	if (msginfo && MSG_IS_COPY(msginfo->flags) && msginfo->to_folder) {
		gint val;

		if ((val = GPOINTER_TO_INT(g_hash_table_lookup
					   (summaryview->folder_table,
					    msginfo->to_folder))))
			g_hash_table_insert(summaryview->folder_table,
					    msginfo->to_folder,
					    GINT_TO_POINTER(val + 1));
		else
			g_hash_table_insert(summaryview->folder_table,
					    g_strdup(msginfo->to_folder),
					    GINT_TO_POINTER(1));

		summaryview->mlist =
			g_slist_append(summaryview->mlist, msginfo);
                MSG_UNSET_FLAGS(msginfo->flags,MSG_COPY);
                gtk_ctree_node_set_text(GTK_CTREE(summaryview->ctree), node, S_COL_MARK, NULL);
                gtk_ctree_node_set_foreground(GTK_CTREE(summaryview->ctree), node, &summaryview->color_normal);
//		gtk_ctree_node_set_row_data(ctree, node, NULL);

//		if (msginfo->msgid && *msginfo->msgid &&
//		    node == g_hash_table_lookup(summaryview->msgid_table,
//						msginfo->msgid))
//			g_hash_table_remove(summaryview->msgid_table,
//					    msginfo->msgid);
	}
}

/* thread functions */

void summary_thread_build(SummaryView *summaryview)
{
	debug_print(_("Building threads..."));
	STATUSBAR_PUSH(summaryview->mainwin, _("Building threads..."));
	main_window_cursor_wait(summaryview->mainwin);

	gtk_clist_freeze(GTK_CLIST(summaryview->ctree));

	gtk_ctree_pre_recursive_to_depth
		(GTK_CTREE(summaryview->ctree), NULL, 1,
		 GTK_CTREE_FUNC(summary_thread_func),
		 summaryview->msgid_table);

	gtk_clist_thaw(GTK_CLIST(summaryview->ctree));

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);
	main_window_cursor_normal(summaryview->mainwin);
}

void summary_unthread(SummaryView *summaryview)
{
	GtkCTreeNode *node;
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);

	debug_print(_("Unthreading..."));
	STATUSBAR_PUSH(summaryview->mainwin, _("Unthreading..."));
	main_window_cursor_wait(summaryview->mainwin);

	gtk_clist_freeze(GTK_CLIST(ctree));

	for (node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list);
	     node != NULL; node = GTK_CTREE_NODE_NEXT(node)) {
		summary_unthread_func(ctree, node, NULL);
	}

	gtk_clist_thaw(GTK_CLIST(ctree));

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);
	main_window_cursor_normal(summaryview->mainwin);
}

static void summary_unthread_for_exec(SummaryView *summaryview)
{
	GtkCTreeNode *node;
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);

	debug_print(_("Unthreading for execution..."));

	gtk_clist_freeze(GTK_CLIST(ctree));

	for (node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list);
	     node != NULL; node = GTK_CTREE_NODE_NEXT(node)) {
		summary_unthread_for_exec_func(ctree, node, NULL);
	}

	gtk_clist_thaw(GTK_CLIST(ctree));

	debug_print(_("done.\n"));
}

static void summary_thread_func(GtkCTree *ctree, GtkCTreeNode *node,
				gpointer data)
{
	MsgInfo *msginfo;
	GtkCTreeNode *parent;
	GHashTable *msgid_table = (GHashTable *)data;

	msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);

	if (!msginfo || !msginfo->inreplyto) return;

	parent = g_hash_table_lookup(msgid_table, msginfo->inreplyto);

	if (parent && parent != node) {
		gtk_ctree_move(ctree, node, parent, NULL);
		gtk_ctree_expand(ctree, parent);
	}
}

static void summary_unthread_func(GtkCTree *ctree, GtkCTreeNode *node,
				  gpointer data)
{
	GtkCTreeNode *child;
	GtkCTreeNode *sibling;

	child = GTK_CTREE_ROW(node)->children;
	sibling = GTK_CTREE_ROW(node)->sibling;

	while (child != NULL) {
		GtkCTreeNode *next_child;

		next_child = GTK_CTREE_ROW(child)->sibling;
		gtk_ctree_move(ctree, child, NULL, sibling);
		child = next_child;
	}
}

static void summary_unthread_for_exec_func(GtkCTree *ctree, GtkCTreeNode *node,
					   gpointer data)
{
	MsgInfo *msginfo;
	GtkCTreeNode *top_parent;
	GtkCTreeNode *child;
	GtkCTreeNode *sibling;

	msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);

	if (!msginfo ||
	    (!MSG_IS_MOVE(msginfo->flags) && !MSG_IS_DELETED(msginfo->flags)))
		return;
	child = GTK_CTREE_ROW(node)->children;
	if (!child) return;

	for (top_parent = node;
	     GTK_CTREE_ROW(top_parent)->parent != NULL;
	     top_parent = GTK_CTREE_ROW(top_parent)->parent)
		;
	sibling = GTK_CTREE_ROW(top_parent)->sibling;

	while (child != NULL) {
		GtkCTreeNode *next_child;

		next_child = GTK_CTREE_ROW(child)->sibling;
		gtk_ctree_move(ctree, child, NULL, sibling);
		child = next_child;
	}
}

void summary_filter(SummaryView *summaryview)
{
	if (!prefs_common.fltlist) return;

	debug_print(_("filtering..."));
	STATUSBAR_PUSH(summaryview->mainwin, _("Filtering..."));
	main_window_cursor_wait(summaryview->mainwin);

	gtk_clist_freeze(GTK_CLIST(summaryview->ctree));

	gtk_ctree_pre_recursive(GTK_CTREE(summaryview->ctree), NULL,
				GTK_CTREE_FUNC(summary_filter_func),
				summaryview);

	gtk_clist_thaw(GTK_CLIST(summaryview->ctree));

	summary_status_show(summaryview);

	debug_print(_("done.\n"));
	STATUSBAR_POP(summaryview->mainwin);
	main_window_cursor_normal(summaryview->mainwin);
}

static void summary_filter_func(GtkCTree *ctree, GtkCTreeNode *node,
				gpointer data)
{
	MsgInfo *msginfo = GTKUT_CTREE_NODE_GET_ROW_DATA(node);
	SummaryView *summaryview = data;
	gchar *file;
	gchar *dest;

	file = g_strdup_printf("%s%c%d", msginfo->folder, G_DIR_SEPARATOR,
			       msginfo->msgnum);
	dest = filter_get_dest_folder(prefs_common.fltlist, file);
        if(dest && !is_dir_exist(dest)) {
                debug_print(_("Folder %s not exist!\n"),dest);
                return;
        }
	g_free(file);

	if (dest && strcmp(dest, FILTER_NOT_RECEIVE) &&
	    path_cmp(summaryview->cur_folder, dest) != 0)
		summary_move_row_to(summaryview, node, dest);
}

/* callback functions */

static void summary_toggle_pressed(GtkWidget *eventbox, GdkEventButton *event,
				   SummaryView *summaryview)
{
	if (!event)
		return;

	if (!summaryview->msg_is_toggled_on && summaryview->selected)
		summary_display_msg(summaryview, summaryview->selected, FALSE);
	else
		summary_toggle_view(summaryview);
}

static void summary_button_pressed(GtkWidget *ctree, GdkEventButton *event,
				   SummaryView *summaryview)
{
	GtkCList *clist = GTK_CLIST(ctree);
	gint row, column;

	if (!event) return;

	if (event->button == 3) {
		/* right clicked */
		if (gtk_clist_get_selection_info(clist, event->x, event->y,
						 &row, &column)) {
			GtkCListRow *clist_row;

			clist_row = g_list_nth(clist->row_list, row)->data;
			if (clist_row->state != GTK_STATE_SELECTED) {
				gtk_clist_unselect_all(clist);
				gtk_clist_select_row(clist, row, column);
			}
			gtkut_clist_set_focus_row(clist, row);
		}
		gtk_menu_popup(GTK_MENU(summaryview->popupmenu), NULL, NULL,
			       NULL, NULL, event->button, event->time);
	} else if (event->button == 2) {
		/* center clicked */
		gtk_widget_grab_focus(ctree);
		gtk_clist_unselect_all(clist);
		if (gtk_clist_get_selection_info(clist, event->x, event->y,
						 &row, &column)) {
			gtk_clist_select_row(clist, row, column);
			gtkut_clist_set_focus_row(clist, row);
                        if(summaryview->displayed == summaryview->selected) {
                            textview_scroll_page(summaryview->messageview->textview, FALSE);
                        } else {
                            if (clist->selection)
                                summary_display_msg(summaryview,
                                                    GTK_CTREE_NODE(clist->selection->data),
                                                    FALSE);
                        }
		}
	} else if (event->button == 1) {
		if (!prefs_common.emulate_emacs)
			summaryview->display_msg = TRUE;
	}
}

static void summary_button_released(GtkWidget *ctree, GdkEventButton *event,
				    SummaryView *summaryview)
{
}

void summary_pass_key_press_event(SummaryView *summaryview, GdkEventKey *event)
{
	summary_key_pressed(summaryview->ctree, event, summaryview);
}

#define BREAK_ON_MODIFIER_KEY() \
	if ((event->state & (GDK_MOD1_MASK|GDK_CONTROL_MASK)) != 0) break

static void summary_key_pressed(GtkWidget *widget, GdkEventKey *event,
				SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(widget);
	GtkCTreeNode *node;
	gchar *to_folder;

	if (!event) return;

	switch (event->keyval) {
	case GDK_g:		/* Go */
	case GDK_G:
		BREAK_ON_MODIFIER_KEY();

		if (gtk_signal_n_emissions_by_name
			(GTK_OBJECT(ctree), "key_press_event") > 0)
			gtk_signal_emit_stop_by_name(GTK_OBJECT(ctree),
						     "key_press_event");

		to_folder = foldersel_folder_sel(NULL);
		if (to_folder) {
			debug_print(_("Go to %s\n"), to_folder);
			folderview_select(summaryview->folderview, to_folder);
		}
		return;
	case GDK_w:		/* Write new message */
		BREAK_ON_MODIFIER_KEY();
		compose_new();
		return;
	case GDK_D:		/* Clean trash */
		BREAK_ON_MODIFIER_KEY();
		if (gtk_signal_n_emissions_by_name
			(GTK_OBJECT(ctree), "key_press_event") > 0)
			gtk_signal_emit_stop_by_name(GTK_OBJECT(ctree),
						     "key_press_event");
		main_window_clean_trash(summaryview->mainwin, TRUE);
		return;
	case GDK_Q:		/* Quit */
		BREAK_ON_MODIFIER_KEY();

		if (prefs_common.confirm_on_exit) {
			if (alertpanel(_("Exit"), _("Exit this program?"),
				       _("OK"), _("Cancel"), NULL)
				       == G_ALERTDEFAULT) {
				app_will_exit(NULL, summaryview->mainwin);
			}
		}
		return;
        case GDK_Escape: /* Move focus */
		gtk_widget_grab_focus(summaryview->folderview->ctree);
		return;
	default:
	}

	if (!summaryview->selected) {
		node = gtk_ctree_node_nth(ctree, 0);
		if (node)
			gtk_ctree_select(ctree, node);
		else
			return;
	}

	switch (event->keyval) {
	case GDK_space:		/* Page down or go to the next */
		if (summaryview->displayed != summaryview->selected) {
			summary_display_msg(summaryview,
					    summaryview->selected, FALSE);
			break;
		}
		if (!textview_scroll_page(summaryview->messageview->textview,
					  FALSE))
			summary_select_next_unread(summaryview);
		break;
	case GDK_n:		/* Next */
	case GDK_N:
		BREAK_ON_MODIFIER_KEY();
		summary_step(summaryview, GTK_SCROLL_STEP_FORWARD);
		break;
	case GDK_BackSpace:	/* Page up */
		textview_scroll_page(summaryview->messageview->textview, TRUE);
		break;
	case GDK_p:		/* Prev */
	case GDK_P:
		BREAK_ON_MODIFIER_KEY();
		summary_step(summaryview, GTK_SCROLL_STEP_BACKWARD);
		break;
	case GDK_v:		/* Toggle summary mode / message mode */
	case GDK_V:
		BREAK_ON_MODIFIER_KEY();

		if (!summaryview->msg_is_toggled_on && summaryview->selected)
			summary_display_msg(summaryview,
					    summaryview->selected, FALSE);
		else
			summary_toggle_view(summaryview);
		break;
        case GDK_Return:	/* Scroll up/down one line */
        case GDK_KP_Enter:
		if (summaryview->displayed != summaryview->selected) {
			summary_display_msg(summaryview,
					    summaryview->selected, FALSE);
			break;
		}
		textview_scroll_one_line(summaryview->messageview->textview,
					 (event->state & GDK_MOD1_MASK) != 0);
		break;
	case GDK_asterisk:	/* Mark */
		summary_mark(summaryview);
		break;
	case GDK_exclam:	/* Mark as unread */
		summary_mark_as_unread(summaryview);
		break;
	case GDK_d:		/* Delete */
	case GDK_Delete:
		BREAK_ON_MODIFIER_KEY();
		summary_delete(summaryview);
		break;
	case GDK_u:		/* Unmark */
	case GDK_U:
		BREAK_ON_MODIFIER_KEY();
		summary_unmark(summaryview);
		break;
	case GDK_o:		/* Move */
	case GDK_O:
		BREAK_ON_MODIFIER_KEY();
		summary_move_to(summaryview);
		break;
	case GDK_c:		/* Copy */
	case GDK_C:
		BREAK_ON_MODIFIER_KEY();
		summary_copy_to(summaryview);
		break;
	case GDK_x:		/* Execute */
	case GDK_X:
		BREAK_ON_MODIFIER_KEY();
		summary_execute(summaryview);
		break;
	case GDK_a:		/* Reply to the message */
		BREAK_ON_MODIFIER_KEY();
		summary_reply_cb(summaryview, 0, NULL);
		break;
	case GDK_A:		/* Reply to the message with quotation */
		BREAK_ON_MODIFIER_KEY();
		summary_reply_cb(summaryview, 1, NULL);
		break;
	case GDK_f:		/* Forward the message */
	case GDK_F:
		BREAK_ON_MODIFIER_KEY();
		summary_reply_cb(summaryview, 2, NULL);
		break;
	case GDK_y:		/* Save the message */
		BREAK_ON_MODIFIER_KEY();
		summary_save_as(summaryview);
		break;
	default:
	}
}

static void summary_row_selected(GtkCList *clist, gint row, gint column,
				 GdkEvent *event, gpointer data)
{
	if (event && event->type == GDK_2BUTTON_PRESS) {
		SummaryView *summaryview = data;
		GtkCTreeNode *node;

		node = gtk_ctree_node_nth(GTK_CTREE(clist), row);
		if (path_cmp(summaryview->cur_folder, DRAFT_DIR) == 0)
			summary_reedit(summaryview);
		else
			summary_display_msg(summaryview, node, TRUE);
	}
}

static void summary_selected(GtkCTree *ctree, GtkCTreeNode *row,
			     gint column, SummaryView *summaryview)
{
	summary_status_show(summaryview);
	summary_set_popup_sensitive(summaryview);

	if (GTK_CLIST(ctree)->selection &&
	     GTK_CLIST(ctree)->selection->next) {
		summaryview->display_msg = FALSE;
		return;
	}

	summaryview->selected = row;

	if (summaryview->msg_is_toggled_on && summaryview->display_msg)
		summary_display_msg(summaryview, row, FALSE);

	summaryview->display_msg = FALSE;
}

static void summary_col_resized(GtkCList *clist, gint column, gint width,
				SummaryView *summaryview)
{
	switch (column) {
	case S_COL_MARK:
		prefs_common.summary_col_mark = width;
		break;
	case S_COL_UNREAD:
		prefs_common.summary_col_unread = width;
		break;
	case S_COL_MIME:
		prefs_common.summary_col_mime = width;
		break;
	case S_COL_NUMBER:
		prefs_common.summary_col_number = width;
		break;
	case S_COL_SIZE:
		prefs_common.summary_col_size = width;
		break;
	case S_COL_DATE:
		prefs_common.summary_col_date = width;
		break;
	case S_COL_FROM:
		prefs_common.summary_col_from = width;
		break;
	case S_COL_SUBJECT:
		prefs_common.summary_col_subject = width;
		break;
	default:
	}
}

static void summary_reply_cb(SummaryView *summaryview, guint action,
			     GtkWidget *widget)
{
	MsgInfo *msginfo;

	msginfo = gtk_ctree_node_get_row_data(GTK_CTREE(summaryview->ctree),
					      summaryview->selected);
	if (!msginfo) return;

	//GTKUT_CTREE_NODE_GET_ROW_DATA(src_node)

	switch (action) {
	case 0:
		compose_reply(msginfo);
		break;
	case 1:
		compose_quote(msginfo);
		break;
	case 2:
		compose_forward(msginfo);
		break;
	case 3:
		compose_forward_as_attach(summaryview);
		break;
	default:
		compose_reply(msginfo);
	}
}

static void summary_num_clicked(GtkWidget *button, SummaryView *summaryview)
{
	summary_sort(summaryview, SORT_BY_NUMBER);
}

static void summary_size_clicked(GtkWidget *button, SummaryView *summaryview)
{
	summary_sort(summaryview, SORT_BY_SIZE);
}

static void summary_date_clicked(GtkWidget *button, SummaryView *summaryview)
{
	summary_sort(summaryview, SORT_BY_DATE);
}

static void summary_from_clicked(GtkWidget *button, SummaryView *summaryview)
{
	summary_sort(summaryview, SORT_BY_FROM);
}

static void summary_subject_clicked(GtkWidget *button,
				    SummaryView *summaryview)
{
	summary_sort(summaryview, SORT_BY_SUBJECT);
}

void summary_change_display_item(SummaryView *summaryview)
{
	GtkCList *clist = GTK_CLIST(summaryview->ctree);

	gtk_clist_set_column_visibility(clist, S_COL_MARK, prefs_common.show_mark);
	gtk_clist_set_column_visibility(clist, S_COL_UNREAD, prefs_common.show_unread);
	gtk_clist_set_column_visibility(clist, S_COL_MIME, prefs_common.show_mime);
	gtk_clist_set_column_visibility(clist, S_COL_NUMBER, prefs_common.show_number);
	gtk_clist_set_column_visibility(clist, S_COL_SIZE, prefs_common.show_size);
	gtk_clist_set_column_visibility(clist, S_COL_DATE, prefs_common.show_date);
	gtk_clist_set_column_visibility(clist, S_COL_FROM, prefs_common.show_from);
	gtk_clist_set_column_visibility(clist, S_COL_SUBJECT, prefs_common.show_subject);
}

/* custom compare functions for sorting */

static gint summary_cmp_by_num(GtkCList *clist,
			       gconstpointer ptr1, gconstpointer ptr2)
{
	MsgInfo *msginfo1 = ((GtkCListRow *)ptr1)->data;
	MsgInfo *msginfo2 = ((GtkCListRow *)ptr2)->data;

	return msginfo1->msgnum - msginfo2->msgnum;
}

static gint summary_cmp_by_size(GtkCList *clist,
				gconstpointer ptr1, gconstpointer ptr2)
{
	MsgInfo *msginfo1 = ((GtkCListRow *)ptr1)->data;
	MsgInfo *msginfo2 = ((GtkCListRow *)ptr2)->data;

	return msginfo1->size - msginfo2->size;
}

static gint summary_cmp_by_date(GtkCList *clist,
			       gconstpointer ptr1, gconstpointer ptr2)
{
	MsgInfo *msginfo1 = ((GtkCListRow *)ptr1)->data;
	MsgInfo *msginfo2 = ((GtkCListRow *)ptr2)->data;

	return msginfo1->date_t - msginfo2->date_t;
}

static gint summary_cmp_by_from(GtkCList *clist,
			       gconstpointer ptr1, gconstpointer ptr2)
{
	MsgInfo *msginfo1 = ((GtkCListRow *)ptr1)->data;
	MsgInfo *msginfo2 = ((GtkCListRow *)ptr2)->data;

	if (!msginfo1->fromname)
		return (msginfo2->fromname != NULL);
	if (!msginfo2->fromname)
		return -1;

	return strcasecmp(msginfo1->fromname, msginfo2->fromname);
}

static gint summary_cmp_by_subject(GtkCList *clist,
			       gconstpointer ptr1, gconstpointer ptr2)
{
	MsgInfo *msginfo1 = ((GtkCListRow *)ptr1)->data;
	MsgInfo *msginfo2 = ((GtkCListRow *)ptr2)->data;

	if (!msginfo1->subject)
		return (msginfo2->subject != NULL);
	if (!msginfo2->subject)
		return -1;

	return strcasecmp(msginfo1->subject, msginfo2->subject);
}

static GtkCTreeNode *summary_find_msg_by_msgnum(SummaryView *summaryview,
						guint msgnum)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	GtkCTreeNode *node;
	MsgInfo *msginfo;

	node = GTK_CTREE_NODE(GTK_CLIST(ctree)->row_list);

	for (; node != NULL; node = GTK_CTREE_NODE_NEXT(node)) {
		msginfo = gtk_ctree_node_get_row_data(ctree, node);
                if (msginfo->msgnum == msgnum) break;
        }
        return node;
};

void summary_select_by_msgnum(SummaryView *summaryview, guint msgnum)
{
    GtkCTreeNode *node;
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);

	node = summary_find_msg_by_msgnum(summaryview, msgnum);

	if (node) {
            summary_unselect_all(summaryview);
            gtk_ctree_select(GTK_CTREE(ctree), node);
            if (summaryview->displayed == node)
                summaryview->displayed = NULL;
            summary_display_msg(summaryview, node, FALSE);
        }
}

guint summary_get_current_msgnum(SummaryView *summaryview)
{
	GtkCTree *ctree = GTK_CTREE(summaryview->ctree);
	MsgInfo *msginfo;

	if (!summaryview->selected)
		return 0;
	msginfo = gtk_ctree_node_get_row_data(ctree, summaryview->selected);
	return msginfo->msgnum;
}
