/*
    Stuphead: (C) 2000,2001 Grigroy Bakunov, Sergey Pinaev
 */
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktext.h>
#include <gtk/gtkstyle.h>

#include "intl.h"
#include "viewsource.h"
#include "utils.h"
#include "gtkutils.h"
#include "prefs_common.h"

static void key_pressed(GtkWidget *widget, GdkEventKey *event,
			ViewSourceWindow *viewsourcewin);

static	GdkFont *normalfont = NULL;

#define FONT_LOAD(font, s) \
{ \
	gchar *fontstr, *p; \
 \
	Xstrdup_a(fontstr, s, ); \
	if ((p = strchr(fontstr, ',')) != NULL) *p = '\0'; \
	font = gdk_fontset_load(fontstr); \
}

static void viewsource_init()
{
  if (!normalfont)
    FONT_LOAD(normalfont, prefs_common.messagefont);
}

ViewSourceWindow *viewsource_window_create(void)
{
	ViewSourceWindow *viewsourcewin;
	GtkWidget *window;
	GtkWidget *scrolledwin;
	GtkWidget *text;

	viewsource_init();

	debug_print(_("Creating view source window...\n"));
	viewsourcewin = g_new0(ViewSourceWindow, 1);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), _("Source"));
	gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);
	gtk_widget_set_usize(window, 600, 400);
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
			   GTK_SIGNAL_FUNC(gtk_widget_hide_on_delete), NULL);
	gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
			   GTK_SIGNAL_FUNC(key_pressed), viewsourcewin);
	gtk_widget_realize(window);

	scrolledwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwin),
				       GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
	gtk_container_add(GTK_CONTAINER(window), scrolledwin);
	gtk_widget_show(scrolledwin);

	text = gtk_text_new(gtk_scrolled_window_get_hadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)),
			    gtk_scrolled_window_get_vadjustment
			    (GTK_SCROLLED_WINDOW(scrolledwin)));
	gtk_container_add(GTK_CONTAINER(scrolledwin), text);
	gtk_widget_show(text);

	viewsourcewin->window = window;
	viewsourcewin->scrolledwin = scrolledwin;
	viewsourcewin->text = text;

	return viewsourcewin;
}

void viewsource_window_show(ViewSourceWindow *viewsourcewin)
{
	GtkAdjustment * adj;

	gtk_widget_hide(viewsourcewin->window);
	adj = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(viewsourcewin->scrolledwin));
	if (adj)
	  gtk_adjustment_set_value(adj, 0);
	gtk_widget_show(viewsourcewin->window);
}

void viewsource_window_append(ViewSourceWindow *viewsourcewin,
			      gchar *str)
{
	GtkText *text;

	g_return_if_fail(viewsourcewin != NULL);
	text = GTK_TEXT(viewsourcewin->text);
	gtk_text_freeze(text);
	gtk_text_insert(text, normalfont, NULL, NULL, str, -1);
	gtk_text_thaw(text);
}

static void key_pressed(GtkWidget *widget, GdkEventKey *event,
			ViewSourceWindow *viewsourcewin)
{
	if (event && event->keyval == GDK_Escape)
		gtk_widget_hide(viewsourcewin->window);
}
